#include <memory.h>
#include "SortMerge.h"
#include "mood_alg.h"


#ifdef __cplusplus
# include <libc.h>
#else
  extern char* getenv(char*);
#endif

#define NBRITEMS  	300 	// this value is used to keep track of
			        // number of fields having the equal values
				// and smallest in the segments

int ItemList1[NBRITEMS];	// holds segment number with the smallest 
				// field value

int ItemList2[NBRITEMS];	// same values for the second file
/* Global Decls. 
*/
 
#define OutFile(file, Oid1, Oid2)  fprintf(file,"%d %d\n",&Oid1,&Oid2) 
						          
extern int seqGroup;
extern static OID OidList[SEGMENTSIZE];
extern static void *KeyList[SEGMENTSIZE],*TempKey;
extern static int LengthofKey;   /* for use of SwapLists() */

/* Function Prototypes  */
void JoinInitCompareFunc(int);
extern void ArrangeHeapForward(int,int);
extern void ArrangeHeapBackward(int);
extern void SwapLists(int,int);
int FindJoinItem(JoinFileInfo *,int *,BOOL *);
int ms_CreateSortedSegments(int , int, JoinFileInfo *);


BOOL JoinGreaterInt(int *,int *);
BOOL JoinGreaterFloat(float *,float *);
BOOL JoinGreaterChar(char *,char *);
BOOL JoinGreaterString(char *,char *);
BOOL JoinGreaterOid(OID *, OID *);
BOOL (*Greater)(void *,void *);

BOOL SmallerInt(int *,int *);
BOOL SmallerFloat(float *,float *);
BOOL SmallerChar(char *,char *);
BOOL SmallerString(char *,char *);
BOOL SmallerOid(OID *, OID *);
BOOL (*Smaller)(void *,void *);


BOOL EqualInt(int *,int *);
BOOL EqualFloat(float *,float *);
BOOL EqualChar(char *,char *);
BOOL EqualString(char *,char *);
BOOL EqualOid(OID *, OID *);
BOOL (*Equal)(void *,void *);

int PrepareKeyOidFile(int Volume,int groupIndex,FID GivenFid,FID *NewFid,
                      int KeyStart,int KeyLength, int *TotObjects)
{
  SCANDESC *ScanDesc;
  USERDESC *UserDesc;
  int LBgi;
  OID Oid,NewOid;
  void *Object;
  BOOL Eof;

  if (sm_OpenBufferGroup(MAXNUMOFPAGES,BF_LRU,&LBgi,TRANS_GROUP))
    return esmFAILURE;

  if (sm_OpenScanWithGroup(&GivenFid,LOGICAL_SCAN,LBgi,&ScanDesc,NULL))
    return esmFAILURE;
  
  if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
    return esmFAILURE;

  (*TotObjects)=0; 

  if (!Eof) {
    if (sm_CreateFile(groupIndex,Volume,NewFid))
      return esmFAILURE;
  }
  else {
    printf("No object in the given file!...");
    return esmFAILURE;
  }
  
  while (!Eof) { 
    Object=UserDesc->basePtr;
    Oid=UserDesc->oid;
    if (sm_CreateObject(groupIndex,NewFid,NEAR_LAST_PHYSICAL,NULL,NULL,
                        sizeof(OID),&Oid,&NewOid))
      return esmFAILURE;

    if (sm_AppendToObject(groupIndex,&NewOid,KeyLength,Object+KeyStart))
      return esmFAILURE;

    (*TotObjects)++;

    if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
      return esmFAILURE;
  } /* while */

  if (sm_CloseScan(ScanDesc))
    return esmFAILURE;
 
  return esmNOERROR;
} //PrepareKeyOidFile 

//
// Function: ms_CreateSortedSegments
//
int ms_CreateSortedSegments(int Volume, int groupIndex, 
			    JoinFileInfo * joinFile)
{
  int ObjectSize,TotObjects,LBgi,HSize,i,Count,j;
  FID SortFID,TempFID,CurrFID;
  SCANDESC *ScanDesc;
  USERDESC *UserDesc,*LUDesc;
  BOOL Eof,AllExhausted;
  TID XactId;
  OID TempOid;

  /* Initialization of Compare function "Greater" */ 
  JoinInitCompareFunc(joinFile->FldType); 

  for(i=0;i<SEGMENTSIZE;i++)
    KeyList[i]=malloc(joinFile->FldLen);
  TempKey=malloc(joinFile->FldLen);
  LengthofKey=joinFile->FldLen;
  
  if (sm_BeginTransaction(&XactId))
    return esmFAILURE;

  if (PrepareKeyOidFile(Volume,groupIndex,joinFile->Fid,&TempFID,
			joinFile->FldOffset,
			joinFile->FldLen,&TotObjects)) 
    return esmFAILURE;

  if (sm_OpenBufferGroup(LIMITEDPAGES,BF_LRU,&LBgi,TRANS_GROUP))
    return esmFAILURE;

  if (sm_OpenScanWithGroup(&TempFID,LOGICAL_SCAN,LBgi,&ScanDesc,NULL))
    return esmFAILURE;
  
  if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
    return esmFAILURE;
 
  /* This while loop forms the initial sorted segments */
  joinFile->NumofSegments=0;
  while (!Eof) {
    HSize=0;
    while (!Eof && (HSize < SEGMENTSIZE)) {
      OidList[HSize]=*((OID *)UserDesc->basePtr);
      memcpy(KeyList[HSize],UserDesc->basePtr+12,joinFile->FldLen);
      ArrangeHeapBackward(HSize);
      if (sm_ScanNextObject(ScanDesc,0,READ_ALL,&UserDesc,&Eof))
        return esmFAILURE;
      HSize++;
    } /* while */
    
    /* Create file to output the sorted segments */
    if (sm_CreateFile(groupIndex,Volume,&CurrFID))
      return esmFAILURE;
    /* Record the segment information to the global structure */
    joinFile->Segment[joinFile->NumofSegments].HeapSize=HSize;
    joinFile->Segment[joinFile->NumofSegments].SegFid=CurrFID;
    joinFile->Segment[joinFile->NumofSegments].EofReached=FALSE;
    joinFile->Segment[joinFile->NumofSegments].CurrItem=
					malloc(joinFile->FldLen);
    joinFile->NumofSegments++;

    /* Output the sorted segments */
    for (i=HSize-1;i>=0;i--) {
      /* Write to the file */
      if (sm_CreateObject(groupIndex,&CurrFID,NEAR_LAST_PHYSICAL,NULL,NULL,
                          sizeof(OID),&OidList[0],&TempOid))
        return esmFAILURE;
      if (sm_AppendToObject(groupIndex,&TempOid,joinFile->FldLen,KeyList[0]))
        return esmFAILURE;
      OidList[0]=OidList[i];
      memcpy(KeyList[0],KeyList[i],joinFile->FldLen);
      ArrangeHeapForward(0,i);
    } /* for */
  } /* while */
  
  if (sm_CloseScan(ScanDesc))
    return esmFAILURE;
  if(sm_CommitTransaction(XactId))
	return esmFAILURE;
  for(i=0;i<SEGMENTSIZE;i++)
     free(KeyList[i]);
  free(TempKey);
  return esmNOERROR;
 } 	// ms_CreateSortedSegments 



//------------------------------------------------------------------
// Join  two files 
//------------------------------------------------------------------
int ms_JoinFiles(int Volume, int groupIndex, JoinFileInfo *joinFile1,
		JoinFileInfo *joinFile2,BULK_SCHEMA *RetDescr)
{

  int ObjectSize,TotObjects,LBgi,HSize,i,Count,j;
  FID SortFID,TempFID,CurrFID;
  SCANDESC *ScanDesc;
  USERDESC *UserDesc,*LUDesc;
  BOOL Eof,AllExhausted1,AllExhausted2;
  TID XactId;
  OID TempOid;
  int  e, NbrMatches1,NbrMatches2;
  char *outbuf;


 // allocate segment buffers for files

 
  if((e = sm_CreateFile(seqGroup,Volume,&RetDescr->fid)) != esmNOERROR) {
      return esmFAILURE;
  }

  outbuf = (char *) malloc(RetDescr->record_width);
  joinFile1->Segment=(struct SegmentInfo *)
                        malloc(SEGMENTSIZE*sizeof(struct SegmentInfo));
  if(joinFile1->Segment==NULL)
     return esmFAILURE;


  joinFile2->Segment=(struct SegmentInfo *)
                         malloc(SEGMENTSIZE*sizeof(struct SegmentInfo));
  if(joinFile2->Segment==NULL)
     return esmFAILURE;
   
  // Create sorted segments for first file
  if(ms_CreateSortedSegments(Volume,groupIndex, joinFile1))
	return esmFAILURE;
  

  // Create sorted segments for second file
  if(ms_CreateSortedSegments(Volume,groupIndex, joinFile2))
	return esmFAILURE;

  // open buffer groups for the files containing segments
  for(i=0;i<joinFile1->NumofSegments;i++) {
    if (sm_OpenBufferGroup(LIMITEDPAGES,BF_LRU,
			   &(joinFile1->Segment[i].SegBgi),TRANS_GROUP))
      return esmFAILURE;

    if (sm_OpenScanWithGroup(&(joinFile1->Segment[i].SegFid),LOGICAL_SCAN,
                               joinFile1->Segment[i].SegBgi,
		               &(joinFile1->Segment[i].SegSDesc),NULL))
      return esmFAILURE;
  
    if (sm_ScanNextObject(joinFile1->Segment[i].SegSDesc,0,READ_ALL,&UserDesc,
                        &(joinFile1->Segment[i].EofReached)))
      return esmFAILURE;

    if (!joinFile1->Segment[i].EofReached) {
      memcpy(joinFile1->Segment[i].CurrItem,UserDesc->basePtr+12,
	     joinFile1->FldLen);
      memcpy(&(joinFile1->Segment[i].CurrOid),UserDesc->basePtr,sizeofOID);
    }
  } /* for */

 // same operations with the second file  
  for(i=0;i<joinFile2->NumofSegments;i++) {
    if (sm_OpenBufferGroup(LIMITEDPAGES,BF_LRU,
			   &(joinFile2->Segment[i].SegBgi),TRANS_GROUP))
      return esmFAILURE;

    if (sm_OpenScanWithGroup(&(joinFile2->Segment[i].SegFid),LOGICAL_SCAN,
                               joinFile2->Segment[i].SegBgi,
		               &(joinFile2->Segment[i].SegSDesc),NULL))
      return esmFAILURE;
  
    if (sm_ScanNextObject(joinFile2->Segment[i].SegSDesc,0,READ_ALL,&UserDesc,
                        &(joinFile2->Segment[i].EofReached)))
      return esmFAILURE;

    if (!joinFile2->Segment[i].EofReached) {
      memcpy(joinFile2->Segment[i].CurrItem,UserDesc->basePtr+12,
	     joinFile2->FldLen);
      memcpy(&(joinFile2->Segment[i].CurrOid),UserDesc->basePtr,sizeofOID);
    }
  } /* for */

//
// first segment values have been read  now search for the smallest items
// put the smallest items in ItemLists and then check if there are any matches
// if so, output (oid, oid) pairs into a file 
//
  AllExhausted1=FALSE; 			// end of data indicator for first file
  AllExhausted2=FALSE;			// end of data indicator for second
  while (TRUE) {

    // search the segments  of the first file and find the smallest
    // items (OID, KEY) pairs into Itemlist
    NbrMatches1=FindJoinItem(joinFile1,ItemList1,&AllExhausted1);
   
    // search the segments  of the second file and find the smallest
    // items (OID, KEY) pairs into Itemlist2, 
    NbrMatches2=FindJoinItem(joinFile2,ItemList2,&AllExhausted2);

    if (AllExhausted1 || AllExhausted2) break; // if any of files has been
					       //  consumed return
   
    if(Equal(joinFile1->Segment[ItemList1[0]].CurrItem,
	     joinFile2->Segment[ItemList2[0]].CurrItem)) {
     // field values are equal , there is a match output
     // corresponding Oid pairs into a file
     for(i=0;i<NbrMatches1;i++)
       for(j=0;j<NbrMatches2;j++)
        if ((e = sm_ReadObject(seqGroup,&joinFile1->Segment[ItemList1[i]].CurrOid,0,READ_ALL,&UserDesc)) != esmNOERROR) {
                   return(FALSE);
        }
        memcpy((void *)outbuf,(void *)(UserDesc->basePtr+joinFile1->SelfOff),joinFile1->ObjLen-joinFile1->SelfOff);

        if (( e = sm_ReleaseObject(UserDesc)) != esmNOERROR) {
           return(FALSE);
        }
        if ((e = sm_ReadObject(seqGroup,&joinFile2->Segment[ItemList2[j]].CurrOid,0,READ_ALL,&UserDesc)) != esmNOERROR) {
                   return(FALSE);
        }
        memcpy((void *)(outbuf+joinFile1->ObjLen-joinFile1->SelfOff),(void *)(UserDesc->basePtr+joinFile2->SelfOff),joinFile2->ObjLen-joinFile2->SelfOff);

        if (( e = sm_ReleaseObject(UserDesc)) != esmNOERROR) {
           return(FALSE);
        }
     } 
   // scan next items for the segments for both files 
    for(i=0;i<NbrMatches1;i++) {
      Count=ItemList1[i];
      if (sm_ScanNextObject(joinFile1->Segment[Count].SegSDesc,0,
			  READ_ALL,&UserDesc, 
			  &(joinFile1->Segment[Count].EofReached)))
      return esmFAILURE;
     
      if (!joinFile1->Segment[Count].EofReached) {
        memcpy(joinFile1->Segment[Count].CurrItem,UserDesc->basePtr+12,
	     joinFile1->FldLen);
        memcpy(&(joinFile1->Segment[Count].CurrOid),UserDesc->basePtr,
		sizeofOID);
      }
    } /* for */


    for(i=0;i<NbrMatches2;i++) {
      Count=ItemList2[i];
      if (sm_ScanNextObject(joinFile2->Segment[Count].SegSDesc,0,
			  READ_ALL,&UserDesc, 
			  &(joinFile2->Segment[Count].EofReached)))
      return esmFAILURE;
     
      if (!joinFile2->Segment[Count].EofReached) {
        memcpy(joinFile2->Segment[Count].CurrItem,UserDesc->basePtr+12,
	     joinFile2->FldLen);
        memcpy(&(joinFile2->Segment[Count].CurrOid),UserDesc->basePtr,
		sizeofOID);
      }

    } /* for */



  } /* while */

 // End Join

// destroy segment files and close buffer groups
// 
  for(i=0;i<joinFile1->NumofSegments;i++) {
    if (sm_DestroyFile(joinFile1->Segment[i].SegBgi,
                       &(joinFile1->Segment[i].SegFid)))

      return esmFAILURE;

    if (sm_CloseScan(joinFile1->Segment[i].SegSDesc))
      return esmFAILURE;

    if (sm_CloseBufferGroup(joinFile1->Segment[i].SegBgi))
      return esmFAILURE;
    free(joinFile1->Segment[i].CurrItem);
  } /* for */
 

  for(i=0;i<joinFile2->NumofSegments;i++) {
    if (sm_DestroyFile(joinFile2->Segment[i].SegBgi,
                       &(joinFile2->Segment[i].SegFid)))

      return esmFAILURE;

    if (sm_CloseScan(joinFile2->Segment[i].SegSDesc))
      return esmFAILURE;

    if (sm_CloseBufferGroup(joinFile2->Segment[i].SegBgi))
      return esmFAILURE;
    free(joinFile2->Segment[i].CurrItem);
  } /* for */

  free(joinFile1->Segment);	// free out segment buffers
  free(joinFile2->Segment);    
  if (sm_CommitTransaction(XactId))
    return esmFAILURE; 
  return esmNOERROR;
} // ms_JoinFile 


void JoinInitCompareFunc(int KeyType)
{
  switch(KeyType) {
    case INTEGER   : Greater=&JoinGreaterInt;
		    Equal=&EqualInt;
		    Smaller=&SmallerInt;
                    break;
    case REAL : Greater=&JoinGreaterFloat;
		    Equal=&EqualFloat;
		    Smaller=&SmallerFloat;
                    break;
    case CHAR: Greater=&JoinGreaterChar;
		    Equal=&EqualChar;
		    Smaller=&SmallerChar;
                    break;
    case TEXT: Greater=&JoinGreaterString;
		    Equal=&EqualString;
		    Smaller=SmallerString;
                    break;
    case REF: Greater=&JoinGreaterOid;
		    Equal=&EqualOid;
		    Smaller=SmallerOid;
                    break;
    
    default       : printf("Invalid Key Type!...");
                    exit(1);
  } /* switch */
} /* InitGreaterFunc */

int FindJoinItem(JoinFileInfo * joinFile,int * ItemList,BOOL *AllExhausted)
{
  int i,ItemPlace;
  int Min,Count;

  for(i=0;joinFile->Segment[i].EofReached && i<joinFile->NumofSegments;i++);


  if (i==joinFile->NumofSegments) {
    *AllExhausted=TRUE;
    return(0);
  }


  Min=i; // minimum value is in ith segment curritem;
  ItemPlace=i;
  for(;i<joinFile->NumofSegments;i++)      // find the smallest item
      if(!joinFile->Segment[i].EofReached)
         if(Smaller(joinFile->Segment[i].CurrItem,
	            joinFile->Segment[Min].CurrItem))
            Min=i;

  i=ItemPlace;
  // search for the items equal to the smallest and put them into the
  // ItemList
  Count=0;
  for(;i<joinFile->NumofSegments;i++) {
    if ((!joinFile->Segment[i].EofReached) &&
      (Equal(joinFile->Segment[i].CurrItem,
             joinFile->Segment[Min].CurrItem)))
        ItemList[Count++]=i;
  } /* for */
  return(Count);
} /* FindItem */

BOOL JoinGreaterInt(int *Param1,int *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterInt */

BOOL JoinGreaterFloat(float *Param1,float *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterFloat */

BOOL JoinGreaterChar(char *Param1,char *Param2)
{
  if (*Param1 > *Param2)
    return TRUE;
  else
    return FALSE;

} /* GreaterChar */

BOOL JoinGreaterString(char *Param1,char *Param2)
{
  if (strcmp(Param1,Param2) > 0)
    return TRUE;
  else
    return FALSE;
} /* GreaterString */

BOOL GreaterOID(OID * oid1, OID * oid2)
{

  if(memcmp(oid1,oid2,sizeof(OID)) > 0)
     return TRUE;
  else 
    return FALSE;
}
// smaller and equal functions 

BOOL SmallerInt(int *Param1,int *Param2)
{
  if (*Param1 > *Param2)
    return FALSE;
  else
    return TRUE;
} /* GreaterInt */

BOOL SmallerFloat(float *Param1,float *Param2)
{
  if (*Param1 < *Param2)
    return TRUE;
  else
    return FALSE;
} /* GreaterFloat */

BOOL SmallerChar(char *Param1,char *Param2)
{
  if (*Param1 < *Param2)
    return TRUE;
  else
    return FALSE;

} /* GreaterChar */

BOOL SmallerString(char *Param1,char *Param2)
{
  if (strcmp(Param1,Param2) < 0)
    return TRUE;
  else
    return FALSE;
}

BOOL SmallerOID(OID * oid1, OID * oid2)
{

  if(memcmp(oid1,oid2,sizeof(OID)) < 0)
     return TRUE;
  else 
    return FALSE;
}
// Equal functions 



BOOL EqualInt(int *Param1,int *Param2)
{
  if (*Param1 != *Param2)
    return FALSE;
  else
    return TRUE;
} /* GreaterInt */

BOOL EqualFloat(float *Param1,float *Param2)
{
  if (*Param1 != *Param2)
    return FALSE;
  else
    return TRUE;
} /* GreaterFloat */

BOOL EqualChar(char *Param1,char *Param2)
{
  if (*Param1 != *Param2)
    return FALSE;
  else
    return TRUE;

} /* GreaterChar */

BOOL EqualString(char *Param1,char *Param2)
{
  if (strcmp(Param1,Param2)!=0)
    return FALSE;
  else
    return TRUE;
}

BOOL EqualOID(OID * oid1, OID * oid2)
{

  if(memcmp(oid1,oid2,sizeof(OID))!=0)
     return FALSE;
  else 
    return TRUE;
}
