/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include "mood_catalog.h"
#include "mood_parser.h"
#include "linklist.h"
#include "userreq.h"
#include <X11/cursorfont.h>
#include <string.h>
#include "m_class.h"
#include "libMv.h"

extern int              language;
extern char 		UidPath[100];
extern MenuItem         m_menu[];
extern UserRequest      *userRequest;
extern XmStringCharSet  charset;
extern Widget           top_method, toplevel, drawing_area;
extern Window           win;
extern Display          *dpy;
extern void             methodCB(Widget list_w, Widget clwid, int metflag);

typedef struct AttStr {
    char name[50][150];
    char type[50][150];
} Atts;

LinkedList              *ll;
ListItem                *p;
FunctionType            *ft;
char                    globclname[100], attributes[500];
int                     attmode, myAttMax;
static char             *uid_filenames[1];
static char             *uid_filenames2[1];
static MrmHierarchy     Mrmdata;
static MrmType          *dummy_class;
static MrmRegisterArg   callbackvec[] = { {"attCB",(caddr_t)attCB} };
static MrmCount         callbacknum = sizeof(callbackvec) / sizeof(MrmRegisterArg);
String                  fallback_resources[] = {
                        "*XmText.fontList: -*-courier-medium-r-*--12-*",
                        "*XmLabelGadget.fontList: -*-new century schoolbook-bold-r-*--12-*",
                        NULL };
String                  types[] = { "System Class", "System Type",
                                    "User Type", "User Class",
                                    "Basic Type" };
String                  cons[] = { "No", "Tuple", "Ref", "Set", "List" };
String                  types2[] = { "Systemklasse", "Benutzerklasse",
                                    "Benutzertyp", "Benutzerklasse",
                                    "Basistyp" };

String                  cons2[] = { "Nein", "Tupel", "Refrenz", "Menge", "Liste" };
XtAppContext            app; 
Widget                  topclass, name_w, methodlist_w;
Pcrclst                 newclass;
Atts                    myAtt;
int                     dropmethodflag = 0;

void        ClassCB(Widget w,Pmine client_data,XtPointer call_data);
void        dropmethodCB(); 
static void methoddispCB(Widget list_w, XtPointer client_data, XmListCallbackStruct *cbs);
            mputBlanks( int indent );
void        mprintAttribute( char *attName, int MaxLength );
int         mfindMaxLength( AttrList *aList );
            mprintDescribe( AttrList *aList, int indent );

/*-------------------------------------------------------------
** BuildPulldownMenu - Build a pulldown menu
*------------------------------------------------------------*/
Widget BuildPulldownMenu(Widget parent,char * menu_title,
                         char menu_mnemonic,MenuItem *items)
{
    Widget PullDown, cascade, widget;
    int i;
    XmString str;

    PullDown = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);

    str = XmStringCreateSimple(menu_title);
    cascade = XtVaCreateManagedWidget(menu_title,
        xmCascadeButtonGadgetClass, parent,
        XmNsubMenuId,   PullDown,
        XmNlabelString, str,
        XmNmnemonic,    menu_mnemonic,
        NULL);
    XmStringFree(str);

    /* Now add the menu items */
    for (i = 0; items[i].label != NULL; i++) {
        /* If subitems exist, create the pull-right menu by calling this
         * function recursively.  Since the function returns a cascade
         * button, the widget returned is used..
         */
        if (items[i].subitems)
            widget = BuildPulldownMenu(PullDown, items[i].label,
                                       items[i].mnemonic, items[i].subitems);
        else
            widget = XtVaCreateManagedWidget(items[i].label,
                *items[i].m_class, PullDown, NULL);
        /* Whether the item is a real item or a cascade button with a
         * menu, it can still have a mnemonic.
         */
        if (items[i].mnemonic)
            XtVaSetValues(widget, XmNmnemonic, items[i].mnemonic, NULL);
        /* any item can have an accelerator, except cascade menus. But,
         * we don't worry about that; we know better in our declarations.
         */
        if (items[i].accelerator) {
            str = XmStringCreateSimple(items[i].accel_text);
            XtVaSetValues(widget,
                XmNaccelerator, items[i].accelerator,
                XmNacceleratorText, str,
                NULL);
            XmStringFree(str);
        }
        /* again, anyone can have a callback -- however, this is an
         * activate-callback.  This may not be appropriate for all items.
         */
        if (items[i].callback)
             XtAddCallback(widget,XmNactivateCallback,items[i].callback,
                           items[i].callback_data);
    }
    return cascade;
}

/*-------------------------------------------------------------
** submenu - Call class display for class creation
*------------------------------------------------------------*/
void submenu(Widget widget, int select)
{
  Pmine pass;

  pass = (Pmine) malloc(sizeof(struct mine));
  pass->flag = 1;
  strcpy(pass->clname, "\0");

  switch(select) {
     case 1:
        ClassCB(widget,pass,NULL);
        break;
  }
}

/*-------------------------------------------------------------
** metmenu - Call for method edition and deletion
*------------------------------------------------------------*/
void metmenu(Widget widget,int select)
{
   switch(select) {
     case 1:
        /* Add method */
        methodCB(widget, name_w, 1);
        break;
     case 2:
        /* Drop method */
        dropmethodCB();
        break;
   }
}

/*-------------------------------------------------------------
** dropmethodCB - Drop method
*------------------------------------------------------------*/
void dropmethodCB() {
  ChangeCursor(methodlist_w, XC_pirate);
  dropmethodflag++;
}

/*-------------------------------------------------------------
** createcl - Sql statements for class creation
*------------------------------------------------------------*/
void createcl()
{
  char   sqlbuf[500];
  extern void recreate_drawing_area();
  extern void dag_algorithm();
  extern void dispose();
  extern void draw_onscreen();

  XtDestroyWidget(topclass);

  /* Get class name */
  strcpy(newclass->clname, (char *) XmTextGetString(name_w));

  /* Get Inheritance list
   * ok..ClassCB 
   */

  /*  Get Attributes
   * ok..do_mine 
   */

  /* SQL statement */
  if (strcmp(newclass->superls,"MoodsRoot") == 0)
    sprintf(sqlbuf, "create class %s %s", newclass->clname, newclass->attls);
  else
    sprintf(sqlbuf, "create class %s inherits from %s %s", newclass->clname,
                   newclass->superls, newclass->attls);
  userRequest->executeMSQL(sqlbuf);
  if ((userRequest->operationType == HANDLE_ERROR) &&
      (userRequest->userError.errorType != NO_ERROR)) {
      char *msg = userRequest->getErrorText();
      Display_Error(drawing_area, msg);  
      free(msg);
  } else { /* Success */
      /* Update Display */
      dispose();
      dag_algorithm();
      recreate_drawing_area();
      draw_onscreen();
      XtRealizeWidget(toplevel);
      dpy = XtDisplay(drawing_area);
      win = XtWindow(drawing_area);
      //char *msg = userRequest->getMessageText();
      //strcpy(msg, "Class created succesfully.");
      //Display_Warning(drawing_area, msg);
      //free(msg);
  }
  // Burada goctu
  // userRequest->reset(); 
  free(newclass);
}

/*-------------------------------------------------------------
** ex_prog - Exit from class display
*------------------------------------------------------------*/
void  ex_prog()
{
    XtDestroyWidget(topclass);
}

/*-------------------------------------------------------------
** Menu items for class display 
*------------------------------------------------------------*/
MenuItem method_menu[] = {
    { "Add Method", &xmPushButtonGadgetClass, 'A', NULL, NULL,
        (void *)metmenu, (XtPointer)1, (MenuItem *)NULL },
    { "Drop Method",  &xmPushButtonGadgetClass, 'D', NULL, NULL,
        (void *)metmenu, (XtPointer)2, (MenuItem *)NULL },
     NULL,
};

MenuItem method_menu2[] = {
    { "Methode Hinzufuegen", &xmPushButtonGadgetClass, 'H', NULL, NULL,
        (void *)metmenu, (XtPointer)1, (MenuItem *)NULL },
    { "Methode Entfernen",  &xmPushButtonGadgetClass, 'E', NULL, NULL,
        (void *)metmenu, (XtPointer)2, (MenuItem *)NULL },
     NULL,
};


MenuItem class_menus[] = {
    {"Method Menu", &xmCascadeButtonGadgetClass, 'M', NULL, NULL,
        0, 0, method_menu },
    {"Exit", &xmPushButtonGadgetClass, 'X', NULL, NULL,
        (void *)ex_prog, 0, NULL },
    NULL,
};

MenuItem class_menus2[] = {
    {"Methodenmenue", &xmCascadeButtonGadgetClass, 'M', NULL, NULL,
        0, 0, method_menu2 },
    {"Beenden", &xmPushButtonGadgetClass, 'B', NULL, NULL,
        (void *)ex_prog, 0, NULL },
    NULL,
};


MenuItem create_menus[] = {
    {"Save", &xmCascadeButtonGadgetClass, 'S', NULL, NULL,
        (void *)createcl, 0, NULL },
    {"Exit", &xmPushButtonGadgetClass, 'X', NULL, NULL,
        (void *)ex_prog, 0, NULL },
    NULL,
};

MenuItem create_menus2[] = {
    {"Sichern", &xmCascadeButtonGadgetClass, 'S', NULL, NULL,
        (void *)createcl, 0, NULL },
    {"Beenden", &xmPushButtonGadgetClass, 'B', NULL, NULL,
        (void *)ex_prog, 0, NULL },
    NULL,
};


/*-------------------------------------------------------------
** Same routines with Text Mode
*------------------------------------------------------------*/

/*-------------------------------------------------------------
** mputBlanks -
*------------------------------------------------------------*/
mputBlanks( int indent )
{
   int i;
   for( i=0; i<=indent; i++ ) printf(" ");
}

/*-------------------------------------------------------------
** mprintAttribute -
*------------------------------------------------------------*/
void mprintAttribute( char *attName, int MaxLength )
{
   int strLength;

   // printf("%s", attName );
   strLength = strlen( attName );
   mputBlanks( MaxLength - strLength  + 1 );
}

/*-------------------------------------------------------------
** mfindMaxLength 
*------------------------------------------------------------*/
int mfindMaxLength( AttrList *aList )
{
   AttrList *p;
 
  int maxLength=-1, currLength;
 
  for( p = aList; p; p = p->next )
   {
      currLength = strlen( p->entry.AttributeName );
      if( maxLength < currLength ) maxLength = currLength;
   }
  return maxLength;
}

/*-------------------------------------------------------------
** mprintDescribe - Print attribute list in a linear format
*------------------------------------------------------------*/
mprintDescribe( AttrList *aList, int indent )
{
    int i, inter_indent = indent, maxLength;
    AttrList *p;

    if( !aList ) return TRUE;

    maxLength = mfindMaxLength( aList );
    mputBlanks( indent );

    switch( aList->entry.AttributeConstructor )
        {
                case 'T' :
                        if( !strcmp( aList->entry.AttributeName, " "))
                                inter_indent =  indent + 3;
                        else {
                                mprintAttribute( aList->entry.AttributeName, maxLength );
                                inter_indent = indent + maxLength + 3;
                        }
                        //printf("TUPLE (\n");
                        mprintDescribe( aList->nextsub, inter_indent);
                        mputBlanks( indent );
                        //printf(" )\n");
                        break;
 
                case 'S' :
                        if( aList->nextsub->entry.AttributeConstructor = 'N') {
                                strcpy(myAtt.name[myAttMax], aList->entry.AttributeName);
                                sprintf(myAtt.type[myAttMax], "SET( %s )" ,
                                        aList->nextsub->entry.AttrTypeName );
                                myAttMax++; 
                                //mprintAttribute( aList->entry.AttributeName, maxLength );
                                //printf(" SET( %s ),\n" , aList->nextsub->entry.AttrTypeName );
                        } else {
                                mprintAttribute( aList->entry.AttributeName, maxLength );
                                //printf(" SET( \n");
                                inter_indent = indent + maxLength + 3;
                                mprintDescribe( aList->nextsub, inter_indent );
                                mputBlanks( indent );
                                //printf(" ),");
                        }
                        break;

                case 'L' :
                        if( aList->nextsub->entry.AttributeConstructor = 'N') {
                                strcpy(myAtt.name[myAttMax], aList->entry.AttributeName);
                                sprintf(myAtt.type[myAttMax], "LIST( %s )" ,
                                        aList->nextsub->entry.AttrTypeName );
                                myAttMax++;
                                //mprintAttribute( aList->entry.AttributeName, maxLength );
                                //printf(" LIST( %s ),\n" , aList->nextsub->entry.AttrTypeName );
                        } else {
                                mprintAttribute( aList->entry.AttributeName, maxLength );
                                //printf(" LIST( \n");
                                inter_indent = indent + maxLength + 3;
                                mprintDescribe( aList->nextsub, inter_indent );
                                mputBlanks( indent );
                                //printf(" ),");
                        }
                        break;
 
                case 'R' :
                        if( aList->nextsub->entry.AttributeConstructor = 'N') {
                                strcpy(myAtt.name[myAttMax], aList->entry.AttributeName);
                                sprintf(myAtt.type[myAttMax], "REF( %s )" ,
                                        aList->nextsub->entry.AttrTypeName );
                                myAttMax++;
                                //mprintAttribute( aList->entry.AttributeName, maxLength );
                                //printf(" REF( %s ),\n" , aList->nextsub->entry.AttrTypeName );
                        } else {
                                //printf("%20s", aList->entry.AttributeName );
                                //printf(" REF( \n");
                                inter_indent = indent + maxLength + 3;
                                mprintDescribe( aList->nextsub, inter_indent );
                                mputBlanks( indent );
                                //printf(" ),");
                        }
                        break;
 
                case 'N' :
                        strcpy(myAtt.name[myAttMax], aList->entry.AttributeName);
                        strcpy(myAtt.type[myAttMax], aList->entry.AttrTypeName);
                        myAttMax++;
                        //mprintAttribute( aList->entry.AttributeName, maxLength );
                        //printf( "%s ",aList->entry.AttrTypeName);
                        //printf( ",\n");
      }
 
   mprintDescribe( aList->next, indent );
 
   return TRUE;
}

/*-------------------------------------------------------------
** ClassCB -  Class display 0-Display mode 1-Creation mode
*------------------------------------------------------------*/
void ClassCB(Widget w,Pmine client_data,XtPointer call_data)
{
    Widget        main_wid, main_rowcol, left, main_w, id_w, separator, separator2; 
    Widget        right, first_box, frame, radio_box1, second_box, frame2, body_label;
    Widget        radio_box2, super_label, super_label2, methods_label, menubar;
    Widget        button, main_w2, list_super, list_sub, menu_pane;
    XmStringTable str_list;
    int           i, n;
    char          *text1, subs[20][40], sups[20][40], methods[20][40];
    char          tempstr[3], togstr[20], sqlstm[300];
    Arg           args[6];
    Catalog       mvcat;
    AttrList      *mvattr;
    LinkedList    *mvsuper;
    LinkedList    *mvsub;
    CatBOOLEAN    mvflag;
    ClassNames    *nt, *xt;
    ListItem      *p;
    int           nsub = 0;
    int           nsup = 0;

    newclass = (Pcrclst) malloc(sizeof(struct crclst));
    strcpy(newclass->superls, client_data->clname);

    attmode = client_data->flag;

    if (client_data->flag==0) { /* display mode */  

       /* CAT... Get Class Info. From Kernel */
       MoodsTypes my;
       mvsuper = mvsub = NULL;
       mvattr = NULL;
       mvsuper = new LinkedList();
       mvsub = new LinkedList();

       userRequest->startXAct();
       userRequest->parseText();
       userRequest->handleOperation();
       my.getClassInfo(client_data->clname, &mvcat, &mvattr, mvsuper, mvsub, &mvflag);
       userRequest->commitXAct();

       /* Attribute structure */
       for (i=0; i<50; i++) {
         strcpy(myAtt.name[i], "\0");
         strcpy(myAtt.type[i], "\0");
       }
     
       /* Linearize attributes */
       myAttMax = 0;
       AttrList *al;
       for( al = mvattr; al->next; al=al->next ); /* find last tuple */
       mprintDescribe( al, 0 );

       i = 0;
       for( p = mvsuper->firstItem; p; p=p->next)
            strcpy(sups[i++],(char *) p->item);
       nsup = i;
       i = 0;
       for( p = mvsub->firstItem; p; p=p->next)
            strcpy(subs[i++],(char *) p->item);
       nsub = i;
    } /* End display mode */

    /* Prepare widgets */
    MrmInitialize();
    topclass = XtVaCreatePopupShell(NULL, topLevelShellWidgetClass, 
         drawing_area,
         XmNallowShellResize, True,
         XmNdeleteResponse,   XmUNMAP,
         NULL);
    main_wid = XtVaCreateManagedWidget("main_wid", xmMainWindowWidgetClass,
        topclass,
        NULL);
    menubar = XmCreateMenuBar(main_wid, "menubar", NULL, 0);

    if (client_data->flag==0) {  /* display mode */
         if (language == ENGLISH)
             BuildPulldownMenu(menubar, "Menu", 'M', class_menus);
         if (language == GERMAN)
             BuildPulldownMenu(menubar, "Menue", 'M', class_menus2);
    }
    else {
         if (language == ENGLISH)
             BuildPulldownMenu(menubar, "Menu", 'M', create_menus);
         if (language == GERMAN)
             BuildPulldownMenu(menubar, "Menu", 'M', create_menus2);
    }

    XtManageChild(menubar);
    main_rowcol = XtVaCreateWidget("mainrowcol", xmRowColumnWidgetClass,
         main_wid,
        XmNspacing, 1,
        XmNorientation, XmHORIZONTAL,
        NULL);
    left = XtVaCreateWidget("left", xmRowColumnWidgetClass, 
        main_rowcol,
        XmNorientation, XmVERTICAL,
        XmNspacing, 5,
        NULL);
   
    /* Get uil path */
    uid_filenames[0] = (char *) malloc(100);
    uid_filenames2[0] = (char *) malloc(100);
    strcpy( uid_filenames[0], UidPath);
    strcpy( uid_filenames2[0], UidPath);
    if (language == ENGLISH)
       strcat( uid_filenames[0],"/class.uid");
    if (language == GERMAN)
       strcat( uid_filenames[0],"/class2.uid");

    strcat( uid_filenames2[0],"/att.uid");

    /* Define the Mrm.hierarchy */
    MrmOpenHierarchy( XtNumber(uid_filenames), uid_filenames, NULL, &Mrmdata );
    MrmFetchWidget(Mrmdata,"main_widg", left, &main_w, (short int *) &dummy_class );
    XtManageChild(main_w);
    separator = XtVaCreateWidget ("sep1",xmSeparatorWidgetClass,
       left,
       XmNorientation , XmHORIZONTAL,
       XmNrightAttachment , XmATTACH_FORM,
       NULL);
    XtManageChild(separator);

    /* Assign values of name and id according to kernel info. */
    if (client_data->flag == 0) { /* display mode */
       if (language == ENGLISH)
          name_w = CreateLabeledTextForm(left, "Type Name", "Type Name",
                                         client_data->clname);
       if (language == GERMAN) 
          name_w = CreateLabeledTextForm(left, "Typname  ", "Typname  ",
                                         client_data->clname);
          strcpy(globclname, client_data->clname);
          sprintf(tempstr,"%d", mvcat.TypeId);
       if (language == ENGLISH)
          id_w =   CreateLabeledTextForm(left, "Type Id      ", "Type Id      ", tempstr);
       if (language == GERMAN)
          id_w =   CreateLabeledTextForm(left, "TypBezeichner", "TypBezeichner", tempstr);
    } else { /* creation mode */
       if (language == ENGLISH) {
          name_w = CreateLabeledTextForm(left, "Type Name", "Type Name", "\0");
          id_w =   CreateLabeledTextForm(left, "Type Id      ", "Type Id      ", "\0");
       }
       if (language == GERMAN) {
          name_w = CreateLabeledTextForm(left, "Typname  ", "Typname  ", "\0");
          id_w =   CreateLabeledTextForm(left, "TypBezeichner", "TypBezeichner", "\0");
       }
    }

    /* Assign values of type and constructor according to kernel info. */
    if (language == ENGLISH)
        first_box = XtVaCreateManagedWidget("Class Type:", xmLabelGadgetClass,
        left,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        first_box = XtVaCreateManagedWidget("Klassentyp:", xmLabelGadgetClass,
        left,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    frame = XtVaCreateManagedWidget("frame", xmFrameWidgetClass, 
        left,
        XmNtopAttachment,   XmATTACH_WIDGET,
        XmNtopWidget,       right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    radio_box1 = XmCreateRadioBox(frame, "radio_box1", NULL, 0);
    XtVaSetValues(radio_box1,
        XmNnumColumns,  2,
        XmNpacking,     XmPACK_COLUMN,
        NULL);

    if (client_data->flag==0) { /* display mode */
         switch (mvcat.ClassType) {
            case 'C':
                if (language == ENGLISH)
                    strcpy(togstr, "System Class");
                if (language == GERMAN)
                    strcpy(togstr, "Systemklasse");
                break;
            case 'c':
                if (language == ENGLISH)
                    strcpy(togstr, "User Class");
                if (language == GERMAN)
                    strcpy(togstr, "Benutzerklasse");
                break;
            case 'T':
                if (language == ENGLISH)
                    strcpy(togstr, "System Type");
                if (language == GERMAN)
                    strcpy(togstr, "Systemtyp"); 
                break;
            case 't':
                if (language == ENGLISH)
                   strcpy(togstr, "User Type");
                if (language == GERMAN)
                    strcpy(togstr, "Benutzertyp");
                break;
            case 'B':
                if (language == ENGLISH)
                    strcpy(togstr, "Basic Type");
                if (language == GERMAN)
                    strcpy(togstr, "Basistyp");
                break;
         }
       if (language == ENGLISH) 
         for (i = 0; i < XtNumber(types); i++)
             XtVaCreateManagedWidget(types[i],
                 xmToggleButtonGadgetClass, radio_box1,
                 XmNset, (strcmp(types[i], togstr) == 0),
                 NULL);
       if (language == GERMAN)
          for (i = 0; i < XtNumber(types2); i++)
             XtVaCreateManagedWidget(types2[i],
                 xmToggleButtonGadgetClass, radio_box1,
                 XmNset, (strcmp(types[i], togstr) == 0),
                 NULL);
    } else { /* creation mode */
         if (language == ENGLISH) {
            strcpy(togstr, "User Class");
            for (i = 0; i < XtNumber(types); i++)
           XtVaCreateManagedWidget(types[i], xmToggleButtonGadgetClass, radio_box1,
               XmNset, (strcmp(types[i], togstr) == 0),
               NULL);
         }
         if (language == GERMAN) {
            strcpy(togstr, "Benutzerklasse");
         for (i = 0; i < XtNumber(types2); i++)
           XtVaCreateManagedWidget(types2[i], xmToggleButtonGadgetClass, radio_box1,
               XmNset, (strcmp(types[i], togstr) == 0),
               NULL);
          }
    }

    XtManageChild(radio_box1);
    if (language == ENGLISH)
        second_box = XtVaCreateManagedWidget("Type Constructors:", xmLabelGadgetClass, 
        left,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        second_box = XtVaCreateManagedWidget("Typkonstruktoren:", xmLabelGadgetClass,
        left,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    frame2 = XtVaCreateManagedWidget("frame2", xmFrameWidgetClass, 
        left,
        XmNtopAttachment,   XmATTACH_WIDGET,
        XmNtopWidget,       right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    radio_box2 = XmCreateRadioBox(frame2, "radio_box2", NULL, 0);
    XtVaSetValues(radio_box2,
        XmNnumColumns,  2,
        XmNpacking,     XmPACK_COLUMN,
        NULL);
    for (i = 0; i < XtNumber(cons); i++)
        XtVaCreateManagedWidget(cons[i],
            xmToggleButtonGadgetClass, radio_box2,
            XmNset, (strcmp(cons[i], "Tuple") == 0),
            NULL);
    XtManageChild(radio_box2);
    XtManageChild(left);
    separator2 = XtVaCreateWidget ("sep2",xmSeparatorWidgetClass, 
         main_rowcol,
         XmNorientation , XmVERTICAL,
         XmNleftAttachment , XmATTACH_FORM,
         XmNrightAttachment , XmATTACH_FORM,
         NULL);
    XtManageChild(separator2);
    right = XtVaCreateWidget("right", xmRowColumnWidgetClass, 
           main_rowcol,
           XmNspacing,    6,
           XmNorientation, XmVERTICAL,
           NULL);

    /* Superclasses... */
    if (language == ENGLISH)
        super_label = XtVaCreateManagedWidget("Superclasses:", xmLabelGadgetClass, 
        right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        super_label = XtVaCreateManagedWidget("Superklassen:", xmLabelGadgetClass,
        right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    if (client_data->flag==1) nsup = 0;
    str_list = (XmStringTable)XtMalloc(nsup * sizeof (XmString *));
    for (i = 0; i < nsup; i++) 
        str_list[i] = XmStringCreateLtoR(sups[i], charset);
    list_super = XmCreateScrolledList(right, "super", NULL, 0);
    XtVaSetValues(list_super, XmNitems, str_list,
        XmNitemCount,        nsup,
        XmNvisibleItemCount, 4,
        XmNselectionPolicy,  XmEXTENDED_SELECT,
        NULL);
    XtManageChild(list_super);
    for (i = 0; i < nsup; i++)
        XmStringFree(str_list[i]);
    XtFree((char*)str_list);

    /* Sublasses... */
    if (language == ENGLISH)
        super_label2 = XtVaCreateManagedWidget("Subclasses:",
        xmLabelGadgetClass, right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        super_label2 = XtVaCreateManagedWidget("Subklassen:",
        xmLabelGadgetClass, right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    if (client_data->flag==1) nsub = 0;
    str_list = (XmStringTable)XtMalloc(nsub * sizeof (XmString *));
    for (i = 0; i < nsub; i++)
        str_list[i] = XmStringCreateLtoR(subs[i], charset);
    list_sub = XmCreateScrolledList(right, "sub", NULL, 0);
    XtVaSetValues(list_sub,
        XmNitems, str_list,
        XmNitemCount,        nsub,
        XmNvisibleItemCount, 4,
        XmNselectionPolicy,  XmEXTENDED_SELECT,
        NULL);
    XtManageChild(list_sub);
    for (i = 0; i < nsub; i++)
        XmStringFree(str_list[i]);
    XtFree((char*)str_list);

    /* Methods... */
    n = 0;
    if (client_data->flag==0) { /* Display mode */
        strcpy(sqlstm, "list function of ");
        strcat(sqlstm, client_data->clname);
        userRequest->executeMSQL(sqlstm);
        ll = new LinkedList();
        ll = ((ListFunction *)userRequest->operation)->functionList;
        for (p = ll->firstItem; p; p=p->next){
             ft = (FunctionType *) p->item;
             strcpy(methods[n++], ft->funcName);
        }
    }
    if (language == ENGLISH)
        methods_label = XtVaCreateManagedWidget("Methods:", xmLabelGadgetClass,
        right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        methods_label = XtVaCreateManagedWidget("Methoden:", xmLabelGadgetClass,
        right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    if (client_data->flag==1) n = 0;
    str_list = (XmStringTable)XtMalloc(n * sizeof (XmString *));
    for (i = 0; i < n; i++) str_list[i] = XmStringCreateLtoR(methods[i], charset);
    methodlist_w = XmCreateScrolledList(right, "methods", NULL, 0);
    XtVaSetValues(methodlist_w,
      XmNitems,            str_list,
      XmNitemCount,        n,
      XmNvisibleItemCount, 4,
      XmNselectionPolicy,  XmEXTENDED_SELECT,
      NULL);
    XtAddCallback(methodlist_w, XmNextendedSelectionCallback,
                  (XtCallbackProc)methoddispCB, NULL);
    XtAddCallback(methodlist_w, XmNdefaultActionCallback,
                  (XtCallbackProc) methoddispCB, NULL);
    XtManageChild(methodlist_w);
    for (i = 0; i < n; i++)
        XmStringFree(str_list[i]);
    XtFree((char*)str_list);

    /* Attributes button... */
    if (language == ENGLISH)
        body_label = XtVaCreateManagedWidget("Attributes:", xmLabelGadgetClass, right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
    if (language == GERMAN)
        body_label = XtVaCreateManagedWidget("Attribute:", xmLabelGadgetClass, right,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

    MrmOpenHierarchy( XtNumber(uid_filenames2), uid_filenames2, NULL, &Mrmdata );
    MrmRegisterNames (callbackvec, callbacknum); 
    MrmFetchWidget(Mrmdata,"main_w2", right, &main_w2, (short int *) &dummy_class );

    free(uid_filenames[0]);
    free(uid_filenames2[0]);

    XtManageChild(main_w2);
    XtManageChild(right);
    XtManageChild(main_rowcol);
    XtPopup(topclass, XtGrabNone);
}

/*-------------------------------------------------------------
** methoddispCB - Display or drop method
*------------------------------------------------------------*/
static void methoddispCB(Widget list_w, XtPointer client_data,
                         XmListCallbackStruct *cbs)
{
  char     *text;
  XmString choice;
  char     sqlstm[200];

   XmStringGetLtoR(cbs->selected_items[0], charset, &text);
   choice = XmStringCreateSimple(text);

   for (p = ll->firstItem; p; p=p->next){
        ft = (FunctionType *) p->item;
        if (strcmp(text, ft->funcName) == 0) break;
   }

   if (dropmethodflag == 1) { /* drop method mode */
     /* SQL... Drop method with its signature */
     strcpy(sqlstm, "drop function with signature \'");
     strcat(sqlstm, ft->signature);
     strcat(sqlstm, "\' from class ");
     strcat(sqlstm, globclname);
     userRequest->executeMSQL(sqlstm);
     if ((userRequest->operationType == HANDLE_ERROR) &&
         (userRequest->userError.errorType != NO_ERROR)) {
        char *msg = userRequest->getErrorText();
        Display_Error(drawing_area, msg);
        free(msg);
     } else { /* Success */
        XmListDeleteItem(list_w, choice);
        //char *msg = userRequest->getMessageText();
        //strcpy(msg, "Class dropped successfully.");
        //Display_Warning(drawing_area, msg);
        //free(msg);
     }
     userRequest->reset();
     dropmethodflag=0;
     ChangeCursor(methodlist_w, XC_hand2);
   }      
   else  /* display method */ 
     methodCB(list_w, name_w, 0);

  XtFree(text);
  XmStringFree(choice);
} 
