/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include "mood_catalog.h"
#include "mood_parser.h"
#include <Xm/ToggleBG.h>
#include "libMv.h"
#include "m_class.h"

extern int              language;
extern char             UidPath[100];
extern char             BitMapPath[100];
extern char             *dldfile;
extern FunctionType     *ft;
extern Widget           methodlist_w, toplevel;
extern UserRequest      *userRequest;
extern Widget           drawing_area;
extern void             redraw_button(Widget button, Pixmap pixmap,
                                      XmDrawnButtonCallbackStruct *cbs);

static char             *uid_filenames[1];
static char             *uid_filenames2[1];
static char             *uid_file[1];
static MrmHierarchy     Mrmdata;
static MrmType          *dummy_class;
static MrmRegisterArg   callbackvec[] = {
       {"penCB",(caddr_t)penCB},
       {"pen2CB",(caddr_t)pen2CB},
       {"editorCB",(caddr_t)editorCB}
};
static MrmCount callbacknum = sizeof(callbackvec) / sizeof(MrmRegisterArg);

MenuItem mcreate_menu[] = {
    { "Add Paramater", &xmCascadeButtonGadgetClass, 'A', NULL, NULL,
        (void *)add_para, 0, NULL },
    { "Drop Paramater", &xmCascadeButtonGadgetClass, 'D', NULL, NULL,
        (void *)drop_para, 0, NULL },
    { "Save", &xmCascadeButtonGadgetClass, 'S', NULL, NULL,
        (void *)save_method, 0, NULL },
    { "Exit", &xmPushButtonGadgetClass, 'X', NULL, NULL,
         (void *)ex_menu, 0, NULL },
    NULL,
};

MenuItem mcreate_menu2[] = {
    { "Paramater Hinzufuegen", &xmCascadeButtonGadgetClass, 'H', NULL, NULL,
        (void *)add_para, 0, NULL },
    { "Paramater Loeschen", &xmCascadeButtonGadgetClass, 'L', NULL, NULL,
        (void *)drop_para, 0, NULL },
    { "Sichern", &xmCascadeButtonGadgetClass, 'S', NULL, NULL,
        (void *)save_method, 0, NULL },
    { "Beenden", &xmPushButtonGadgetClass, 'B', NULL, NULL,
         (void *)ex_menu, 0, NULL },
    NULL,
};

MenuItem mdisplay_menu[] = {
    { "Exit", &xmPushButtonGadgetClass, 'X', NULL, NULL,
         (void *)ex_menu, 0, NULL },
    NULL,
};

MenuItem mdisplay_menu2[] = {
    { "Beenden", &xmPushButtonGadgetClass, 'B', NULL, NULL,
         (void *)ex_menu, 0, NULL },
    NULL,
};


char    *para[]  = {""};
char    *appl[]  = {""};
char    *comps[] = {"Set", "List",  "Ref"};
char    *atom[]  = {"int", "float", "double", "char", "string", "boolean"};
Widget  popup, top_method, addprc, mstype, name, para_w;
Widget  name_w, ret_w, classwid;
int     delete_flag = 0;
int     update_flag = 0;
int     update_sec_flag = 0;
char    *text;
GC      gc;
int     toggle_item_set, globmetmode;

void toggled(Widget widget, int which, XmToggleButtonCallbackStruct *state);

/*-------------------------------------------------------------
** methodCB - Display method
*------------------------------------------------------------*/
void methodCB(Widget list_w, Widget clwid, int metmode)
{
  char    *choice;
  Widget  main_method, menubar, appl_w;
  Widget  rowcol, name_w2, met_main_rowcol;
  Widget  method_pane, pen_w;
  Widget  separator, main_w;
  String  spec[]={ "public", "private"};
  Widget  first_box, radio_box, frame;
  Widget  zero, one;
  int     i;
  char    sspec[50];

  classwid = clwid;
  globmetmode= metmode;

  /* Get uil path */
  uid_filenames[0] = (char *) malloc(100);
  uid_filenames2[0] = (char *) malloc(100);
  uid_file[0] = (char *) malloc(100);
  strcpy( uid_filenames[0], UidPath);
  strcpy( uid_filenames2[0], UidPath);
  strcpy( uid_file[0], UidPath);
  strcat( uid_filenames[0],"/pen.uid");
  strcat( uid_filenames2[0],"/pen2.uid");
  if (language == ENGLISH)
      strcat( uid_file[0],"/method.uid");
  if (language == GERMAN)
      strcat( uid_file[0],"/method2.uid");

  /* Prepare widgets */
  top_method = XtVaCreatePopupShell(NULL, topLevelShellWidgetClass,
          XtParent(list_w),
          XmNallowShellResize, True,
          XmNdeleteResponse,   XmUNMAP,
          NULL);
  main_method = XtVaCreateWidget("main_w", xmMainWindowWidgetClass,
                                  top_method, NULL);
  menubar = XmCreateMenuBar(main_method, "menubar", NULL, 0);
  if (metmode == 1) {
       if (language == ENGLISH)
          BuildPulldownMenu(menubar, "Menu", 'M', mcreate_menu);
       if (language == GERMAN)
          BuildPulldownMenu(menubar, "Menue", 'M', mcreate_menu2);
  }
  if (metmode == 0) {
        if (language == ENGLISH)
          BuildPulldownMenu(menubar, "Menu", 'M', mdisplay_menu);
        if (language == GERMAN)
          BuildPulldownMenu(menubar, "Menue", 'M', mdisplay_menu2);
  }
  XtManageChild(menubar);
  method_pane = XtVaCreateWidget("panemethod", xmPanedWindowWidgetClass,
          main_method,
          NULL);
   met_main_rowcol = XtVaCreateWidget("mainrowcol", xmRowColumnWidgetClass,
          method_pane, 
          NULL);
   MrmOpenHierarchy( XtNumber(uid_file), uid_file, NULL, &Mrmdata );
   MrmRegisterNames (callbackvec, callbacknum);
   MrmFetchWidget(Mrmdata,"main_widg", met_main_rowcol, &main_w,
       (short int *) &dummy_class );
   XtManageChild(main_w);
   separator = XtVaCreateWidget ("sep1",xmSeparatorWidgetClass, met_main_rowcol,
          XmNorientation , XmHORIZONTAL,
          XmNrightAttachment , XmATTACH_FORM,
          NULL);
   XtManageChild(separator);

   /* Method Name */
   if (language == ENGLISH)
       name_w = CreateLabeledTextForm(met_main_rowcol, "Method Name ",
                                  "Method Name", " ");;
   if (language == GERMAN)
       name_w = CreateLabeledTextForm(met_main_rowcol, "Methodenname",
                                  "Methodenname", " ");;

   if (globmetmode == 0) XtVaSetValues(name_w, XmNvalue, ft->funcName, NULL);
   rowcol = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass,
          met_main_rowcol,
          XmNorientation, XmHORIZONTAL,
          NULL);

   /* Return Type */
   if (language == ENGLISH)
       ret_w = CreateLabeledTextForm(rowcol, "Return Type   ",
                                 "Return Type   ", " ");
   if (language == GERMAN)
       ret_w = CreateLabeledTextForm(rowcol, "Returntyp     ",
                                 "Returntyp     ", " ");

   if (globmetmode == 0) { /* Display mode */
       sscanf(ft->prototype,  "%s :", sspec);
       XtVaSetValues(ret_w, XmNvalue, ft->retTypeName, NULL);
   }

   MrmOpenHierarchy(XtNumber(uid_filenames), uid_filenames, NULL, &Mrmdata );
   MrmRegisterNames (callbackvec, callbacknum);
   MrmFetchWidget(Mrmdata,"pen_w", rowcol, &pen_w, (short int *) &dummy_class);
   XtManageChild(pen_w);
   XtManageChild(rowcol);

   /* Paramaters */
   if (globmetmode == 1) { /* Creation mode */
       para_w = BuildList(met_main_rowcol, "Parameters:", para, 0);
       XtAddCallback(para_w, XmNextendedSelectionCallback,
                    (XtCallbackProc)list_proc, NULL);
       XtAddCallback(para_w, XmNdefaultActionCallback,
                    (XtCallbackProc)list_proc, NULL);
       XtManageChild(para_w);
       addprc = vir_para(method_pane, "", "");
   }

   /* Specifier Public, Private */
   if (language == ENGLISH)
        first_box = XtVaCreateManagedWidget("Specifier:", xmLabelGadgetClass,
        met_main_rowcol,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);
   if (language == GERMAN)
        first_box = XtVaCreateManagedWidget("Bezeichner:", xmLabelGadgetClass,
        met_main_rowcol,
        XmNrightAttachment,  XmATTACH_FORM,
        NULL);

   frame = XtVaCreateManagedWidget("frame", xmFrameWidgetClass,
        met_main_rowcol,
        XmNtopAttachment,   XmATTACH_WIDGET,
        XmNtopWidget,       para_w,
        XmNrightAttachment, XmATTACH_FORM,
        NULL);
   radio_box = XmCreateRadioBox(frame, "radio_box", NULL, 0);
   XtVaSetValues(radio_box, XmNnumColumns, 2, XmNpacking,XmPACK_COLUMN, NULL);

   if (globmetmode == 0) /* Display mode */
      zero = XtVaCreateManagedWidget(spec[0], xmToggleButtonGadgetClass,
         radio_box,
         XmNset, (strcmp(spec[0], sspec) == 0),
         NULL);

   if (globmetmode == 1) /* Creation mode */
       zero = XtVaCreateManagedWidget(spec[0], xmToggleButtonGadgetClass,
         radio_box,
         XmNset, (strcmp(spec[0],"public") == 0), 
         NULL);

   XtAddCallback(zero, XmNvalueChangedCallback, toggled, 0); 
   one = XtVaCreateManagedWidget(spec[1], xmToggleButtonGadgetClass, radio_box,
      XmNset, (strcmp(spec[1], sspec) == 0),
      NULL);
   XtAddCallback(one, XmNvalueChangedCallback, toggled, 1);
   XtManageChild(radio_box);
   
   free(uid_filenames[0]);
   free(uid_filenames2[0]);
   free(uid_file[0]);

   XtFree(choice);
   XtManageChild(met_main_rowcol);
   XtManageChild(method_pane);  
   XtManageChild(main_method);
   XtPopup(top_method, XtGrabNone);
}

/*-------------------------------------------------------------
** toggled -  callback for toggle buttons
*------------------------------------------------------------*/
void toggled(Widget widget, int which, XmToggleButtonCallbackStruct *state)
{
  if (state->set)
      toggle_item_set = which;
  else
      toggle_item_set = 0;
}

/*-------------------------------------------------------------
** list_proc -
*------------------------------------------------------------*/
void list_proc(Widget list_w, XtPointer client_data, XmListCallbackStruct *cbs)
{
  XmString choice;
  char up_type[30];
  char up_name[30];

  if (delete_flag == 1) {
     XmStringGetLtoR(cbs->selected_items[0], charset, &text);
     choice = XmStringCreateSimple(text);
     XmListDeleteItem(list_w, choice);
     XtFree(text);
     XmStringFree(choice);
     delete_flag=0;
     ChangeCursor(para_w, XC_hand2);
  }
  if (update_flag == 1) {
     XmStringGetLtoR(cbs->selected_items[0], charset, &text);  
     sscanf(text, "%s %s", up_type, up_name);
     XtManageChild(addprc);
     XtVaSetValues(name, XmNvalue, up_name, NULL);
     XtVaSetValues(mstype, XmNvalue, up_type, NULL);
     update_flag = 0;
     update_sec_flag ++;
     ChangeCursor(para_w, XC_hand2);
  }  
}

/*-------------------------------------------------------------
** penCB - Display pencil
*------------------------------------------------------------*/
static void penCB(Widget w, XtPointer client_data,XmListCallbackStruct *cbs)
{
  pencil(0, w, cbs);
}

/*-------------------------------------------------------------
** pen2CB - Display pencil
*------------------------------------------------------------*/
static void pen2CB(Widget w, XtPointer client_data, XmListCallbackStruct *cbs)
{
  pencil(1, w, cbs);
}

/*-------------------------------------------------------------
** pencil - Pencil pixmap
*------------------------------------------------------------*/
void pencil(int select, Widget w,XmListCallbackStruct * cbs)
{
  Widget buttong;
  int i;

  popup = XmCreatePopupMenu(XtParent(w), "popup", NULL, 0);
  for (i=0; i<5; i++) {
     buttong = XtCreateManagedWidget(atom[i], xmPushButtonGadgetClass, popup,NULL, 0);
     if (select == 0)
         XtAddCallback (buttong, XmNactivateCallback,(XtCallbackProc) ChangeRet1,
                       (XtPointer)atom[i]);
     else
         XtAddCallback (buttong, XmNactivateCallback, (XtCallbackProc)ChangeRet2,
                       (XtPointer)atom[i]);
  }
  buttong = XtCreateManagedWidget("xmSep",xmSeparatorWidgetClass, popup,NULL, 0);
  for (i=0; i<3; i++) {
     buttong = XtCreateManagedWidget(comps[i], xmPushButtonGadgetClass, popup, NULL, 0);
     if (select == 0)
         XtAddCallback (buttong, XmNactivateCallback,(XtCallbackProc) ChangeRet1,
                       (XtPointer)comps[i]);
     else
         XtAddCallback (buttong, XmNactivateCallback, (XtCallbackProc)ChangeRet2,
                       (XtPointer)comps[i]);
  }
  XmMenuPosition(popup,(XButtonPressedEvent *)(cbs->event));
  XtManageChild(popup);
}

/*-------------------------------------------------------------
** ChangeRet1 - 
*------------------------------------------------------------*/
static void ChangeRet1(Widget w, char *client_data, XmListCallbackStruct *cbs)
{
  XtVaSetValues(ret_w, XmNvalue, client_data , NULL);
  XtDestroyWidget(popup);
}

/*-------------------------------------------------------------
** ChangeRet2 -
*------------------------------------------------------------*/
static void ChangeRet2(Widget w, char *client_data, XmListCallbackStruct *cbs)
{
  XtVaSetValues(mstype, XmNvalue, client_data , NULL);
  XtDestroyWidget(popup);
}

/*-------------------------------------------------------------
** vir_para - Editor for paramaters
*------------------------------------------------------------*/
Widget vir_para(Widget w, char *def_name, char *def_type)
{
  Widget addprc, pen_w, rowcol_p, rowcol_s;
  void   dialogue();

  addprc = XtVaCreateWidget("frame", xmFrameWidgetClass, w , NULL);
  rowcol_p = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, addprc,
          XmNspacing, 5,
          XmNorientation, XmVERTICAL,
          NULL);
  name = CreateLabeledTextForm(rowcol_p, "Name", "Name", def_name);
  rowcol_s = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, rowcol_p,
          XmNorientation, XmHORIZONTAL,
          NULL);
  mstype = CreateLabeledTextForm(rowcol_s, "Type ", "Type ", def_type);
  MrmOpenHierarchy(XtNumber(uid_filenames2), uid_filenames2, NULL, &Mrmdata );
  MrmRegisterNames (callbackvec, callbacknum);
  MrmFetchWidget(Mrmdata,"pen_w", rowcol_s, &pen_w, (short int *) &dummy_class);
  XtManageChild(pen_w);
  XtManageChild(rowcol_s);
  dialogue(rowcol_p);
  XtManageChild(rowcol_p);
  return(addprc);
}

/*-------------------------------------------------------------
** add_para - Realize paramater editor
*------------------------------------------------------------*/
void add_para()
{
  XtManageChild(addprc);
}

/*-------------------------------------------------------------
** drop_para - Drop a paramater
*------------------------------------------------------------*/
void drop_para()
{
  ChangeCursor(para_w, XC_pirate);
  delete_flag++;
}

/*-------------------------------------------------------------
** save_method - SQL statement for method creation
*------------------------------------------------------------*/
void save_method()
{
 char     sqlstm[500];
 int      i,cnt;
 XmString *strlist = NULL;

 strcpy(sqlstm, "create ");

 /* Get specifier (Public, Private) */
 if (toggle_item_set == 0) strcat(sqlstm, "public ");
 if (toggle_item_set == 1) strcat(sqlstm, "private ");
 strcat(sqlstm, "function for class ");

 /* Get Class Name */
 strcat(sqlstm, (char *) XmTextGetString(classwid));
 strcat(sqlstm, " with prototype \'");

 /* Get return type */
 strcat(sqlstm, (char *) XmTextGetString(ret_w));
 strcat(sqlstm, " ");

 /* Get function name */
 strcat(sqlstm, (char *) XmTextGetString(name_w));
 strcat(sqlstm, "(");

 /* Get paramaters */
 XtVaGetValues(para_w, XmNitemCount, &cnt, XmNitems, &strlist, NULL);
 i = 0;
 while (i < cnt){ 
    XmStringGetLtoR(strlist[i], charset, &text);
    strcat(sqlstm, text);
    i++;
    if (i <cnt)
       strcat(sqlstm, ",");
 }
 strcat(sqlstm, ")\' using \'");

 /* Get filename */
 if (dldfile)
    strcat(sqlstm, dldfile);
 strcat(sqlstm, "\'");

 printf("\n%s\n",sqlstm);
 /* Execute MSQL */
 userRequest->executeMSQL(sqlstm);
 XtDestroyWidget(top_method);
 if ((userRequest->operationType == HANDLE_ERROR) &&
     (userRequest->userError.errorType != NO_ERROR)) {
      char *msg = userRequest->getErrorText();
      printf("\n%s\n",msg);
      Display_Error(drawing_area, msg);
      free(msg);
  } else {
     XmListAddItem(methodlist_w, 
                   XmStringCreateSimple((char *) XmTextGetString(name_w)), 0);
     globmetmode = 0; /* Display mode */
     //char *msg = userRequest->getMessageText();
     //strcpy(msg, "Method created succesfully.");
     //Display_Warning(drawing_area, msg);
     //free(msg);
  }
 userRequest->reset();
}

/*-------------------------------------------------------------
** ex_menu - Exit from method display
*------------------------------------------------------------*/
void ex_menu(Widget widget)
{
  XtDestroyWidget(top_method);
}

/*-------------------------------------------------------------
** make_frame - Create a frame widget
*------------------------------------------------------------*/
Widget make_frame(Widget w)
{
  Widget frame;

  frame = XtVaCreateManagedWidget("frame", xmFrameWidgetClass, w, NULL);
  return(frame);
}

/*-------------------------------------------------------------
** make_label - Create a label widget
*------------------------------------------------------------*/
Widget make_label(Widget w, char *title)
{
  Widget label;

  label = XtVaCreateManagedWidget(title, xmLabelGadgetClass, w, NULL);
  return(label);
}

/*-------------------------------------------------------------
** make_rowcol - Create a rowcolumn widget
*------------------------------------------------------------*/
Widget make_rowcol(Widget w)
{
  Widget rowcol;

  rowcol = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, w,
        XmNorientation, XmVERTICAL,
        NULL);
  return(rowcol);
} 

/*-------------------------------------------------------------
** dialogue - ok, cancel dialogue
*------------------------------------------------------------*/
void dialogue(Widget w)
{
    Widget pb, pb2, rowcol, rowcol2, rowcol3;
    Widget frame, frame2, label, label2;
    Pixel fg, bg;
    Pixmap pixmap, pixmap2;
    char okPath[100];

    rowcol = XtVaCreateWidget("rowcol", xmRowColumnWidgetClass, w,
        XmNorientation, XmHORIZONTAL,
        XmNspacing, 5,
        NULL);
    XtVaGetValues(rowcol, XmNbackground, &fg, XmNbackground, &bg, NULL);
    gc = XCreateGC(XtDisplay(rowcol), RootWindowOfScreen(XtScreen(rowcol)), NULL, 0);
    XSetForeground(XtDisplay(rowcol), gc, fg);
    XSetBackground(XtDisplay(rowcol), gc, bg);
    sprintf(okPath,"%s/ok",BitMapPath);
    pixmap = XmGetPixmap(XtScreen(rowcol), okPath, GetPixel("blue"), bg);
    frame =make_frame(rowcol);
    rowcol2 = make_rowcol(frame);
    pb = XtVaCreateManagedWidget("Ok", xmDrawnButtonWidgetClass, rowcol2,
        XmNwidth,             55,
        XmNheight,            45,
        XmNpushButtonEnabled, True,
        XmNshadowType,        XmSHADOW_ETCHED_OUT,
        NULL);
    label = make_label(rowcol2, "   Ok!");
    XtAddCallback(pb, XmNactivateCallback, (XtCallbackProc)do_it,(XtPointer) 0);
    XtAddCallback(pb, XmNexposeCallback, (XtCallbackProc) redraw_button,(XtPointer) pixmap);
    XtAddCallback(pb, XmNresizeCallback, (XtCallbackProc) redraw_button,(XtPointer) pixmap);
    XtManageChild(rowcol2);
    sprintf(okPath,"%s/cancel",BitMapPath);
    pixmap2 = XmGetPixmap(XtScreen(rowcol), okPath, GetPixel("red"), bg);
    frame2 = make_frame(rowcol);
    rowcol3 = make_rowcol(frame2);
    pb2 = XtVaCreateManagedWidget("pb", xmDrawnButtonWidgetClass, rowcol3,
        XmNwidth,             55,
        XmNheight,            45,
        XmNpushButtonEnabled, True,
        XmNshadowType,        XmSHADOW_ETCHED_OUT,
        NULL);
    label2 = make_label(rowcol3, " Cancel");
    XtAddCallback(pb2, XmNactivateCallback,(XtCallbackProc) do_it,(XtPointer) 1);
    XtAddCallback(pb2, XmNexposeCallback, (XtCallbackProc)redraw_button, (XtPointer)pixmap2);
    XtAddCallback(pb2, XmNresizeCallback, (XtCallbackProc)redraw_button, (XtPointer)pixmap2);
    XtManageChild(rowcol3);
    XtManageChild(rowcol);
}

/*-------------------------------------------------------------
** do_it -
*------------------------------------------------------------*/
void do_it(Widget w,int select,XmListCallbackStruct *cbs)
{
  char *text1 =  (char *) XmTextGetString(mstype);
  char *text2 =  (char *) XmTextGetString(name);
  XmString choice;

  if ((select == 0) && (update_sec_flag)) {
     strcat(text1," ");
     strcat(text1, text2);
     choice = XmStringCreateSimple(text);
     XmListDeleteItem(para_w, choice);
     XmListAddItem(para_w, XmStringCreateSimple(text1), 0);
     XtFree(text1);
     XtFree(text2);
     update_sec_flag = 0;
     XtUnmanageChild(addprc);
  }
  else if (select == 0){ 
     strcat(text1," ");
     strcat(text1, text2);
     XmListAddItem(para_w, XmStringCreateSimple(text1), 0);
     XtFree(text1);
     XtFree(text2);
     XmStringFree(choice);
     update_sec_flag = 0;
     XtUnmanageChild(addprc);
  }
  else if (select)
           XtUnmanageChild(addprc);
} 
