/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <Cursor.h>
#include <Xm/BulletinB.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include "me.h"
#include "userreq.h"
#include "mood_alg.h"

/*-------------------------------------------------------------
**  Global Variables
*/
Widget fn_main_window;
Widget fn_menu_bar,fn_frame;
Widget fn_toplevel,fn_BBoard,fn_Label_BBoard;
extern Widget fn_parent;
extern char *GetRefName(BULK_SCHEMA *, int);
extern BULK_SCHEMA *FindRef(BULK_SCHEMA *, int,int);
extern int find_len(BULK_SCHEMA *bs);
extern void Display_Error(char *);
extern UserRequest *userRequest;
extern int AttributeShowed;


struct Ref_Struct{
     BULK_SCHEMA *bs;
     int ROWNUM;
     int COLNUM;
    };

void DisplayRef (Widget, XtPointer,XtPointer);
int fn_main(BULK_SCHEMA *bs, Widget Parentw);


void DisplayRef (Widget w,XtPointer RSS,XtPointer call_data)
{ BULK_SCHEMA *Rbs; 
  Ref_Struct  *RS;
  char *reqMessage;
  Widget Pw;

   Pw = XtParent(w);
   Pw = XtParent(Pw);
   Pw = XtParent(Pw);
   Pw = XtParent(Pw);
   Pw = XtParent(Pw);
   Pw = XtParent(Pw);

   RS = (Ref_Struct *)RSS; 

   Rbs = FindRef(RS->bs,RS->ROWNUM,RS->COLNUM);

   if (Rbs) {
      if (((int)Rbs) != -1)
         fn_main(Rbs, Pw);
   }
   else {
        reqMessage = userRequest->getErrorText();
        Display_Error(reqMessage);
        free(reqMessage);
   }
}


/*-------------------------------------------------------------
**  CreateApplication  - create main window
*/
int Display_Query_Results(BULK_SCHEMA *bs)
{
  Widget   button;   
  Widget   label;
  Arg      args[20]; 
  int      pos[40], attrlen[40];
  int      RowNum = 0;
  int      ColNum = 0;
  unsigned x,y,i,n,len;
  char     name[50];
  char     Flag;
  char     fn_label[255];
  char*    LabelStr[6];
  XmString label_string;
  XmString title_string;
  MOODSCursor *ResCur;
  BOOL        eof = FALSE;
  XFontStruct *font1, *font2;
  XmFontList  fontlist1;
  Ref_Struct *RS;
  void *mybuf;

  font1 = XLoadQueryFont(XtDisplay(fn_toplevel),
          "-adobe-times-bold-i-normal--14-140-75-75-p-77-iso8859-1");
  if (!font1){
     font1 = XLoadQueryFont(XtDisplay(fn_toplevel),charset);
     fontlist1 = XmFontListCreate(font1, "chset1");
  }
  else
  {
     fontlist1 = XmFontListCreate(font1, "chset1");
  }

     font2 = XLoadQueryFont(XtDisplay(fn_toplevel),
           "-adobe-times-medium-r-normal--12-120-75-75-p-64-iso8859-1");
 
  if (!font2){
     font2 = XLoadQueryFont(XtDisplay(fn_toplevel),charset);
     fontlist1 = XmFontListAdd(fontlist1, font2, "chset2");
   }
  else
  {
     fontlist1 = XmFontListAdd(fontlist1, font2, "chset2");
  }
 

  pos[0] = 20;
  for (i = 0 ; i < bs->arity; i++) {
      if ((strstr(bs->attr[i]->name,"typeId")) ||
          (strstr(bs->attr[i]->name,"self"))) {
          pos[i+1] = pos[i];
          continue;
      }
      switch(bs->attr[i]->attr_type) {
         case REAL : if (strlen(bs->attr[i]->name)<15)
                       attrlen[i] = 15;
                    else
                       attrlen[i] = strlen(bs->attr[i]->name);
                    break;
        case INTEGER: if (strlen(bs->attr[i]->name)<8)
                         attrlen[i] = 8;
                      else
                         attrlen[i] = strlen(bs->attr[i]->name);
                    break;
        case TEXT : 
        case REF  :
        case SET  :
        case LIST :
        case SIGN_VOID :
        case DB_TYPE :
                    attrlen[i] = 20;
                    break;
        default :
                  pos[i+1] = pos[i];
                  continue;
       }

      n = 0;
      label_string = XmStringCreateLtoR(bs->attr[i]->name,"chset1");
      XtSetArg(args[n], XmNfontList,fontlist1);n++;
      XtSetArg(args[n], XmNlabelString,label_string);n++;
      XtSetArg(args[n], XmNwidth,attrlen[i]*8);n++;
      XtSetArg(args[n], XmNx,pos[i]);n++;
      XtSetArg(args[n], XmNy,10);n++;
      XtSetArg(args[n], XmNbackground,fn_GetPixel("pink"));n++;
      XtSetArg(args[n], XmNforeground,fn_GetPixel("red"));n++;
      label = XmCreateLabel(fn_Label_BBoard,"fnlb",args,n);
      XtManageChild (label);
      XmStringFree (label_string);
      pos[i+1] = pos[i]+ 10 + attrlen[i] * 8;
  }

  ResCur = new MOODSCursor;
 
  if (!OpenCursor(bs,ResCur)) {
      return(FALSE);
   }
   if (!GetFirstObject(ResCur,&Flag)) {
       return(FALSE);
   }

  y = 10;

  n = 0;
  XtSetArg (args[n], XmNfontList,fontlist1);n++;
  
  while(!Flag) {
  RowNum++;
  mybuf = (void *)ResCur->CurUserdesc->basePtr;
     for(x=0,i=0,ColNum=0;i<bs->arity;++i,ColNum++) {
      
      if ((strstr(bs->attr[i]->name,"typeId")) ||
          (strstr(bs->attr[i]->name,"self")))
          continue;
      switch(bs->attr[i]->attr_type) {
        case REAL : sprintf(fn_label,"%f\0",
          *(float *)(ResCur->CurUserdesc->basePtr+bs->attr[i]->offset));
                    break;
        case INTEGER: sprintf(fn_label,"%d\0",
            *(int *)(ResCur->CurUserdesc->basePtr+bs->attr[i]->offset));
                    break;
        case TEXT : sprintf(fn_label,"%s\0",
            (char *)(ResCur->CurUserdesc->basePtr+bs->attr[i]->offset));
                    break;
        case REF : sprintf(fn_label,"REF(%s)\0",GetRefName(bs,i));
                   break;
        case SET : sprintf(fn_label,"SET(%s)\0",GetRefName(bs,i));
                   break;
        case LIST : sprintf(fn_label,"LIST(%s)\0",GetRefName(bs,i));
                   break;
        case SIGN_VOID : strcpy(fn_label,"FUNCTION");
                   break;
        case DB_TYPE : sprintf(fn_label,"%s",GetRefName(bs,i));
                   break;
        default :
                   continue;
       }

      label_string = XmStringCreateLtoR(fn_label, "chset2");
      n = 1;
      XtSetArg (args[n], XmNlabelString,label_string);n++;
      // XtSetArg (args[n], XmNwidth, (pos[i+1]-pos[i]-10));n++;
      XtSetArg (args[n], XmNwidth, attrlen[i]*8 );n++;
      XtSetArg (args[n], XmNx, pos[i]-10);n++;
      XtSetArg (args[n], XmNy, y);n++;
      switch(bs->attr[i]->attr_type) {
        case REF  : 
        case SET  :
        case LIST :
        case SIGN_VOID :
        case DB_TYPE :
          button = XtCreateManagedWidget("BTREF",
                   xmPushButtonGadgetClass,fn_BBoard,args,n);
          RS = new Ref_Struct;
          RS->bs = bs;
          RS->COLNUM = ColNum;
          RS->ROWNUM = RowNum;
          XtAddCallback(button,XmNactivateCallback,DisplayRef,RS);
          break;
        default :
          button = XtCreateManagedWidget("pbfn",xmLabelGadgetClass,
                   fn_BBoard,args,n);
          break;
      }
      XmStringFree (label_string);
      }
     y += 25;

  if (!GetNextObject(ResCur,&Flag))
       return(FALSE);
   }
  CloseCursor(ResCur);
  return TRUE;
}
 
/*-------------------------------------------------------------
**  main for displaying query results 
*/
int fn_main(BULK_SCHEMA *bs, Widget Parentw)
{
  Widget swindow,fn_line;
  Arg args[10];
  int n,e,len=0,i;

 fn_toplevel = XtVaCreatePopupShell(NULL,
            topLevelShellWidgetClass, Parentw,
            XmNallowShellResize, True,
            XmNdeleteResponse,   XmUNMAP,
            NULL);

  fn_main_window =XtCreateManagedWidget("main1",xmMainWindowWidgetClass,
                  fn_toplevel,NULL,0);

  fn_create_help();

  n=0;
  XtSetArg(args[n],XmNbackground,fn_GetPixel("LightSkyBlue"));n++;
  XtSetArg(args[n],XmNheight,200);n++;
  XtSetArg(args[n],XmNwidth,300);n++;
  fn_Label_BBoard = XtCreateManagedWidget("fn_Label_BBoard",
                    xmBulletinBoardWidgetClass,fn_main_window,args,n);

  len = find_len(bs);

  n = 0;
  XtSetArg(args[n],XmNx,0);n++;
  XtSetArg(args[n],XmNy,35);n++;
  XtSetArg(args[n],XmNbackground,fn_GetPixel("blue"));n++;
  XtSetArg(args[n],XmNmarginWidth,2);n++;
  XtSetArg(args[n],XmNwidth,len*8+(AttributeShowed-1)*10);n++;
  XtSetArg(args[n],XmNmarginHeight,2);n++;
  XtSetArg(args[n],XmNheight,2);n++;
  XtSetArg(args[n],XmNshadowThickness,1);n++;
  XtSetArg(args[n],XmNshadowType, XmSHADOW_ETCHED_OUT);n++;
  fn_line = XtCreateManagedWidget("fn_line",xmFrameWidgetClass,
            fn_Label_BBoard,args, n);


  //  Create Frame MainWindow and ScrolledWindow in Frame.
  n = 0;
  XtSetArg(args[n],XmNforeground,fn_GetPixel("white"));n++;
  XtSetArg(args[n],XmNy,45);n++;
  XtSetArg(args[n],XmNmarginWidth,2);n++;
  XtSetArg(args[n],XmNmarginHeight,2);n++;
  XtSetArg(args[n],XmNshadowThickness,1);n++;
  XtSetArg(args[n],XmNshadowType, XmSHADOW_ETCHED_OUT);n++;
  fn_frame = XtCreateManagedWidget("fn_frame", xmFrameWidgetClass,
             fn_Label_BBoard,args,n);

  n = 0;
  XtSetArg(args[n],XmNscrollBarDisplayPolicy,XmAS_NEEDED);n++;
  XtSetArg(args[n],XmNscrollingPolicy,XmAUTOMATIC);n++;
  XtSetArg(args[n],XmNwidth,len*8+(AttributeShowed-1)*10+20);n++;
  XtSetArg(args[n],XmNheight,400);n++;
  swindow = XtCreateManagedWidget("swindow",xmScrolledWindowWidgetClass,
            fn_frame,args, n);

  n=0;
  XtSetArg(args[n],XmNbackground,fn_GetPixel("white"));n++;
  fn_BBoard =  XtCreateManagedWidget("fn_BBoard",
               xmBulletinBoardWidgetClass,swindow,args,n);

  if (!Display_Query_Results(bs))
     return FALSE;

  XtPopup(fn_toplevel,XtGrabNone);
  return TRUE;
}
