/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/SelectioB.h>
#include "me.h"

extern Widget toplevel;
extern XmStringCharSet charset;
extern Widget cut_button;           /* Clipboard cut button         */
extern Widget copy_button;          /* Clipboard copy button        */
extern Widget copy_by_name_button;  /* Clipboard copy by name button*/
extern Widget paste_button;         /* Clipboard paste button       */
extern Widget paste_by_name_button; /* Clipboard paste by name button*/
extern Widget open_dialog;          /* File selection dialog         */
extern Widget new_dialog;           /* File name prompt dialog       */
extern Widget save_dialog;   

/*-------------------------------------------------------------
**  GetPixel - Convert string to pixel
*------------------------------------------------------------*/
XtArgVal GetPixel(char *colorstr)
{
   XrmValue from, to;

   from.size = strlen(colorstr) +1;

       if (from.size < sizeof(String))
           from.size = sizeof(String);

   from.addr = colorstr;
   to.addr = NULL;
   XtConvert(toplevel, XmRString, &from, XmRPixel, &to);

   return ((XtArgVal) *((XtArgVal *) to.addr));
}

/*-------------------------------------------------------------
**  DialogApplyCB
**    Process callback from Dialog apply actions.
*------------------------------------------------------------*/
void DialogApplyCB (Widget w, XtPointer client_data,XtPointer  call_data)
{
  switch ((int)client_data)
  {
    case DIALOG_CWARNING:
      CloseFile();
      file_saved = True;
      break;

    case DIALOG_XWARNING:
      CloseFile();
      exit(1);
      break;

    default:
      fprintf (stderr, "Warning: in apply callback\n");
      break;
  }
}

/*-------------------------------------------------------------
**  DialogCancelCB
**    Process callback from Dialog cancel actions.
*------------------------------------------------------------*/
void DialogCancelCB (Widget w, XtPointer client_data,XtPointer  call_data)
{
  switch ((int)client_data)
  {
    case DIALOG_FSELECT:
        XtUnmanageChild (open_dialog);
      break;

    case DIALOG_CWARNING:
    case DIALOG_XWARNING:
    case DIALOG_NEW:
    case DIALOG_SAVE:
      // No action is necessary at this time
      break;

    default:
      fprintf (stderr, "Warning: in cancel callback\n");
      break;
  }
}

/*-------------------------------------------------------------
**  DialogAcceptCB-  Process callback from Dialog actions.
*------------------------------------------------------------*/
void DialogAcceptCB (Widget w, XtPointer client_data,XtPointer  call_data)
{
  switch ((int)client_data)
  {
    case DIALOG_FSELECT:
            // Open the file and read it into the text widget
      if (filename != NULL) {
         XmStringFree((XmString)filename);
         filename = NULL;
      }
      {
         XmFileSelectionBoxCallbackStruct *fcb =
         (XmFileSelectionBoxCallbackStruct *) call_data;

         // Get the filename from the file selection box
         XmStringGetLtoR(fcb->value, charset, &filename);

         // Open file, print error if it does not exist.
         if (!OpenFile())
               fprintf(stderr, "Warning: unable to open file\n");

         // Popdown the file selection box
           XtUnmanageChild (open_dialog);
      }
      break;

    case DIALOG_NEW:
            // Open the file and read it into the text widget
      if (filename != NULL) {
         XmStringFree((XmString)filename);
         filename = NULL;
      }
      {
         XmSelectionBoxCallbackStruct *scb =
         (XmSelectionBoxCallbackStruct *) call_data;

         XmStringGetLtoR(scb->value, charset, &filename);

 if (!OpenFile()) {
              // make appropriate item sensitive
              XtSetSensitive(text, True);
              XtSetSensitive(cut_button, True);
              XtSetSensitive(copy_button, True);
              XtSetSensitive(copy_by_name_button, True);
              XtSetSensitive(clear_button, True);
         }
           XtUnmanageChild (new_dialog);
      }
      break;

    case DIALOG_CWARNING:
      if (SaveFile()) {
         CloseFile();
      } else
         fprintf(stderr,
             "Warning: unable to save file, file not closed");
      break;

    case DIALOG_XWARNING:
      if (SaveFile()) {
         CloseFile();
         exit(0);
         } else
         fprintf(stderr,
         "Warning: unable to save file, exit aborted");
      break;

    case DIALOG_SAVE:
      {
         XmSelectionBoxCallbackStruct *scb =
         (XmSelectionBoxCallbackStruct *) call_data;

         // get the filename string from the file selection box
         XmStringGetLtoR(scb->value, charset, &filename);

         SaveFile();

         XtUnmanageChild (save_dialog);
      }
      break;

    default:
      fprintf (stderr, "Warning: in accept callback\n");
      break;
  }
}
/*-------------------------------------------------------------
**  CreateSpecialWarningDialog
**    Create special 4 button message box out of a
**  Selection box.
*------------------------------------------------------------*/
Widget CreateSpecialWarningDialog (Widget parent, String name,String  image_string,String  message,
ArgList  arglist, int argcount)
{
  Widget        warning_dialog; /*  Special warning selection box */
  Widget        work_area;      /*  Rowcolumn for pixmap and text */
  Widget        pixmap_label;   /*  Pixmap label                  */
  Widget        text_label;     /*  Text label                    */
  Widget        apply_button;   /*  Apply button                  */
  Widget        ok_button;      /*  Ok button                     */
  Widget        kid[5];         /*  Buttons                       */
  Pixel         foreground;     /*  Dialog foreground             */
  Pixel         background;     /*  Dialog background             */
  Pixmap        pixmap;         /*  Dialog pixmap                 */
  Arg           args[10];       /*  Arg list                      */
  register int  i;              /*  Kid index                     */
  register int  n;              /*  Arg count                     */


  warning_dialog = XmCreatePromptDialog(parent, name, arglist, argcount);

  n = 0;
  XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
  work_area = XmCreateRowColumn(warning_dialog, "workarea", args, n);
  XtManageChild(work_area);

  n = 0;
  XtSetArg(args[n], XmNforeground, &foreground); n++;
  XtSetArg(args[n], XmNbackground, &background); n++;
  XtGetValues(warning_dialog, args, n);

  n = 0;
  XtSetArg(args[n], XmNlabelType, XmPIXMAP); n++;
  pixmap = XmGetPixmap(XtScreen(warning_dialog), image_string,
           foreground, background);
  XtSetArg(args[n], XmNlabelPixmap, pixmap); n++;
  pixmap_label = XmCreateLabel(work_area, "pixmap_label", args, n);
  XtManageChild(pixmap_label);

  n = 0;
  XtSetArg(args[n], XmNlabelString,XmStringCreateLtoR(message, charset)); n++;
  text_label = XmCreateLabel(work_area, "text_label", args, n);
  XtManageChild(text_label);

  apply_button = XmSelectionBoxGetChild (warning_dialog,XmDIALOG_APPLY_BUTTON);

  n = 0;
  XtSetArg(args[n], XmNlabelString,
  XmStringCreateLtoR("Discard", charset)); n++;
  XtSetValues(apply_button, args, n);
  XtManageChild(apply_button);

  ok_button = XmSelectionBoxGetChild (warning_dialog,XmDIALOG_OK_BUTTON);
  n = 0;
  XtSetArg(args[n], XmNlabelString,XmStringCreateLtoR("Save", charset)); n++;
  XtSetValues(ok_button, args, n);


        //      Unmanage unneeded children.
        i = 0;
        kid[i++] = XmSelectionBoxGetChild (warning_dialog, XmDIALOG_TEXT);
        kid[i++] = XmSelectionBoxGetChild (warning_dialog,
                 XmDIALOG_SELECTION_LABEL);
        XtUnmanageChildren (kid, i);

  return(warning_dialog);
}
