/**********************************************************************
* MoodView ,The Graphical User Interface To MOODS
* Copyright (c) 1994 Scientific and Technical Research Council of Turkiye
*                    Software Research and Development Center, Ankara
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof, and that both notices appear in supporting documentation.
*
* THE SOFTWARE RESEARCH AND DEVELOPMENT CENTER ALLOWS FREE USE OF THIS
* SOFTWARE IN ITS "AS IS" CONDITION.
* THE CENTER DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES
* WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
*
* The MOODS Project Group requests users of this software to return
* any improvements or extensions that they make to:
*
*   MOODS Project Group
*   TUBITAK Software Research and Development Center
*   Department of Computer Engineering
*   Middle East Technical University
*   Ankara, 06531
*
*        or budak@karakoy.srdc.metu.edu.tr
*
* In addition, the MOODS Project Group requests that users grant the
* Software Research and Development Center rights to redistribute these
* changes.
**********************************************************************/

#include <stdio.h>
#include <Xm/Xm.h>
#include <Xm/CutPaste.h>
#include <Xm/Text.h>
#include "me.h"

int m_private=0;  // private id for clipboard copy by name
char Error[128];
extern XmStringCharSet charset;
extern Widget paste_by_name_button;
extern Widget paste_button;

/*-------------------------------------------------------------
**  CopyByName
**    Copy data item to the clipboard by name.
*------------------------------------------------------------*/
void CopyByName(Widget widget,unsigned long * dataid,int * privateid, int *reason)
{
  int buffer_size, status;

  if (*reason == XmCR_CLIPBOARD_DATA_REQUEST) {
   buffer_size = strlen (text_buffer);
   status = XmClipboardCopyByName (XtDisplay(text), XtWindow(text),
    *dataid, text_buffer, buffer_size, *privateid);
     if (status == ClipboardLocked) {
     printf ("xmeditor procedure CopyByName: Clipboard is locked.");
     return;
     }
  }
  if (*reason == XmCR_CLIPBOARD_DATA_DELETE)
   printf ("Clipboard data delete request occurred.");
}

/*-------------------------------------------------------------
**  CopyFileToClipboard
**    Copy the present file to the clipboard.
*------------------------------------------------------------*/
void CopyFileToClipboard(Time time)
{
   char *selected_string = XmTextGetSelection (text); /* Text selection    */
   long item_id = 0;     /* Clipboard item id */
   int data_id = 0;              /* Clipboard data id */
   int status = 0;              /* Clipboard status  */
   XmString clip_label;

   if (by_name) {
   // Copy the selected by name text to the clipboard 
   by_name = False;
   text_buffer = XmTextGetString (text);  // Text in the text widget 
   m_private ++;  // Unique data name used in call to XmClipboardCopy 
  clip_label = XmStringCreateLtoR ("MOODSQL", charset);
  /* Start copy to clipboard, and continue till
     a successful start copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardStartCopy (XtDisplay(text), XtWindow(text),
                 clip_label, time,
                 (Widget)text,(VoidProc) CopyByName, (long*)&item_id);

     /* XmClipboardCopy will put the name of the data on the clipboard */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardCopy (XtDisplay(text), XtWindow(text),
              item_id, "STRING", NULL, 0, m_private, &data_id);
  /* End the copy to the clipboard and continue till
     a successful end copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardEndCopy (XtDisplay(text), XtWindow(text),
           item_id);
  /* Allow pasting when an item is successfully copied to the clipboard */
  XtSetSensitive(paste_by_name_button, True);
   }
   else {     /* Copy is not by name */
   /* Copy the selected text to the clipboard */
   if (selected_string != NULL) {
  clip_label = XmStringCreateLtoR ("XM_EDITOR", charset);
  /* Start copy to clipboard, and continue until
     a successful start copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardStartCopy (XtDisplay(text), XtWindow(text),

                 clip_label, time,
                 text, NULL, &item_id);

     /* Move the data to the clipboard, and
     continue until a successful copy is made */
  status = 0;
  while (status != ClipboardSuccess)
      status = XmClipboardCopy (XtDisplay(text), XtWindow(text),
              item_id, "STRING", selected_string,
              (long)strlen(selected_string)+1, 0, &data_id);

  /* End the copy to the clipboard and continue until
     a successful end copy is made */
  status = 0;
  while (status != ClipboardSuccess)
     status = XmClipboardEndCopy (XtDisplay(text), XtWindow(text),item_id);

  /* Allow pasting when an item is successfully copied to the clipboard */
  XtSetSensitive(paste_button, True);
     }
   }
}

/*-------------------------------------------------------------
**  PasteItemFromClipboard
**    Paste item from the clipboard to the current cursor location
*------------------------------------------------------------*/
void PasteItemFromClipboard()
{
   /* Retrieve the current data from the clipboard
      and paste it at the current cursor position */
   char * selected_string = XmTextGetSelection (text);
                                      /* Contents of selection   */
   int status = 0;                    /* Clipboard status        */
   char *buffer;                      /* Temporary text buffer   */
   unsigned long length;                        /* Length of buffer        */
   unsigned long outlength = 0;                 /* Length of bytes copie   */
   int private_id = 0;                /* Id of item on clipboard */
   XmTextPosition cursorPos;          /* Text cursor position    */
   register int n;                    /* Arg count               */
   Arg args[10];                      /* Arg list                */

   /* Find the length of the paste item, continue until the length is found */
   while (status != ClipboardSuccess) {
     status = XmClipboardInquireLength(XtDisplay(text), XtWindow(text),
              (char *)"STRING", &length);
     if (status == ClipboardNoData) {
  length = 0;
  break;
     }
   }

   if (length == 0) {
  fprintf(stderr, "Warning: paste failed, no items to paste.\n");
  return;
   }

   /* Malloc to necessary space */
   buffer = XtMalloc(length);

   status = XmClipboardRetrieve (XtDisplay(text), XtWindow(text), "STRING",
         buffer, (unsigned long)length, &outlength, &private_id);

   /* Error in pasting */
   if (status != ClipboardSuccess) {
  fprintf(stderr, "Warning: paste failed, status = %d\n", status);
  return;
   }

   /* Get cursor position for pasting */
   XtSetArg(args[0], XmNcursorPosition, &cursorPos);
   XtGetValues(text, args, 1);

   /* Add new text */
   XmTextReplace(text, cursorPos, cursorPos, buffer);
}

/*-------------------------------------------------------------
**  PasteByName
**    paste by name from the clipboard to the current cursor location
*------------------------------------------------------------*/
void PasteByName()
{
   int status = 0;           /* Clipboard status        */
   char *buffer;             /* Temporary text buffer   */
   unsigned long length;               /* Length of buffer        */
   unsigned long outlength = 0;        /* Length of bytes copied  */
   int private_id = 0;       /* Id of item on clipboard */
   XmTextPosition cursorPos; /* Text cursor position    */
   register int n;           /* Arg count               */
   Arg args[10];             /* Arg list                */

   /* find the length of the paste item, continue until the length is found */
   length = strlen(text_buffer);
   if (length == 0) {
  fprintf(stderr, "Warning: paste by name failed, no items to paste.\n");
  return;
   }

   /* Malloc to necessary space */
   buffer = XtMalloc(length);

  status = XmClipboardRetrieve (XtDisplay(text), XtWindow(text),
           (char *)"STRING",(char *)buffer, (unsigned long)length, &outlength, &private_id);

   /* Error in pasting */
   if (status != ClipboardSuccess) {
  fprintf(stderr, "Warning: paste by name failed, status = %d\n", status);
  return;
   }

   /* Get cursor position for pasting */
   XtSetArg(args[0], XmNcursorPosition, &cursorPos);
   XtGetValues(text, args, 1);

   /* Add new text */
   XmTextReplace(text, cursorPos, cursorPos, buffer);
}

/*-------------------------------------------------------------
**  VersionCBQ      - callback for help button
*------------------------------------------------------------*/
void VersionCBQ (Widget w,XtPointer client_data,XtPointer call_data)
{
  Widget    message_box;    /*  message box    */

  /*  Create message window. */
  message_box = CreateVersionQ (w);

  /*  Display version window. */
  XtManageChild (message_box);
}
