#ifndef __MOODS_CATALOG
#define __MOODS_CATALOG

#define sm_check( _errNo, _errorMsg );\
	if( err ){ ur->setError( EXODUS_ERROR, _errNo, _errorMsg ); \
		   return FALSE; }\

// external definitions 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setjmp.h>
#include <sm_client.h>
#include "mood_catdef.h"
#include "mood_error.h"
#include "mood_DBrecord.h"


// external definitions 
extern  FILE  *sm_ErrorStream ;
extern  VOLID VolumeId ;
extern  TID   TransactionId;
extern  int   BufferGroup;
extern  FID   FileIdentifier;
extern  FID   DBFile;
extern  FID   CatalogID;
extern  FID   UnsizedFID;
extern  FID   RefFID;
extern  IID   RefIID;
extern  OBJHDR ObjectHeader;
extern  FILE  *ErrStrm;
extern  char noDBInUse;
extern  MoodsError catalogError;
extern  struct DBrecord DBrec;

extern  ERTYPE ErrorCheck(int e,char *routine) ;
extern  int Test_Oid( OID Pfirst , OID Psecond , int Pcondition ) ;
extern  void display_Oid(OID oid) ;
extern  void read_Oid(OID *oid);
extern  int ESM_Startup() ;
extern  int ESM_ShutDown() ;
extern  int ms_LocateObject( FID *, int, char *, int, OID *, int = -1 );
extern  char *MoodHome;
extern  char *MoodBin;
extern  char DatabaseHome[];
extern  FILE *myalloc;
extern  FILE *dealloc;
extern	int  interfaceMode;
extern  char *myAlloc( int );
extern  myFree( char *);

OID     nullOid;
FID     nullFid;
OID     rootOid;

class CatRoot
{
public :
	int	__typeId;
};

/* protection flags */
typedef enum {
                cpPUBLIC,
                cpPRIVATE,
                cpPROTECTED
             } protFlags ;


/* the structure of the inheritance list */
struct inherits{
   char                 cname[MAXSTRLEN];
   char                 virtualFlag;
   protFlags            protectionFlag; // PUBLIC or PRIVATE
   struct inherits      *next;
};

// data structure for class definition
struct memberNode{
    int                 type;
    int                 size;
    char                memberName[MAXSTRLEN];
    char 		typeName[MAXSTRLEN];
    struct memberNode   *subAttributes;
    struct memberNode   *attributeList;
};

class LinkedList;
class List;
class SortedStringList;
class MoodsTypes;
//class UserRequest;


class CreateHeader{
private :
         FILE   *whichFile;
	void manipulateset(memberNode *);
	void manipulatelist(memberNode *);
	void maketpldef(memberNode *);
	void traverse(memberNode *);
	void writeinheritList(LinkedList *);
	void defInheritList(LinkedList *);
	char *generateName();
	void itoa( int, char * );
public:
	void createheaderfile( char *, LinkedList *, memberNode * );
};


struct ForwardBackward{
	IID   forward;
	IID   backward;
};


class SeqTemp
{
public :
	char 	seqname[SEQ_MAXNAME];
	int  	minval ;
	int  	maxval ;
	char 	cyclic ;
	int  	initial;
	int  	increment;
	int  	curval ;
	OID  	SeqId ;
};

              
class Sequence 
{
private : 
	char 		seqname[SEQ_MAXNAME] ;
	int  		minval ;
	int  		maxval ;
	char 		cyclic ;
	int  		initial;
	int  		increment;
	int  		curval ;
public :
	OID  		SeqId ;
	SeqTemp  	SeqIntern ;
	USERDESC 	*UserDesc;


		Sequence(){}
	//int	initSequence(FID *, char *, int , int , int, char ,  int ,CatBOOLEAN *);
	int	initSequence( FID *, char *, CatBOOLEAN *);
	int  	getNextVal(int *, CatBOOLEAN *PStatus);
	int  	getCurVal( int *);
	int 	setCurVal(int ,CatBOOLEAN *);
	int 	setMaxVal(int ,CatBOOLEAN *);
	int 	setMinVal(int ,CatBOOLEAN *);
	int 	setCycOn();
	int 	setCycOff() ;
	int 	setInitial(int );
	int 	setIncr(int ,CatBOOLEAN *) ;
	int 	delSeq(CatBOOLEAN *);
};

class Database
{
public :
        char            dbname[MAXDBNAME];
        FID             FCfile;   //  associated catalog file .
                                //  get or create db entry . see definition !!
	FID		Unsizedfile;
                        Database(char * , FID * , CatBOOLEAN *);
                        Database();

        void            initSeq( SeqTemp *, char * );
        int             useDatabase( char * );
        int             createDatabase( char * );
        int             deleteDatabase( char * );
        int             listDatabases( LinkedList *);
        int             initDB();
	int		inUse();	
	void		setUnused();
        int             createType(MoodsTypes *, char *,int, char , OID *, OID *);

                       // delete db & assoc. information see definiton !!
        void            deleteDB(CatBOOLEAN *) ;

};


// this definition is needed for exodus storage manager....
class AttributeType : public CatRoot
{
public :
	char	AttributeName[MAXSTRLEN];
	//OID     AttributeType;
	OID     AttrType;
	char    AttrTypeName[MAXSTRLEN];
	int     AttributeOffset;
	int     AttributeLength;
	char    AttributeConstructor;
	char    ClassMember;
	AttributeType(){ CatRoot::__typeId = TYPE_ATTRIBUTE; }
};

class AttrList{
public :
        AttributeType	entry;
        AttrList	*next;
        AttrList	*nextsub;
			AttrList();
			~AttrList();
};
    
class CatATTRIBUTE 
{
public:
	USERDESC 	*UserDesc ;
	FID      	AttributeFid;

//		CatATTRIBUTE(FID ,OID *); // constructor for Attribute..
		CatATTRIBUTE(FID);        // constructor for Attribute..
	int 	writeAttr(AttributeType ,OID *);
	int 	readAttr(OID ,AttributeType *, CatBOOLEAN *);
	void 	disp_attr(AttributeType );
	int 	updateAttr( OID, AttributeType );

};


class Catalog : public CatRoot
{
public :
	char 	TypeName[MAXSTRLEN];
	int  	TypeId;
	int  	Size;
	char 	ClassType; // 'C' system class , 'c' user class, 
		             // 'T' system type, 't' user type
			          // 'B' basic type....
	OID	List_Attributes;
	OID	List_SuperClass;
	OID	List_SubClass;
	OID	List_Function;
	FID	Instance_File;
	int	Cardinality;
	int	RefCount; // for unnamed type only

	Catalog() { CatRoot::__typeId = TYPE_CATALOG; }
};

class CatalogTypeItem
{
public :
	char	TypeName[MAXSTRLEN];
	int	TypeId;
	int	Size;
	char	ClassType;
		CatalogTypeItem( Catalog );
		~CatalogTypeItem();
};

class FunctionType : public CatRoot
{
public :
        int   retType;
        char  funcName[100];
        int   isMember; // 0 not member, 1 member
        char  className[100];
        char  signature[100];
        int   argc;
	char  retTypeName[32];
	char  prototype[200];

	FunctionType();
	~FunctionType();
};
        

class LevelInfo 
{
public :
	char		ClassName[MAXSTRLEN] ;
	LevelInfo 	*next ;
			LevelInfo();
			~LevelInfo();
};

class Hierarchy
{
public :
	LevelInfo 	*thislevel;
	Hierarchy     	*nextlevel;
			Hierarchy();
			~Hierarchy();
};

class ClassNames
{
public :
	char 		Name[MAXSTRLEN] ;
	ClassNames 	*next ;
			ClassNames();
			~ClassNames();
};

class MoodsTypes
{
private:
	FID  		CatalogId ;
	FID  		AttributeId ;
	FID  		FunctionId ;
	FID  		InstanceId;
	FID		SequenceId;	
	IID		bjoinindices;
	char		seqName[MAXSTRLEN];
         
public :
	USERDESC 	*UserDesc ;
	char 		typeName[MAXSTRLEN];
	int  		typeId;
	int  		typeSize;
  //         int  	classVariables;  
	char 		classType;  //   System class | User Class | User typ 
		              //   | Basic typ | Sys typ
	char 		constructor; //  Tuple,set,list,ref,noconstructor
	OID  		attributes;  //  Attribute Oid's
	OID  		subClasses;  //  
	OID  		superClasses;//

		//MoodsTypes(FID );
		MoodsTypes();
	int	init( FID );
	int 	addType( char *, LinkedList *,  memberNode *, int , CatBOOLEAN *);
	int	createEmptyClass( char * );
	int 	fillEmptyClass( char *, LinkedList *, memberNode *, int, CatBOOLEAN * );
	int 	MoodsAttr( memberNode *, int *, List *, CatBOOLEAN *) ;
	int 	createUnnamedType( memberNode *, int *, OID *, CatBOOLEAN *);
	int 	setClassType( char *, int );
	int 	getTypOid( char *, OID *, CatBOOLEAN *);
	int 	getTypOid( int   , OID *, CatBOOLEAN *);
	int 	copyInheritedTypes( OID *, List *, int *, CatBOOLEAN * );
	int  	findType( char *, CatBOOLEAN * );
	int	checkReference( LinkedList *, SortedStringList *);
	int 	createInstanceFile( FID *, CatBOOLEAN *);
	int 	linkToSuper( OID, OID );
	int 	readType( OID, Catalog *, CatBOOLEAN *);
	int 	readType( char *, Catalog *, CatBOOLEAN *);
	int 	writeType( Catalog, OID *, CatBOOLEAN * );
	int 	deleteType( OID );
	int 	dropFromSuperClass( OID , OID , CatBOOLEAN *);
	int 	dropFromSubClass( OID , OID , CatBOOLEAN *);
	int 	deleteClass( OID , int *, CatBOOLEAN *) ;
	int 	deleteClass( char * , int *, CatBOOLEAN *) ;
	int 	updateType( OID , Catalog ) ;
	int 	deleteAttributes( Catalog , CatBOOLEAN *);
	int 	deleteFiles(); 
	int 	getAttrFile( FID * );
	int 	describe (OID, int = 0 );
	int 	describe ( char * );
	int 	listTypes( LinkedList *);
	int	updateRefCount( OID, int );
             
	int 	getClassInfo(char *, Catalog *, AttrList**, CatBOOLEAN*);
	int 	getClassInfo(char *, Catalog *, AttrList**, LinkedList *, LinkedList *, CatBOOLEAN*);
	int 	getAttrInfo( OID, AttrList**, CatBOOLEAN*);
	int 	getClassHierarchy(char *,Hierarchy **,CatBOOLEAN *);
	int 	changeNameOfClass(char *, char *);
	int 	addAttributeToClass(char *, AttributeType, CatBOOLEAN *);
	int 	addAttributeToClass(char *, memberNode *, CatBOOLEAN *);
	void 	printAllAtt(AttrList* );
	void 	dispLevel( LevelInfo *);
	void 	dispHier( Hierarchy *);
	int 	dispSuperClass( char *) ;
	int 	dispSubClass( char *) ;
	int 	getSubClass( char *, LinkedList * );
	int 	getSuperClass( char *, LinkedList *) ;
	int 	createIndex( char *, OID , CatBOOLEAN * );
	int 	dropIndex( char *, OID, CatBOOLEAN *);
	int 	myPrintAtt(AttrList* ); // Budakin rutininde var
             
	// Dynamic Function Linker routines...

	int  	addFunction( FunctionType, CatBOOLEAN *);
	int  	searchSign( char *, char *, char **, int, CatBOOLEAN *, 
		FunctionType *, int *, int); // 0 for found 1 for not found...
	int 	signCheck( char *, char *, FunctionType *, CatBOOLEAN *);
	int  	signCheck( char *, char *);
	int 	readFunc( OID , FunctionType *, CatBOOLEAN *);
	int 	searchFuncName( char *,char *, FunctionType *, CatBOOLEAN *); 
	int 	getFuncBySign(char *, char *, FunctionType *, CatBOOLEAN *);
	int	search_signature( char *, CatBOOLEAN * );
	int	deleteFunction(char *, char *,CatBOOLEAN *, FunctionType *);
	int	addFunc( FunctionType , CatBOOLEAN *);
	int	listFunction(  char *, LinkedList *);




	// Bjoin indices routines

	//int  	insertIndexName( char *, struct ForwardBackward * );
	//struct ForwardBackward*	getIndexRecord( char * );
};

#endif // __MOODS_CATALOG
