#ifndef DBIH
#define  DBIH

#include "dbi_error_text.h"
#include <math.h>
#include <setjmp.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <malloc.h>
//#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <sm_client.h>
#include "keywords.h"

/* Definitions */
#define MAX_ARITY         80
#define MAX_ATTR_NAME_LEN 100
#define MAX_CONDITIONS    10
#define MAX_STRING_LEN    100

typedef enum {JOIN, SELECT, GET_SET, MAT, UNNEST, UNION, FIX} OPERATOR ;
typedef enum {HASH_JOIN, MERGE_JOIN, FILTER, FILE_SCAN, BTREE_SELECT,
	      TRAVERSE, PTR_HH_JOIN, PTR_SM_JOIN, UNNEST_ALG, NESTED_LOOP_JOIN,
              UNION_ALG, FIX_ALG, SORT} ALGORITHM ;

/*--------------------------------- Cost -------------------------------*/

/* Define the structure of cost. */
typedef struct COST_ITEM
{
    double io;
} COST_ITEM;
typedef struct COST
{
    COST_ITEM cost;
} COST;

/* Initialize a cost variable pointed to by cp. */
#define init_cost(cp) { (cp)->cost.io = 0.0; }

/* Test to see if cost has been set. */
#define cost_not_set(cp) ((cp)->cost.io == 0.0)

#define print_cost(cp) { if( (cp) == NULL) {FAIL(ERR_NULL_POINTER);}\
                          else  {printi("%f \n",(cp)->cost.io);} }

#define copy_cost(c1,c2) (*(c1) = *(c2))

#if defined(__cplusplus) || defined(ANSI)
//void add_cost(COST*,COST*,COST*);
void compare_cost(COST*,COST*,COMPARE*);
void subtract_cost(COST*,COST*,COST*);
#else
void add_cost();
void compare_cost();
void subtract_cost();
#endif


/*--------------------------- Physical Property ------------------------*/

typedef struct PROPERTY_VECTOR
{
    int  sort_arity ; /* number of valid entries in array */
    char sort_attr[5][MAX_ATTR_NAME_LEN]; 
} PROPERTY_VECTOR;

#define property_vec_not_set(pv) ((pv)->sort_arity == -1) 

#define property_vec_dont_care(pv) ((pv)->sort_arity == -2) 

#define copy_property_vec(pv1,pv2) (*(pv1) = *(pv2))

#define init_property_vec(pv) ((pv)->sort_arity = 0)


/*---------------------------- Operator Arguments -----------------------*/

/* Attribute information */
#define REAL   3222
#define TEXT   3223
#define METHOD 3224

#define EXTENT 3225
#ifndef TYPE
#define TYPE   3226
#endif

typedef OID MOODS_SET;
typedef FID MOODS_FID;
typedef OID MOODS_LIST;
typedef OID MOODS_REF;
typedef union
{
    float      real_value;
    int        int_value;
    MOODS_LIST list;
    MOODS_REF  ref;
    MOODS_SET  set;
    char       text_value[MAX_STRING_LEN];
} CONSTANT;

typedef struct
{
    char FcnName[MAX_ATTR_NAME_LEN];
    struct JOIN_ARG *ParExpr[15];
    int ret_type;
    int argnum;
    char ret_name[MAXSTRLEN];
} FCN_INPUT;

class ATTRIBUTE
{
public:
    char     name[MAX_ATTR_NAME_LEN];
    char     class_name[20];
    short    attr_type;  // INTEGER METHOD .. can be a short
    CONSTANT domain_min;
    CONSTANT domain_max;
    int      unique;     // number of distinct values
    int      attr_width;
    int      offset;
    IID      index_id;
    float    fan;
    float    totref;
    float    notnull;      //Pinar
    char     btree_avail;  // TRUE if a btree is available
    char     set_id;  // set_id and set_no must be unique for each  
    char     set_no;  // occurance of a class that occurs in the query tree 
    char     use;
};


/* Pinar */
class TR_ST
{
public:
     char name[40];
     float tr;   /* totref */
     float ntr;  /* new totref */
};
/**/


class RELATION_SCHEMA
{
public:
    char      name[MAX_STRING_LEN];
    char      set_id; 
    char      type;               // EXTENT veya degil
    MOODS_FID fid;
    int       arity;              // number of attrs.
    ATTRIBUTE *attr[MAX_ARITY];
    int       cardinality;
    int       record_width;
    int       TypeId;
    void      copy( RELATION_SCHEMA * copyFrom);
    ATTRIBUTE *contains( char *attName);
    void      printOn();
};

typedef RELATION_SCHEMA BULK_SCHEMA;

/* General types */
#define EQUAL_OP  		EQ
#define LESS_THAN_OP		LT
#define GREATER_THAN_OP		GT
#define LESS_THAN_EQUAL_OP	LE
#define GREATER_THAN_EQUAL_OP	GE
#define NOT_EQUAL_OP		NE
#define NO_OP			3221

typedef int  RELATION;

typedef enum { AND_OP, OR_OP } BOOL_RELATION;

typedef enum { OPER, ATTRIB, CONST } ELEM_TYPE;

typedef struct
{
    int kind;
    CONSTANT value;
} CNST;
 
typedef union
{
    char      attr[MAX_ATTR_NAME_LEN];
    RELATION  op;
    CNST      cnst;
} ELEMENT;  
 
typedef struct JOIN_ARG
{
   ELEM_TYPE        type;
   ELEMENT          data;
   struct JOIN_ARG *left;
   struct JOIN_ARG *right;
   FCN_INPUT       *fcn;
} JOIN_ARG;

typedef struct
{
   int              num;
   char             attr[5][MAX_ATTR_NAME_LEN];
   BULK_SCHEMA      input[5];
   FCN_INPUT	    *fcn[5];
   char	            ref_or_set[5]; // 0 ref, 1 set
} MAT_ARG;

typedef struct
{
   char             attr[MAX_ATTR_NAME_LEN];
   BULK_SCHEMA      input;
   FCN_INPUT       *fcn;
} PTR_HH_JOIN_ARG;

typedef struct
{
    int             terms;                  /* terms in cond array */
    JOIN_ARG       *cond[MAX_CONDITIONS];   /* op[i] occurs between */
    BOOL_RELATION   op[MAX_CONDITIONS - 1]; /* cond[i] and cond[i+1] */
} SELECT_ARG;

typedef struct
{
    int             use_self;
    char            attr[MAX_ATTR_NAME_LEN];
    BULK_SCHEMA     input;
} GET_SET_ARG;

typedef struct 
{
   char             attr[MAX_ATTR_NAME_LEN];
   BULK_SCHEMA     input;
} UNNEST_ARG;

typedef struct
{
   char             attr[MAX_ATTR_NAME_LEN];
   BULK_SCHEMA      input;
} FIX_ARG;


typedef int UNION_ARG;



/* An OPERATOR_ARGUMENT is one of the operator arguments */
typedef union
{
    JOIN_ARG        join;
    SELECT_ARG      select;
    GET_SET_ARG     get_set;
    MAT_ARG         mat;
    UNNEST_ARG      unnest;    
    UNION_ARG       un;    
    FIX_ARG         fix;    
} OPERATOR_ARGUMENT;

/* Just initialize an operator argument to all binary 0s. */
#define init_operator_arg(oa) bzero((char *) (oa), sizeof(OPERATOR_ARGUMENT))

/* copy from arg2 to arg1 */
#define copy_operator_arg(arg1,arg2) (*(arg1) = *(arg2))

 
/*--------------------------- Algorithm Argument -----------------------*/

/* Both the MERGE_JOIN_ARG and HASH_JOIN_ARG are identical to the JOIN_ARG */
typedef JOIN_ARG    MERGE_JOIN_ARG;
typedef JOIN_ARG    NESTED_LOOP_JOIN_ARG;
typedef JOIN_ARG    HASH_JOIN_ARG;
typedef SELECT_ARG  FILTER_ARG;
typedef GET_SET_ARG FILE_SCAN_ARG;
typedef MAT_ARG     TRAVERSE_ARG;
typedef PTR_HH_JOIN_ARG     PTR_SM_JOIN_ARG;
typedef UNNEST_ARG  UNNEST_ALG_ARG;
typedef UNION_ARG   UNION_ALG_ARG;
typedef FIX_ARG     FIX_ALG_ARG;

/* A SORT_ARG contains the resulting sort property vector from the sort */
typedef struct
{
    int  	    sort_arity ; 
    char 	    sort_attr[5][MAX_ATTR_NAME_LEN]; 
} SORT_ARG;

typedef struct
{
    JOIN_ARG       *cond;
    BULK_SCHEMA     input;
    char            attr[MAX_ATTR_NAME_LEN];
    IID             index_id;
} BTREE_SELECT_ARG;

/* An ALGORITHM_ARGUMENT is one of the algorithm arguments */
typedef union
{
    MERGE_JOIN_ARG  	 merge_join;
    HASH_JOIN_ARG   	 hash_join;
    FILTER_ARG      	 filter;
    FILE_SCAN_ARG   	 file_scan;
    BTREE_SELECT_ARG  	 btree_select;
    TRAVERSE_ARG     	 traverse;
    SORT_ARG        	 sort;
    PTR_HH_JOIN_ARG      ptr_hh_join;
    PTR_SM_JOIN_ARG      ptr_sm_join;
    NESTED_LOOP_JOIN_ARG nested_loop_join;
    UNNEST_ALG_ARG  	 unnest_alg;
    UNION_ALG_ARG        un_alg;
    FIX_ALG_ARG          fix_alg;
} ALGORITHM_ARGUMENT;

#define init_algorithm_arg(aa) bzero((char *) (aa), sizeof(ALGORITHM_ARGUMENT)) 

#define copy_algorithm_arg(arg1,arg2) (*(arg1) = *(arg2))


/*--------------------------- Logical Property -------------------------*/

/* A LOGICAL_PROPERTY is simply a schema */
typedef BULK_SCHEMA LOGICAL_PROPERTY;

/* Initialize a LOGICAL_PROPERTY variable pointed to by lp. */
/* For macros with loops, use mi as the index variable.     */
#define init_logical_prop(lp) \
    {(lp)->name[0] = '\0'; (lp)->set_id = -1;}

#define copy_logical_prop(lp1, lp2) (*(lp1) = *(lp2))

/* If the object type is unknown, assume not set. */
#define logical_prop_not_set(lp) ((lp)->set_id == -1)

#endif

