#ifndef __OPT_H
#define __OPT_H
/* 
* Volcano Optimizer Generator 
* =========================== 
*   
*
*/ 

typedef int STATUS ;

typedef enum
	 {PV_EQUAL, PV1_COV_PV2, PV2_COV_PV1,
	  PV_NOT_COMPARABLE, PV_COMPARE_NOT_SET} PV_COMPARE ;

typedef enum
	 {PV_EXCL_TRUE, PV_EXCL_FALSE, PV_EXCL_NOT_SET}
	                                     PV_EXCL_COMPARE ;

typedef enum 
	{INITIALIZED_ONLY, IN_PROGRESS, COMPLETE}
			PROGRESS_FLAG ;

typedef enum
	{NO_TRANSFORMATIONS, TRANSFORMATIONS_IN_PROGRESS,
	 TRANSFORMATIONS_COMPLETE} TRANSFORMATION_FLAG ;

typedef enum {SUBEXPRESSION, TRANSFORMATION} LOG_EXPR_TYPE ;

typedef enum {OPTIMIZATION_FAILED, OPTIMIZATION_SUCCESSFUL}
	OPT_RETURN_VALUE ;

typedef enum {ALGORITHM_IS_CURRENT, ENFORCER_IS_CURRENT}
	 ALGORITHM_OR_ENFORCER ;

typedef enum {PERFORM_TRANSFORMATIONS, PERFORM_COST_ANALYSIS}
	 STRATEGY ;

typedef enum {FOUND_PLAN_COST_EXCEEDS_LIMIT,
	      LIMITING_COST_BOUND_EXISTS,
	      COST_LIMIT_EXCEEDED,
	      OPTIMIZE_FAILED,
	      OPTIMIZATION_NOT_POSSIBLE,
	      USED_COVERING_PLAN,
	      REASON_NOT_SET, 
              EXPR_IN_PROGRESS,}
	FAILURE_REASON ;

#define print_failure_reason(reason) \
	switch (reason) \
	{ \
	    case (FOUND_PLAN_COST_EXCEEDS_LIMIT) : \
		printf ("FOUND_PLAN_COST_EXCEEDS_LIMIT\n") ; \
		break ; \
	    case (LIMITING_COST_BOUND_EXISTS) : \
		printf ("LIMITING_COST_BOUND_EXISTS\n") ; \
		break ; \
	    case (COST_LIMIT_EXCEEDED) : \
		printf ("COST_LIMIT_EXCEEDED\n") ; \
		break ; \
	    case (OPTIMIZE_FAILED) : \
		printf ("OPTIMIZE_FAILED\n") ; \
		break ; \
	    case (USED_COVERING_PLAN) : \
		printf ("USED_COVERING_PLAN\n") ; \
		break ; \
	    case (OPTIMIZATION_NOT_POSSIBLE) : \
		printf ("OPTIMIZATION_NOT_POSSIBLE\n") ; \
		break ; \
	    case (REASON_NOT_SET) : \
		printf ("REASON_NOT_SET\n") ; \
		break ; \
	    case (EXPR_IN_PROGRESS) : \
		printf ("EXPR_IN_PROGRESS\n") ; \
		break ; \
	    default: \
		FAIL (ERR_UNKNOWN_VALUE) ;\
		break ; \
	}

#define print_optimization_error(error) \
	switch (error) \
	{ \
	    case (OPTIMIZATION_FAILED) : \
		printf ("OPTIMIZATION_FAILED\n") ; \
		break ; \
	    case (OPTIMIZATION_SUCCESSFUL) : \
		printf ("OPTIMIZATION_SUCCESSFUL\n") ; \
		break ; \
	    default: \
		FAIL (ERR_UNKNOWN_VALUE) ;\
		break ; \
	}

#define print_progress_flag(progress_flag) \
	switch (progress_flag) \
	{ \
	    case (INITIALIZED_ONLY) : \
		printi ("INITIALIZED ONLY\n") ; \
		break ; \
	    case (IN_PROGRESS) : \
		printi ("IN PROGRESS\n") ; \
		break ; \
	    case (COMPLETE) : \
		printi ("COMPLETE\n") ; \
		break ; \
	    default : \
	        printi ("error: progress flag = %d\n",\
			 progress_flag) ; \
		FAIL (ERR_UNKNOWN_VALUE) ; \
		break ; \
	} 

#define print_transformation_flag(tf) \
	switch (tf) \
	{ \
	    case (NO_TRANSFORMATIONS) : \
		printf ("NO TRANSFORMATIONS\n") ; \
		break ; \
	    case (TRANSFORMATIONS_IN_PROGRESS) : \
		printf ("TRANSFORMATIONS IN PROGRESS\n") ; \
		break ; \
	    case (TRANSFORMATIONS_COMPLETE) : \
		printf ("TRANSFORMATIONS COMPLETE\n") ; \
		break ; \
	    default : \
	        printi ("error: transformation flag = %d\n",\
			 tf) ; \
		FAIL (ERR_UNKNOWN_VALUE) ; \
		break ; \
	} \

#define init_opt_return(p) \
	(p) -> pele = NULL ;\
	(p) -> reason = REASON_NOT_SET ;\
	(p) -> error = OPTIMIZATION_FAILED ;	

typedef struct SYSTEM_PROPERTY
    {
	long			cardinality ;
	long			record_width ;
    } SYSTEM_PROPERTY ;

/* ----------------------------------------------------------------- */

typedef struct OPT_RETURN 
    {
	struct PHY_EXPR_LIST_ELEM	*pele ;
	OPT_RETURN_VALUE		error ;
        FAILURE_REASON			reason ;
    } OPT_RETURN ;

typedef struct ALGO_APPL_RETURN
    {
	BOOLEAN				apply_truth_value ;
	ALGORITHM_ARGUMENT		algorithm_argument ;
	int				process_input_count ;
	int				enforcer_inputs ;
	BOOLEAN				reestimate_cost ;
    } ALGO_APPL_RETURN ;

typedef struct LOG_EXPR
    {
	OPERATOR			my_operator ;
	OPERATOR_ARGUMENT		argument ;
	struct LOG_EXPR_LIST_ELEM	*input [MAX_OPERATOR_ARITY] ;
    } LOG_EXPR ;

typedef struct QUERY
    {
	OPERATOR			my_operator ;
	OPERATOR_ARGUMENT		argument ;
	struct QUERY			*input [MAX_OPERATOR_ARITY] ;
    } QUERY ;

#ifdef KEEP_TRANS_INFO
typedef struct TRANS_RULE_ADDR
    {
	short				rule_id ;
	struct LOG_EXPR_LIST_ELEM	*trans_ptr ;
    } TRANS_RULE_ADDR ;

#define INITIAL_TRANS_LIST_SIZE		TRANS_RULE_COUNT

typedef struct TRANS_LIST 
    {
	struct TRANS_RULE_ADDR		trans_list [INITIAL_TRANS_LIST_SIZE] ;
    } TRANS_LIST ;

typedef struct TRANS_INFO
    {
	int				count ;
	int				list_size ;
	struct TRANS_LIST		*trans_list_ptr ;
    } TRANS_INFO ;
#endif

typedef struct LOG_EXPR_LIST_ELEM
    {
	LOG_EXPR		        expr ;
	int				eq_class_index ;
	int				last_rule ;
	int				rule_generated_from ;
	int				depth ;
#ifdef KEEP_TRANS_INFO
	struct TRANS_INFO		trans_info ;
#endif
#ifdef GET_STATS
        int 				ref_count ;
#endif
	struct LOG_EXPR_LIST_ELEM	*next, *prev ;
	struct LOG_EXPR_LIST_ELEM	*hash_next, *hash_prev ;
    } LOG_EXPR_LIST_ELEM ;

typedef struct LOG_EXPR_LIST
    {
	LOG_EXPR_LIST_ELEM	*first, *last ;
#ifdef GET_STATS
	int			count ;
#endif
    } LOG_EXPR_LIST ;

/* ----------------------------------------------------------------- */

typedef struct PHY_EXPR
    {
	ALGORITHM			algorithm ;
	ALGORITHM_ARGUMENT		algorithm_argument ;
	COST				cost ;
	long				ref_count ; 
	int 				num_inputs ;
	struct PLAN			*plan ;
	struct PHY_EXPR_LIST_ELEM	*input [MAX_ALGORITHM_ARITY] ;
	struct PHY_EXPR_LIST_ELEM	*derived_pele ;
    } PHY_EXPR ;

typedef struct PLAN
    {
	ALGORITHM			algorithm ;
	ALGORITHM_ARGUMENT		algorithm_argument ;
	OPERATOR			my_operator ;
	OPERATOR_ARGUMENT		operator_argument ;
	COST				cost ;
	LOGICAL_PROPERTY		logical_property ;
	PROPERTY_VECTOR			property_vector ;
	PROPERTY_VECTOR			excl_property_vector ;
	SYSTEM_PROPERTY			system_property ;
	struct PLAN			*input [MAX_ALGORITHM_ARITY] ;
    } PLAN ;

typedef struct PHY_EXPR_LIST_ELEM
    {
	PHY_EXPR			*expr ;
	int				eq_class_index ;
	PROPERTY_VECTOR			property_vector ;
	PROPERTY_VECTOR			excl_property_vector ;
	struct PHY_EXPR_LIST_ELEM	*next, *prev ;
	struct LOG_EXPR_LIST_ELEM	*lele ;
#ifdef USE_PE_HASH
        struct PHY_EXPR_LIST_ELEM       *hash_next ;
#endif
        COST				fail_cost ;
	FAILURE_REASON			fail_reason ;
	BOOLEAN				failure_exists ;
	PROGRESS_FLAG			progress_flag ;
    } PHY_EXPR_LIST_ELEM ;

typedef struct PE_HASH_ENTRY
    {
	struct PHY_EXPR_LIST_ELEM	*first ;
#ifdef GET_STATS
	int				count ;
#endif
    } PE_HASH_ENTRY ;


typedef struct PHY_EXPR_LIST
    {
	PHY_EXPR_LIST_ELEM		*first, *last ;
#ifdef USE_PE_HASH
	PE_HASH_ENTRY       		hash_table [PE_HT_SIZE] ;
#endif
#ifdef GET_STATS
	long				count ;
#endif
    } PHY_EXPR_LIST ;

typedef struct EQ_CLASS
    {
	SYSTEM_PROPERTY		system_property ;
	LOGICAL_PROPERTY	logical_property ;
	LOG_EXPR_LIST		log_expr_list ;
        PHY_EXPR_LIST		phy_expr_list ;
	TRANSFORMATION_FLAG	trans_flag ;
        int			eq_class_index ;
#ifdef GET_STATS
	double			count ;
#endif
    } EQ_CLASS ;

#define init_sys_prop(sp) (sp) ->cardinality = -1 ;\
		          (sp) -> record_width = -1 ;

#define sys_prop_not_set(sp) (((sp) -> cardinality == -1) \
		         	 || ((sp) -> record_width == -1))

#define print_sys_props(sys_prop) \
	if ((sys_prop) == NULL) \
	     FAIL (ERR_NULL_POINTER) ; \
	printi ("cardinality: %ld\n", (sys_prop) -> cardinality) ; \
	printi ("record_width: %ld\n", (sys_prop) -> record_width) ; 

#define copy_sys_prop(sp1,sp2) \
	(sp1) -> cardinality = (sp2) -> cardinality ; \
	(sp1) -> record_width = (sp2) -> record_width ; 

#define allocate_opt_return(p) \
	(p) = (OPT_RETURN *) \
		malloc ((unsigned) sizeof (OPT_RETURN)) ;\
	if ((p) == NULL) {FAIL (ERR_MEM_ALLOC_FAILED) ;}\
	int_opt_return(p) ;

#define init_query(q) init_operator_arg\
			     (&((q) -> argument));\
			{int i_yiy ;\
			 for (i_yiy = MAX_OPERATOR_ARITY;\
			      --i_yiy >= 0;)\
			     (q) -> input[i_yiy] = NULL;} 

#define allocate_query(q) \
	(q) = (QUERY *) \
		malloc ((unsigned) sizeof (QUERY)) ;\
	if ((q) == NULL) {FAIL (ERR_MEM_ALLOC_FAILED) ;}\
        init_query(q) ;

#define init_plan(p) init_algorithm_arg\
			 (&((p) -> algorithm_argument));\
			init_operator_arg\
			 (&((p) -> operator_argument));\
			init_property_vec (&((p) -> property_vector));\
			init_property_vec (&((p) -> excl_property_vector));\
			init_logical_prop \
				(&((p) -> logical_property)) ;\
			init_sys_prop (&((p) -> system_property)) ;\
                        init_cost(&((p) -> cost)) ;\
			{int i_xix ;\
			 for (i_xix = MAX_ALGORITHM_ARITY;\
			      --i_xix >= 0;)\
			     (p) -> input[i_xix] = NULL;} 

#define allocate_plan(p) \
	(p) = (PLAN *) \
		malloc ((unsigned) sizeof (PLAN)) ;\
	if ((p) == NULL) {FAIL (ERR_MEM_ALLOC_FAILED) ;}\
        init_plan(p) ;

#define equal_operators(op1,op2)	(((op1) == (op2)))

#define NO_MORE	-99

#if defined(__cplusplus) || defined (ANSI)

#ifndef copy_operator_arg
STATUS copy_operator_arg (OPERATOR_ARGUMENT *, OPERATOR_ARGUMENT *) ;
#endif

#ifndef copy_algorithm_arg
STATUS copy_algorithm_arg (ALGORITHM_ARGUMENT *, ALGORITHM_ARGUMENT *) ;
#endif

#ifndef copy_logical_prop
STATUS copy_logical_prop (LOGICAL_PROPERTY *, LOGICAL_PROPERTY *) ;
#endif

#ifndef compare_property_vec
STATUS compare_property_vec (PROPERTY_VECTOR *, PROPERTY_VECTOR *,
    SYSTEM_PROPERTY *, LOGICAL_PROPERTY *, PV_COMPARE *) ;
#endif

#ifndef compare_excl_property_vec
STATUS compare_excl_property_vec (PROPERTY_VECTOR *, 
    PROPERTY_VECTOR *,  SYSTEM_PROPERTY *,  LOGICAL_PROPERTY *, 
    PV_EXCL_COMPARE *) ;
#endif

#ifndef compare_operator_args
STATUS compare_operator_args (OPERATOR_ARGUMENT *, OPERATOR *,
    OPERATOR_ARGUMENT *, OPERATOR *, COMPARE *) ;
#endif

#ifndef print_logical_prop
STATUS print_logical_prop (LOGICAL_PROPERTY *) ;
#endif

#ifndef print_property_vec
STATUS print_property_vec (PROPERTY_VECTOR *) ;
#endif

#ifndef print_operator_arg
STATUS print_operator_arg (OPERATOR_ARGUMENT *, OPERATOR *) ;
#endif

#ifndef print_algorithm_arg
STATUS print_algorithm_arg (ALGORITHM_ARGUMENT *, ALGORITHM *) ;
#endif

#ifndef copy_cost
STATUS copy_cost (COST *, COST *) ;
#endif

#else

#ifndef copy_operator_arg
STATUS copy_operator_arg () ;
#endif

#ifndef copy_algorithm_arg
STATUS copy_algorithm_arg () ;
#endif

#ifndef copy_logical_prop
STATUS copy_logical_prop () ;
#endif

#ifndef compare_property_vec
STATUS compare_property_vec () ;
#endif

#ifndef compare_excl_property_vec
STATUS compare_excl_property_vec () ;
#endif

#ifndef compare_excl_property_vec
STATUS compare_operator_args () ;
#endif

#ifndef compare_operator_args
STATUS compare_operator_args () ;
#endif

#ifndef print_logical_prop
STATUS print_logical_prop () ;
#endif

#ifndef print_property_vec
STATUS print_property_vec () ;
#endif

#ifndef print_operator_arg
STATUS print_operator_arg () ;
#endif

#ifndef print_algorithm_arg
STATUS print_algorithm_arg () ;
#endif

#ifndef copy_cost
STATUS copy_cost () ;
#endif

#endif
#endif
