/* 
* Volcano Optimizer Generator 
* =========================== 
*   
* $Header: /tmp_mnt/home/chainsaw/bill/optgen/optgen/RCS/gen.c,v 1.16 92/11/18 01:32:57 bill Exp Locker: bill $ 
*
*/ 

#if defined(__cplusplus) || defined ANSI
#include <stdlib.h>
#endif
 
#include "opt_defs.h" 
#include "parse.h"

extern FILE 		*yyin ;
SYM_TAB			symbol_table [MAX_SYMBOLS] ;
int			max_algorithm_arity = -1, 
			max_operator_arity = -1,
			sym_tab_index = -1 ,
			var_tab_index = -1, 
			impl_tab_index = -1 ,
			trans_rule_num = -1,
			impl_rule_num = -1,
			node_id = -1 ,
			oper_count = 0,
			algo_count = 0,
			end_parens ,
			max_level = -1 ; 
RULE_GROUP_TAB		rule_group_table[MAX_GROUPS];
CURRENT_RULE_TYPE	current_rule_type ;
RULE_EXPR 		*lhs_queue[MAX_EXPR_PER_RULE] ;
RULE_EXPR 		*rhs_queue[MAX_EXPR_PER_RULE] ;
int 			lhs_tail, rhs_tail ;
BOOLEAN			in_loop = FALSE ;
extern USER_CODE	current_user_code ;
RULE_PRIORITY_INFO	trans_rule_priority[MAX_RULES],
			impl_rule_priority[MAX_RULES] ;

void 			print_expr_tree () ;
static int		rule_group_count = 0 ;

#if defined(__cplusplus) 
extern "C" int yyparse();
#endif

#if defined(ANSI)
extern int yyparse();
#endif

#if defined(__cplusplus) || defined(ANSI)
void write_trans_rule_code (RULE_EXPR *,
                                RULE_DIRECTION) ;
void write_trans_rule_lhs_code (FILE *, RULE_EXPR *, int,
                                RULE_DIRECTION) ;
void write_trans_rule_rhs_code (FILE *) ;
void write_impl_rule_code (RULE_EXPR *,
                                RULE_DIRECTION) ;
void write_impl_rule_rhs_code (FILE *, RULE_EXPR *) ;
void clear_rule_group_table (RULE_GROUP_TAB *) ;
void print_rule_group_table (RULE_GROUP_TAB *) ;
void add_rule_group_table (RULE_GROUP_TAB *, char *, int *) ;
void find_rule_group_table (RULE_GROUP_TAB *, char *, int *, BOOLEAN *) ;
void print_expr_tree(FILE *, RULE_EXPR *, int) ;
void number_nodes (RULE_EXPR *, int, CURRENT_RULE_SIDE) ;
char *my_strstr( char*, char* ) ;
#else
void                    write_trans_rule_code () ;
void                    write_trans_rule_lhs_code () ;
void                    write_trans_rule_rhs_code () ;
void                    write_impl_rule_code () ;
void                    write_impl_rule_lhs_code () ;
void                    write_impl_rule_rhs_code () ;
void 			clear_rule_group_table () ;
void 			print_rule_group_table () ;
void 			add_rule_group_table () ;
void 			find_rule_group_table () ;
void 			print_expr_tree() ;
void 			number_nodes () ;
char                    *my_strstr () ;
#endif

main ()
{
    void 	write_dbi_gen () ;
    int 	i ;
    FILE 	*fp ;
#if defined(__cplusplus) || defined(ANSI)
    void write_trans_rule_def (FILE *) ;
#else
    void write_trans_rule_def ( ) ;
#endif

    for (i = -1; ++i<MAX_SYMBOLS;)
    {
	strcpy (symbol_table[i].name, "") ;
	symbol_table[i].arity = -99 ;
	symbol_table[i].defined_in = DEFINED_NOT_SET ;
	symbol_table[i].var_set = VAR_NOT_SET ;
	symbol_table[i].func = FUNC_NOT_SET ;
	symbol_table[i].id = -99 ;
	symbol_table[i].bind_id = -99 ;
	symbol_table[i].rhs_count = 0 ;
    }

    clear_rule_group_table (&(rule_group_table[0])) ;

    yyin = fopen ("model.input", "r") ;
    if (yyin == NULL)
    {
	fprintf (stderr, "Error: open failed for model.input file.\n") ;
	exit (-1) ;
    }

    yyparse () ;

    if (trans_rule_num == -1)
    {
        fp = fopen ("temp_match_trans_rule.c", "w") ;
        write_trans_rule_def (fp) ;
	fclose (fp) ;
    }

    fp = fopen ("temp_match_trans_rule.c", "a+") ;
    fprintf (fp, "default:lele_in->last_rule = NO_MORE;return (OKAY);\n");
    fprintf (fp, "}\n}\n") ;
    fprintf (fp, "}\n") ;
    fclose (fp) ;

    fp = fopen ("temp_match_impl_rule.c", "a+") ;

    if (impl_rule_num == -1)
    {
	fprintf (stderr, "Error: No implementation rules defined.\n") ;
    }

    fprintf (fp, 
     "default:*algorithm_or_enforcer = (ALGORITHM) NO_MORE; return (OKAY);\n");
    fprintf (fp, "}\n}\n") ;
    fprintf (fp, "}\n") ;
    fclose (fp) ;
    write_dbi_gen () ;
    return (0) ;
    
}

#if defined(__cplusplus) || defined(ANSI)
void clear_rule_group_table (RULE_GROUP_TAB *t)
#else
void clear_rule_group_table (t)
RULE_GROUP_TAB	*t ;
#endif
{
    int i ;
    
    for (i = -1 ; ++i < MAX_GROUPS;)
    {
	t[i].group_id = -1 ; 
	strcpy (t[i].group_name, "") ;  
    }
}

#if defined(__cplusplus) || defined(ANSI)
void print_rule_group_table (RULE_GROUP_TAB *t)
#else
void print_rule_group_table (t)
RULE_GROUP_TAB  *t ;
#endif
{
    int i ;
    
    printf ("\nRule group table:\n") ;
    for (i = -1 ; ++i < rule_group_count;)
    {
	printf ("entry %d:\n") ;
	printf ("  id    : %d\n", t[i].group_id) ; 
	printf ("  name  : %s\n", t[i].group_name) ; 
    }
}

#if defined(__cplusplus) || defined(ANSI)
void add_rule_group_table (RULE_GROUP_TAB *t, char *name, int *id)
#else
void add_rule_group_table (t, name, id)
RULE_GROUP_TAB	*t ;
char		*name ;
int		*id ;
#endif
{
    if (rule_group_count == MAX_GROUPS)
    {
	fprintf (stderr,
		     "Increase MAX_GROUPS in parse.h and re-make\n") ;
	exit (-1) ;
    }

    t[rule_group_count].group_id = rule_group_count ;
    strcpy (t[rule_group_count].group_name, name) ; 
    *id = rule_group_count ;
    ++rule_group_count ;
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void find_rule_group_table (RULE_GROUP_TAB *t, char *name, int *id, BOOLEAN *found)
#else
void find_rule_group_table (t, name, id, found)
RULE_GROUP_TAB	*t ;
char		*name ;
int		*id ;
BOOLEAN		*found ;
#endif
{
    int i ;

    *found = FALSE ;
    if (*id == -1)
	for (i = -1 ; ++i < rule_group_count;)
	{
	    if (strcmp (name, t[i].group_name) == 0)
	    {
		*id = i ;
		*found = TRUE ;
		break ;	
	    }
	}
    else
	for (i = -1 ; ++i < rule_group_count;)
	{
	    if (*id == t[i].group_id)
	    {
		strcpy (name, t[i].group_name) ;
		*found = TRUE ;
		break ;	
	    }
	}

    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void process_user_code (char* in_str)
#else
void process_user_code (in_str)
char 		*in_str ;
#endif
{
    char 	*s, *var, variable[80], *last, out_str[200],
	 	out_str2[200] ;
    char 	id[80] ;
    int 	var_len, i ;
    BOOLEAN	found ;
    FILE 	*fp ;

    s = (char *) malloc ((unsigned) ((strlen (in_str)) + 1)) ;
    strcpy (s, in_str) ;

    var = index (s, '?') ;
    last = s ;
    bzero (out_str, 200) ;

    while (var)
    {
        if (strncmp (var + 1, " ", 1) != 0 
		&& strncmp (var + 1, "\n", 1) != 0
		&& strncmp (var + 1, "\t" , 1) != 0)
	{
	    bzero (variable,80) ;
	    
	    if (strncmp (var + 1, "op_arg", 6) == 0)
	    {
		found = FALSE ;
		strncat (out_str, last, var - last) ;
		var_len = strspn (var, "?op_arg0123456789") ;
		strncpy (variable, var, var_len) ;
		for (i = sym_tab_index; ++i <= var_tab_index;)
		    if (symbol_table[i].func == IS_OP_ARG_VARIABLE)
		    {
			if (strcmp (variable, symbol_table[i].name) == 0) 
			{
			    found = TRUE ;
			    if (symbol_table[i].defined_in ==
				   DEFINED_IN_RHS) 
			    {
				strcat (out_str, "(&(temp_op_argument[") ;
				sprintf (id, "%d", symbol_table[i].bind_id) ;	
				strcat (out_str, id) ;
				strcat (out_str, "]))") ;
				if (current_user_code == USER_APPL_CODE) 
				{
				    symbol_table[i].var_set 
				       = VAR_SET_IN_USER_CODE ;
				}
				else if (current_user_code == USER_COND_CODE)
				{
				    fprintf (stderr, 
					"Error: oper. arg. variables (%s) defined\n", variable) ;
				    fprintf (stderr, 
					"       in right hand side of rules cannot be \n") ;
				    fprintf (stderr, 	
					"       referenced in user condition code.\n") ;
				    if (current_rule_type == TRANS_RULE)
					fprintf (stderr, "       trans. rule number %d\n",
						    trans_rule_num) ;
				    else
					fprintf (stderr, "       impl. rule number %d\n",
						    impl_rule_num) ;
				}
			    }
			    else
			    {
				if (symbol_table[i].defined_in ==
				       DEFINED_IN_LHS) 
				{
				    strcat (out_str, "op_argument[") ;
				    sprintf (id, "%d", symbol_table[i].bind_id) ;	
				    strcat (out_str, id) ;
				    strcat (out_str, "]") ;
				}
				else
				{
				    fprintf (stderr, 
					"Error: File %s, line %d - bad symbol table.\n",
					__FILE__, __LINE__) ;
				    exit (-1) ;
				}
			    }
			}
		    }
		if (found == FALSE)
		{
		    fprintf (stderr, 
			"Error: Unknown oper. argument var. %s in user code.\n",
				variable) ;
		    if (current_rule_type == TRANS_RULE)
			fprintf (stderr, "       trans. rule number %d\n",
				    trans_rule_num) ;
		    else
			fprintf (stderr, "       impl. rule number %d\n",
				    impl_rule_num) ;
		    exit (-1) ;
		}
            }
	    else if (strncmp (var + 1, "al_arg", 6) == 0)
	    {
		found = FALSE ;
		strncat (out_str, last, var - last) ;
		var_len = strspn (var, "?al_arg0123456789") ;
		strncpy (variable, var, var_len) ;
		found = FALSE ;
                for (i = sym_tab_index; ++i <= var_tab_index;)
                    if (symbol_table[i].func == IS_AL_ARG_VARIABLE
                             &&
                        strcmp (variable, symbol_table[i].name) == 0)
                    {
                        strcat (out_str, "aa_param") ;
                        found = TRUE ;
                    }
                if (found == FALSE)
                {
                    fprintf (stderr,
                        "Error: Unknown alg. argument var %s in user code.\n",
                                variable) ;
                    if (current_rule_type == TRANS_RULE)
                        fprintf (stderr, "      trans. rule number %d\n",
                                    trans_rule_num) ;
                    else
                        fprintf (stderr, "      impl. rule number %d\n",
                                    impl_rule_num) ;
                    exit (-1) ;
                }

            }
	    else if (strncmp (var + 1, "op", 2) == 0)
	    {
		strncat (out_str, last, var - last) ;
		var_len = strspn (var, "?op0123456789") ;
		strncpy (variable, var, var_len) ;
		found = FALSE ;
		for (i = sym_tab_index; ++i <= var_tab_index;)
		    if (symbol_table[i].func == IS_OPER_VARIABLE
			     &&
			strcmp (variable, symbol_table[i].name) == 0)
		    {
			strcat (out_str, "operator_array[") ;
			sprintf (id, "%d", 
			    symbol_table[i].bind_id) ;	
			strcat (out_str, id) ;
			strcat (out_str, "]") ;
			found = TRUE ;
		    }
		if (found == FALSE)
		{
		    fprintf (stderr, 
			"Error: Unknown operator var. %s in user code.\n",
				variable) ;
		    if (current_rule_type == TRANS_RULE)
			fprintf (stderr, "      trans. rule number %d\n",
				    trans_rule_num) ;
		    else
			fprintf (stderr, "      impl. rule number %d\n",
				    impl_rule_num) ;
		    exit (-1) ;
		}
	    }
	    else 
	    {
		strncat (out_str, last, var - last) ;
		var_len = strspn (var, "?0123456789") ;
		strncpy (variable, var, var_len) ;
		found = FALSE ;
		for (i = sym_tab_index; ++i <= var_tab_index;)
		    if (symbol_table[i].func == IS_INPUT_VARIABLE
			     &&
			strcmp (variable, symbol_table[i].name) == 0)
		    {
			strcat (out_str, "input[") ;
			sprintf (id, "%d", 
			    symbol_table[i].bind_id) ;	
			strcat (out_str, id) ;
			strcat (out_str, "]") ;
			found = TRUE ;
		    }
		if (found == FALSE)
		{
		    fprintf (stderr, 
			"Error: Unknown input var. %s in user code.\n",
				variable) ;
		    if (current_rule_type == TRANS_RULE)
			fprintf (stderr, "      trans. rule number %d\n",
				    trans_rule_num) ;
		    else
			fprintf (stderr, "      impl. rule number %d\n",
				    impl_rule_num) ;
		    exit (-1) ;
		}
	    }
	    last = var + var_len ;
	}
	++var ;
	var = index (var, '?') ;
    }

    strcat (out_str, last) ;
    s = &out_str[0] ;
    var = my_strstr (s, "REJECT") ;
    last = s ;
    bzero (out_str2, 200) ;
    while (var)
    {
	strncat (out_str2, last, var - last) ;
	if (in_loop == TRUE)
	    strcat (out_str2, "continue") ;
	else
	{
	    if (current_rule_type == TRANS_RULE)
	    {
		char	irs[4] ;

		sprintf (irs, "%d", trans_rule_num) ;
		strcat (out_str2, "goto incr_rule_iteration_") ;
		strcat (out_str2, irs) ; 
		strcat (out_str2, ";") ; 
	    }
	    else
		strcat (out_str2, "goto try_rule ;") ;
	}
	last = var + 6 ;
	++var ;
	var = my_strstr (var, "REJECT") ;
    }

    strcat (out_str2, last) ;
    last = &out_str2[0] ;
    while (strncmp (last, " ", 1) == 0) ++last ;

    if (current_rule_type == TRANS_RULE)
    {
	fp = fopen ("temp_match_trans_rule.c", "a+") ;
	fprintf (fp, "%s\n",last) ;
	fclose (fp) ;
    }
    else 
    {
	fp = fopen ("temp_match_impl_rule.c", "a+") ;
	fprintf (fp, "%s\n",last) ;
	fclose (fp) ;
    }

    free ((char *) s) ;
    return ;

}

#if defined(__cplusplus) || defined(ANSI)
void write_impl_rule_code (RULE_EXPR *lhs,
                                RULE_DIRECTION rule_direction )
#else
void write_impl_rule_code (lhs, rule_direction)
RULE_EXPR               *lhs ;
RULE_DIRECTION          rule_direction ;
#endif
{
    FILE *fp ; 

    if (impl_rule_num == 0)
    {
	fp = fopen ("temp_match_impl_rule.c", "w") ;
	fprintf (fp,"#include \"opt_defs.h\"\n") ;
	fprintf (fp, "#include \"dbi_gen.h\"\n") ;
	fprintf (fp,"#include \"dbi.h\"\n") ;
	fprintf (fp,"#include \"opt.h\"\n") ;
	fprintf (fp,"#include \"expr_str.h\"\n") ;
	fprintf (fp,"#include \"global.h\"\n\n") ;
	fprintf (fp,"extern int indent_value ;\n") ;

	fprintf (fp,"extern int              opt_arity[] ;\n\n") ;
	fprintf (fp,"extern int              algo_arity[] ;\n\n") ;
	fprintf (fp,"#if DEBUG>=1\n") ;
	fprintf (fp,"extern TRACE            trace ;\n\n") ;
	fprintf (fp,"#endif\n") ;
        fprintf (fp, "#if defined(__cplusplus) || defined(ANSI)\n") ;
	fprintf (fp,"extern char * error_text (int) ;\n");
	fprintf (fp,"#else\n") ;
	fprintf (fp, "char *error_text () ;\n") ;
	fprintf (fp,"#endif\n") ;

        fprintf (fp, "#if defined(__cplusplus) || defined(ANSI)\n") ;
	fprintf (fp, "STATUS match_impl_rule(LOG_EXPR_LIST_ELEM *lele_in,\n");
        fprintf (fp, "LOG_EXPR *log_expr, int rule_group,\n") ;
        fprintf (fp, "STORAGE_SEG_ANCHOR *ssa,\n") ;
        fprintf (fp, "int count,\n") ;
        fprintf (fp, "LOG_EXPR_LIST_ELEM *inputs_to_search[],\n") ; 
        fprintf (fp, "ALGORITHM *algorithm_or_enforcer,\n") ;
        fprintf (fp, "ALGORITHM_ARGUMENT *aa_param,\n") ;
        fprintf (fp, "int *ret_count)\n") ;
        fprintf (fp,
         "{\nSTATUS generate_all_transformations(LOG_EXPR_LIST_ELEM *,int,");
        fprintf (fp, "STORAGE_SEG_ANCHOR *) ;\n") ;
        fprintf (fp, "#else\n") ;
	fprintf (fp,
	  "STATUS match_impl_rule(lele_in, log_expr, rule_group, ssa, count, inputs_to_search, algorithm_or_enforcer, aa_param, ret_count)\n") ;
	fprintf (fp,"LOG_EXPR_LIST_ELEM	*lele_in ;\n") ;
	fprintf (fp,"LOG_EXPR		*log_expr ;\n") ;
	fprintf (fp,"int		rule_group ;\n") ;
	fprintf (fp,"STORAGE_SEG_ANCHOR *ssa ;\n") ;
	fprintf (fp,"LOG_EXPR_LIST_ELEM	*inputs_to_search[] ;\n") ;
	fprintf (fp,"int		count, *ret_count ;\n") ;
	fprintf (fp, "ALGORITHM		*algorithm_or_enforcer; \n") ;
	fprintf (fp, "ALGORITHM_ARGUMENT *aa_param; \n\n") ;
        fprintf (fp,
         "{\nSTATUS generate_all_transformations() ;\n");
        fprintf (fp, "#endif\n") ;
	fprintf (fp,"LOG_EXPR_LIST_ELEM	*lele[%d] ;\n",
			 MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR		operator_array[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR_ARGUMENT	*op_argument[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR_ARGUMENT	temp_op_argument[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "LOG_EXPR_LIST_ELEM	*input[%d];\n", 
	  MAX_EXPR_PER_RULE*max_operator_arity) ;
	fprintf (fp,
	  "EQ_CLASS		*temp_eq_class ;\n") ;
	fprintf (fp,"BOOLEAN			found_rule ;\n") ;
	fprintf (fp,"int		i ;\n") ;
	fprintf (fp,"STATUS                     error ;\n\n") ;
       /*
	fprintf (fp,
	  "get_eq_class_addr(ssa,lele_in->eq_class_index,&eq_class) ;\n\n") ;
       */
	fprintf (fp,"#if SAFETY>=1\n") ;
	fprintf (fp,
	  "for (i = -1; ++i<opt_arity[(int) (log_expr->my_operator)];)\n{\n");
	fprintf (fp,
	  "get_eq_class_addr(ssa,log_expr->input[i]->eq_class_index,\n"); 
	fprintf (fp,"&temp_eq_class) ;\n") ;
	fprintf (fp,
	  "if (log_expr->input[i] != temp_eq_class->log_expr_list.first)\n") ;
	fprintf (fp,"{printi (\"Problem with input.\\n\") ;\n") ;
	fprintf (fp,"FAIL (ERR_OPTIMIZER_INTERNAL_ERROR) ;\n}}\n") ;
	fprintf (fp, "#endif\n\n") ;
	fprintf (fp, "--count ;\n") ;
	fprintf (fp, "for (;;)\n{\n\n") ;
	fprintf (fp, "try_rule:\n") ;
	fprintf (fp, "switch (++count)\n{\n") ;
    }
    else
	fp = fopen ("temp_match_impl_rule.c", "a+") ;

    fprintf (fp, "case (%d):\n",impl_rule_num) ;
    write_trans_rule_lhs_code (fp, lhs, -1, rule_direction) ;

    fclose (fp) ;
    
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void free_tree (RULE_EXPR *ptr)
#else
void free_tree (ptr)
RULE_EXPR       *ptr ;
#endif
{
    int i ;

    for (i = -1; ++i < ptr -> arity;)
	if (ptr -> input[i] != NULL)
	    free_tree (ptr -> input[i]) ;
    free ((char *) ptr) ;
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void write_trans_rule_def ( FILE *fp )
#else
void write_trans_rule_def ( fp )
FILE	*fp ;
#endif
{
	fprintf (fp,"#include \"opt_defs.h\"\n") ;
	fprintf (fp, "#include \"dbi_gen.h\"\n") ;
	fprintf (fp,"#include \"dbi.h\"\n") ;
	fprintf (fp,"#include \"opt.h\"\n") ;
	fprintf (fp,"#include \"expr_str.h\"\n") ;
	fprintf (fp,"#include \"global.h\"\n\n") ;

	fprintf (fp,"extern int indent_value ;\n") ;

	fprintf (fp,"#if defined(__cplusplus) || defined(ANSI)\n") ;
	fprintf (fp,"extern char * error_text (int) ;\n");
	fprintf (fp,"STATUS link_eq_classes(STORAGE_SEG_ANCHOR *,int,");
        fprintf (fp, "int) ;\n") ;
	fprintf (fp,
	 "STATUS generate_all_transformations(LOG_EXPR_LIST_ELEM *,int,");
	fprintf (fp, "STORAGE_SEG_ANCHOR *) ;\n") ;
	fprintf (fp, "STATUS find_log_expr_list_elem (STORAGE_SEG_ANCHOR *,");
        fprintf (fp, "LOG_EXPR *,") ;
        fprintf (fp, "int,") ;
        fprintf (fp, "LOG_EXPR_LIST_ELEM *,") ;
        fprintf (fp, "LOG_EXPR_LIST_ELEM **) ;\n") ;
	fprintf (fp, "STATUS insert_log_expr (STORAGE_SEG_ANCHOR *,") ;
	fprintf (fp, "LOG_EXPR *,") ;
	fprintf (fp, "int,") ;
	fprintf (fp, "int *,") ;
	fprintf (fp, "LOG_EXPR_LIST_ELEM **,") ;
	fprintf (fp, "BOOLEAN) ;\n") ;
	fprintf (fp, "#if defined(KEEP_TRANS_INFO)\n") ;
	fprintf (fp, "STATUS insert_trans_list(LOG_EXPR_LIST_ELEM *,") ;
	fprintf (fp, "int, LOG_EXPR_LIST_ELEM *) ; \n") ;
	fprintf (fp,"#endif\n") ;
	fprintf (fp, "#if DEBUG>=3 && defined(PRINT)\n") ;
	fprintf (fp, "STATUS print_log_expr_list_elem") ;
	fprintf (fp, " (LOG_EXPR_LIST_ELEM *, BOOLEAN) ;\n") ;
	fprintf (fp,"#endif\n") ;
	fprintf (fp,"#else\n") ;
	fprintf (fp,"STATUS link_eq_classes () ;\n") ;
	fprintf (fp, "char *error_text () ;\n") ;
	fprintf (fp,"STATUS generate_all_transformations () ;\n") ;
	fprintf (fp, "STATUS find_log_expr_list_elem () ;\n");
	fprintf (fp, "STATUS insert_log_expr () ;\n") ;
	fprintf (fp, "#if defined(KEEP_TRANS_INFO)\n") ;
	fprintf (fp, "STATUS insert_trans_list() ;\n") ;
	fprintf (fp,"#endif\n") ;
	fprintf (fp, "#if DEBUG>=3 && defined(PRINT)\n") ;
	fprintf (fp, "STATUS print_log_expr_list_elem () ;\n") ;
	fprintf (fp,"#endif\n") ;
	fprintf (fp,"#endif\n") ;

	fprintf (fp,"extern int              opt_arity[] ;\n\n") ;
	fprintf (fp,"extern TRACE            trace ;\n") ;
	fprintf (fp, "#ifdef GET_STATS\n") ;
	fprintf (fp,"extern int		     num_reallocs;\n") ;
	fprintf (fp, "#endif\n") ;

	fprintf (fp,"#if defined(__cplusplus) || defined(ANSI)\n") ;
	fprintf (fp,
	 "STATUS match_trans_rule(LOG_EXPR_LIST_ELEM *lele_in,\n") ;
	fprintf (fp, "LOG_EXPR *log_expr,int rule_group,\n") ;
	fprintf (fp, "STORAGE_SEG_ANCHOR *ssa)\n");
	fprintf (fp,"#else\n") ;
	fprintf (fp,
	 "STATUS match_trans_rule(lele_in,log_expr,rule_group,ssa)\n");
	fprintf (fp,"LOG_EXPR_LIST_ELEM	*lele_in ;\n") ;
	fprintf (fp,"LOG_EXPR		*log_expr ;\n") ;
	fprintf (fp,"int		rule_group ;\n") ;
	fprintf (fp,"STORAGE_SEG_ANCHOR	*ssa ;\n") ;
	fprintf (fp,"#endif\n") ;
	fprintf (fp,
	  "{\nEQ_CLASS		*eq_class, *temp_eq_class ;\n") ;
	fprintf (fp,"LOG_EXPR		temp_log_expr ;\n") ;
	fprintf (fp,"int			ret_eq_class_index, i ;\n") ;
	fprintf (fp,"/* BOOLEAN			insert_expr ; */ \n") ;
	fprintf (fp,"LOG_EXPR_LIST_ELEM	*lele[%d], *ret_lele[%d] ;\n",
			 MAX_EXPR_PER_RULE, MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR		operator_array[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR_ARGUMENT	*op_argument[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "OPERATOR_ARGUMENT	temp_op_argument[%d];\n", MAX_EXPR_PER_RULE);
	fprintf (fp,
	  "LOG_EXPR_LIST_ELEM	*input[%d];\n", 
	  MAX_EXPR_PER_RULE*max_operator_arity) ;
	fprintf (fp,"BOOLEAN			found_rule ;\n") ;
	if (current_rule_type == TRANS_RULE)
	{
	    fprintf (fp,"int			RULE_ITERATION ;\n") ;
	    fprintf (fp,"int			FIRE_RULE_COUNT ;\n") ;
	}
	fprintf (fp,"STATUS			error ;\n\n") ;
	fprintf (fp,
	  "get_eq_class_addr(ssa,lele_in->eq_class_index,&eq_class) ;\n\n") ;
       /*
	fprintf (fp,
	  "for (i = %d;--i >= 0;)\n{\n", MAX_EXPR_PER_RULE);
	fprintf (fp, "init_operator_arg (&temp_op_argument[i]);}\n") ;
       */
	fprintf (fp,"#if SAFETY>=1\n") ;
	fprintf (fp,
	  "for (i = opt_arity[(int) (log_expr->my_operator)];--i >= 0;)\n{\n");
	fprintf (fp, "if (!(log_expr->input[i])) continue;\n") ;
	fprintf (fp,
	  "get_eq_class_addr(ssa,log_expr->input[i]->eq_class_index,\n"); 
	fprintf (fp,"&temp_eq_class) ;\n") ;
	fprintf (fp,
	  "if (log_expr->input[i] != temp_eq_class->log_expr_list.first)\n") ;
	fprintf (fp,"{printi (\"Problem with input.\\n\") ;\n") ;
	fprintf (fp,"FAIL (ERR_OPTIMIZER_INTERNAL_ERROR) ;\n}}\n") ;
	fprintf (fp, "#endif\n\n") ;
	fprintf (fp, "#if DEBUG>=3\n") ;
        fprintf (fp, "if (trace ==TRACE_ON) {\n") ;
	fprintf (fp, 
	  "printi (\"in match_trans_rule for list element:\\n\") ;\n") ;
	fprintf (fp,
	  "CALL(print_log_expr_list_elem(lele_in,TRUE)) ;\n") ;
        fprintf (fp, "}\n") ;
	fprintf (fp, "#endif\n\n") ;
	fprintf (fp, "for (;;)\n{\n\n") ;
	fprintf (fp, "try_rule:\n") ;
	fprintf (fp, "switch (++ (lele_in -> last_rule))\n{\n") ;

	return ;
    
}

#if defined(__cplusplus) || defined(ANSI)
void write_trans_rule_code (RULE_EXPR *lhs, 
                            RULE_DIRECTION rule_direction)
#else
void write_trans_rule_code (lhs, rule_direction)
RULE_EXPR               *lhs ;
RULE_DIRECTION          rule_direction ;
#endif
{

#if defined(__cplusplus) || defined(ANSI)
    void write_trans_rule_def (FILE *) ;
#else
    void write_trans_rule_def ( ) ;
#endif

    FILE *fp ; 

    if (trans_rule_num == 0)
    {
	fp = fopen ("temp_match_trans_rule.c", "w") ;
	write_trans_rule_def (fp) ;
    }
    else
	fp = fopen ("temp_match_trans_rule.c", "a+") ;
    fprintf (fp, "case (%d):\n",trans_rule_num) ;
    write_trans_rule_lhs_code (fp, lhs, -1, rule_direction) ;
    fclose (fp) ;
    
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void write_impl_rule_rhs_code (FILE *fp, RULE_EXPR *rhs)
#else
void write_impl_rule_rhs_code (fp, rhs)
FILE                    *fp ;
RULE_EXPR               *rhs ;
#endif
{
    int i, k, found ;

    fprintf (fp, "*algorithm_or_enforcer = %s ;\n",
		rhs -> name) ;
    fprintf (fp, "for (i = MAX_ALGORITHM_ARITY; --i >= 0;)\n") ;
    fprintf (fp, "inputs_to_search[i] = NULL ;\n") ;

    for (i = -1; ++i < rhs -> arity;)
    {
	if (rhs -> input[i] -> func == IS_INPUT_VARIABLE)
	{
	    for (k = sym_tab_index, found = 0; ++k <= var_tab_index;)
	    {
		if (symbol_table[k].func == IS_INPUT_VARIABLE &&
		    symbol_table[k].id == rhs -> input[i] -> id)
		{
		    found = 1 ;
		    fprintf (fp, 
		      "inputs_to_search[%d] = input[%d] ;\n",
				i,
				symbol_table[k].bind_id) ; 
		    break ;
		}
	    }
	    if (found == 0)
	    {
		fprintf (stderr, 
			"Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		exit (-1) ;
	    }
	}
	else if (rhs -> input[i] -> func == IS_NULL)
	{
	    fprintf (fp, 
	      "inputs_to_search[%d] = NULL ;\n",
			i) ;
	}
	else
	{
	    fprintf (stderr, "Error: File %s, line %d - unknown value\n",
			    __FILE__, __LINE__) ;
	    exit (-1) ;
	}
    }
    fprintf (fp, "found_rule = TRUE ;\n") ;
    fprintf (fp, "*ret_count = %d ;\n", impl_rule_num) ;
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void write_trans_rule_rhs_code (FILE *fp)
#else
void write_trans_rule_rhs_code (fp)
FILE                    *fp ;
#endif
{
    RULE_EXPR	*ptr ;
    int i, j, k, found ;

    for (j = rhs_tail + 1; --j >= 0;) 
    {
	ptr = rhs_queue[j] ;
	if (ptr -> func == IS_OPERATOR)
	{
	    fprintf (fp, "init_log_expr(&temp_log_expr);\n") ;
	    fprintf (fp, "temp_log_expr.my_operator = %s ;\n",
			ptr -> name) ; 
	    fprintf (fp, "/* insert_expr = TRUE ; */ \n") ;
	}
	if (ptr -> func == IS_OPER_VARIABLE)
	{
	    fprintf (fp, "init_log_expr(&temp_log_expr);\n") ;
	    for (i = sym_tab_index, found = 0; ++i <= var_tab_index;)
	    {
		if (symbol_table[i].func == ptr -> func &&
		    symbol_table[i].id == ptr -> id)
		{
		    found = 1 ;
		    if (symbol_table[i].bind_id < 0)
		    {
			fprintf (stderr, 
			 "Error: File %s, line %d - bad symbol table.\n",
					__FILE__, __LINE__) ;
			exit (-1) ;
		    }
		    fprintf (fp, "temp_log_expr.my_operator = operator_array[%d] ;\n",
				symbol_table[i].bind_id) ; 
		    break ;
		}
	    }
	    if (found == 0)
	    {
		fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		exit (-1) ;
	    }
	}

       /* added May 14,1992 at TI */
	if (ptr -> func == IS_INPUT_VARIABLE && ptr -> node_id == 0)
	{
	    for (k = sym_tab_index, found = 0; ++k <= var_tab_index;)
	    {
		if (symbol_table[k].func == IS_INPUT_VARIABLE &&
		    symbol_table[k].id == ptr -> id)
		{
		    found = 1 ;
		    fprintf (fp,
		      "temp_lele = input[%d] ;\n",
				symbol_table[k].bind_id) ;
		    break ;
		}
	    }
	    if (found == 0)
	    {
		fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		exit (-1) ;
	    }
	    fprintf (fp, "if (temp_lele -> eq_class_index != lele[0] -> eq_class_index)\n") ;
	    fprintf (fp, 
	      "{\nget_eq_class_addr(ssa,temp_lele -> eq_class_index,&temp_eq_class);\n",
	       ptr -> node_id) ;
	    fprintf (fp, "if (temp_eq_class -> trans_flag == ") ;
	    fprintf (fp, "TRANSFORMATIONS_IN_PROGRESS)\n") ;
	    if (in_loop == TRUE)
		fprintf (fp, "continue; \n") ;
	    else
	    {
		fprintf (fp, "goto incr_rule_iteration_%d; \n", trans_rule_num) ;
	    }
	    fprintf (fp, 
		"if (temp_eq_class -> trans_flag == TRANSFORMATIONS_COMPLETE ||\n") ;
	    fprintf (fp, 
		"    temp_eq_class -> trans_flag == NO_TRANSFORMATIONS)\n") ;
	    fprintf (fp, 
		"{ \nCALL (link_eq_classes (ssa, lele[0] -> eq_class_index, \n") ;
	    fprintf (fp, "temp_eq_class -> eq_class_index)) ;}\n}\n") ;
	}
	
	if (ptr -> func == IS_OPER_VARIABLE || ptr -> func == IS_OPERATOR)
	{
	    for (i = sym_tab_index, found = 0; ++i <= var_tab_index;)
	    {
		if (symbol_table[i].func == IS_OP_ARG_VARIABLE &&
		    symbol_table[i].id == ptr -> arg_id)
		{
		    found = 1 ;
		    if (symbol_table[i].bind_id < 0)
		    {
			fprintf (stderr, 
			 "Error: File %s, line %d - bad symbol table.\n",
					__FILE__, __LINE__) ;
			exit (-1) ;
		    }
		    fprintf (fp, 
		    "copy_operator_arg(&(temp_log_expr.argument),&(temp_op_argument[%d]));\n",
					    symbol_table[i].bind_id) ;
		    break ;
		}
	    }
	    if (found == 0)
	    {
		fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		exit (-1) ;
	    }
	    for (i = -1; ++i < ptr -> arity;)
	    {
		if (ptr -> input[i] -> func == IS_INPUT_VARIABLE)
		{
		    for (k = sym_tab_index, found = 0; ++k <= var_tab_index;)
		    {
			if (symbol_table[k].func == IS_INPUT_VARIABLE &&
			    symbol_table[k].id == ptr -> input[i] -> id)
			{
			    found = 1 ;
			    fprintf (fp, 
			      "temp_log_expr.input[%d] = input[%d] ;\n",
					i,
					symbol_table[k].bind_id) ; 
			    break ;
			}
		    }
		    if (found == 0)
		    {
			fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
					__FILE__, __LINE__) ;
			exit (-1) ;
		    }
		}
		else if (ptr -> input[i] -> func == IS_NULL)
		{
		    fprintf (fp, 
		      "temp_log_expr.input[%d] = NULL ;\n",
				i) ;
		}
		else if (ptr -> input[i] -> func == IS_OPERATOR || 
			 ptr -> input[i] -> func == IS_OPER_VARIABLE)
		{
		    fprintf (fp, 
		      "temp_log_expr.input[%d] = ret_lele[%d] ;\n",
				i, ptr -> input[i] -> node_id) ;
		}
		else
		{
		    fprintf (stderr, "Error: File %s, line %d - unknown value\n",
				    __FILE__, __LINE__) ;
		    exit (-1) ;
		}
	    }
	    fprintf (fp, "CALL (find_log_expr_list_elem (ssa, &temp_log_expr, (int) -1,\n") ;
	    fprintf (fp, 
		      "(LOG_EXPR_LIST_ELEM *) NULL, &ret_lele[%d])) ;\n",
			    ptr -> node_id) ;
	    fprintf (fp, "if (ret_lele[%d] == NULL)\n{\n", ptr -> node_id) ;
	    if (ptr -> node_id == 0)
		fprintf (fp,
			  "get_eq_class_addr(ssa,lele_in->eq_class_index,&eq_class) ;\n\n") ;
	    fprintf (fp, "CALL (insert_log_expr (ssa, &temp_log_expr,\n") ;
	    if (ptr -> node_id == 0)
		fprintf (fp, "eq_class->eq_class_index, &ret_eq_class_index, \n") ;
	    else 
		fprintf (fp, "(int) -1, &ret_eq_class_index, \n") ;
	    fprintf (fp, "&ret_lele[%d], FALSE)) ;\n", ptr -> node_id) ;
	    fprintf (fp, "ret_lele[%d] -> rule_generated_from = %d ;\n",
			ptr -> node_id, trans_rule_num) ;
	    if (ptr -> node_id != 0)
	    {
		fprintf (fp, 
		  "CALL (generate_all_transformations (ret_lele[%d],\n",
				    ptr -> node_id) ;
		fprintf (fp, "rule_group, ssa)) ;\n") ;
	    }
	    fprintf (fp, "} \n") ; 
	    fprintf (fp, 
	      "else {\nget_eq_class_addr(ssa,ret_lele[%d] -> eq_class_index,&temp_eq_class);\n",
	       ptr -> node_id) ;
	    fprintf (fp, "ret_lele[%d] = temp_eq_class -> log_expr_list.first;\n",
			    ptr -> node_id) ;
	    if (ptr -> node_id != 0)
	    {
	       /*
		fprintf (fp, 
		    "if (ret_lele[%d] -> eq_class_index == lele[0] -> eq_class_index)\n",
			ptr -> node_id) ;
		if (in_loop == TRUE)
		    fprintf (fp, "continue; \n") ;
		else
		    fprintf (fp, "goto incr_rule_iteration_%d; \n", trans_rule_num) ;
	       */
	    }
	    fprintf (fp, 
		"if (ret_lele[%d] -> eq_class_index != lele[0] -> eq_class_index)\n{\n",
		    ptr -> node_id) ;
	    fprintf (fp, "if (temp_eq_class -> trans_flag == ") ;
	    fprintf (fp, "TRANSFORMATIONS_IN_PROGRESS)\n") ;
	    if (in_loop == TRUE)
		fprintf (fp, "continue; \n") ;
	    else
	    {
		fprintf (fp, "goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		fprintf (fp, "/* FAIL (ERR_OPTIMIZER_INTERNAL_ERROR) ; */ \n") ;
	    }
	    if (ptr -> node_id == 0)
	    {
		fprintf (fp, 
		    "if (temp_eq_class -> trans_flag == TRANSFORMATIONS_COMPLETE ||\n") ;
		fprintf (fp, 
		    "    temp_eq_class -> trans_flag == NO_TRANSFORMATIONS)\n") ;
		fprintf (fp, 
		    "{ \nCALL (link_eq_classes (ssa, lele[0] -> eq_class_index, \n") ;
		fprintf (fp, "temp_eq_class -> eq_class_index)) ;}\n") ;
	    }
	    else 
	    {
		fprintf (fp, 
		    "if (temp_eq_class -> trans_flag == NO_TRANSFORMATIONS)\n") ;
		fprintf (fp, "{\n CALL (generate_all_transformations (ret_lele[%d],\n",
				    ptr -> node_id) ;
		fprintf (fp, "rule_group, ssa)) ;\n}\n") ;
	    }
	    fprintf (fp, "} \n } \n") ;
	    if (ptr -> node_id == 0)
	    {
                fprintf (fp, "#ifdef KEEP_TRANS_INFO\n") ;
		fprintf (fp, "CALL (insert_trans_list(lele_in,%d,ret_lele[0]));\n",
				trans_rule_num) ;
                fprintf (fp, "#endif\n") ;
		fprintf (fp, "found_rule = TRUE ;\n") ;
	    }
	}
    }
    return ;
}

#if defined(__cplusplus) || defined(ANSI)
void write_trans_rule_lhs_code (FILE *fp, RULE_EXPR *lhs, int input_no,
                                RULE_DIRECTION rule_direction)
#else
void write_trans_rule_lhs_code (fp, lhs, input_no, rule_direction)
FILE                    *fp ;
RULE_EXPR               *lhs ;
int                     input_no ;
RULE_DIRECTION          rule_direction ;
#endif
{
    int i ;
    int found = 0 ;
    BOOLEAN	found_group ;
    char 	temp_name [MAX_NAME] ;

    if (lhs -> func == IS_NULL)
	return ;

    if (lhs -> parent == NULL)
    {
	in_loop = FALSE ;
	end_parens = 0 ;
	fprintf (fp, "found_rule = FALSE ;\n", lhs -> node_id) ;
	if (current_rule_type == TRANS_RULE)
	{
	    fprintf (fp, "FIRE_RULE_COUNT = 1 ;\n") ;
	    fprintf (fp, "RULE_ITERATION = 1 ;\n") ;
	    fprintf (fp, "fire_rule_%d :\n", trans_rule_num) ;
	}
	fprintf (fp, "lele[%d] = lele_in;\n", lhs -> node_id) ;
	if (rule_direction == L_TO_R1)
	    fprintf (fp, 
	    "if (lele_in -> rule_generated_from == %d) goto try_rule ;\n",
		 trans_rule_num) ;

	found_group = FALSE ;
        find_rule_group_table (&(rule_group_table[0]), &(temp_name[0]),
				&(lhs -> group_id), &found_group) ;

	if (found_group == FALSE)
	{
	    fprintf (stderr, 
		"Error: File %s, line %d - bad symbol table.\n",
		__FILE__, __LINE__) ;
	    exit (-1) ;
	}

	fprintf (fp, "if (rule_group != %s)\n goto try_rule;\n", 
	    temp_name) ;
    }
    else
    {
	if (lhs -> loop_over == TRUE)
	{
	    fprintf (fp, "get_eq_class_addr\n") ;
	    fprintf (fp, "(ssa,lele[%d] -> expr.input[%d]->eq_class_index,\n",
	                  lhs -> parent -> node_id, input_no) ;
	    fprintf (fp, " &temp_eq_class) ;\n");
	    fprintf (fp, "if (lele[%d] -> expr.input[%d] !=\n", 
	                  lhs -> parent -> node_id, input_no) ;
	    fprintf (fp, " temp_eq_class->log_expr_list.first)\n");
	    fprintf (fp, "FAIL(ERR_OPTIMIZER_INTERNAL_ERROR);\n") ;
	    fprintf (fp, "if (temp_eq_class -> trans_flag == ") ;
	    fprintf (fp, "TRANSFORMATIONS_IN_PROGRESS)\n") ;
	    if (in_loop == TRUE)
		fprintf (fp, "continue; \n") ;
	    else
	    {
		if (current_rule_type == TRANS_RULE)
		    fprintf (fp, 
			"goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		else 
		    fprintf (fp, "goto try_rule; \n") ;
	    }
#ifdef EXHAUSTIVE
	    fprintf (fp, "/* FAIL(ERR_OPTIMIZER_INTERNAL_ERROR); */ \n") ;
#else
           /*
	    fprintf (stderr, "Error: File %s, Line %d - not implemented",
				__FILE__, __LINE__) ;
	    exit (-1) ;
           */
	   /* should generate all transformations here */
#endif
	    fprintf (fp, "if (temp_eq_class -> trans_flag == ") ;
	    fprintf (fp, "NO_TRANSFORMATIONS) \n{\n") ;

	    fprintf (fp, 
		"CALL (generate_all_transformations (temp_eq_class -> log_expr_list.first,\n") ;
		fprintf (fp, "rule_group, ssa)) ;}\n") ;
/*
	    fprintf (fp, 
	      "FAIL (ERR_OPTIMIZER_INTERNAL_ERROR) ;\n}\n") ;
*/
	    fprintf (fp, 
	      "for (lele[%d]=temp_eq_class->log_expr_list.first;\n",
		         lhs -> node_id) ;
	    in_loop = TRUE ;
	    ++end_parens ;
	    fprintf (fp, "lele[%d] != NULL;\n", lhs -> node_id) ;
	    fprintf (fp, "lele[%d] = lele[%d] -> next)\n{", 
			       lhs -> node_id, lhs -> node_id) ;
	}
	else
	{
	    fprintf (fp, "lele[%d] = lele[%d] -> expr.input[%d];\n", 
		lhs -> node_id, lhs -> parent -> node_id, input_no) ;
	    fprintf (fp, "get_eq_class_addr\n") ;
	    fprintf (fp, "(ssa,lele[%d] ->eq_class_index, &temp_eq_class) ;\n",
				lhs -> node_id) ;
	    fprintf (fp, "if (temp_eq_class -> trans_flag == ") ;
	    fprintf (fp, "TRANSFORMATIONS_IN_PROGRESS)\n") ;
	    if (in_loop == TRUE)
		fprintf (fp, "continue; \n") ;
	    else
		if (current_rule_type == TRANS_RULE)
		    fprintf (fp, 
			"goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		else 
		    fprintf (fp, "goto try_rule; \n") ;
	}
    }

    if (lhs -> func != IS_INPUT_VARIABLE)
    {
	fprintf (fp, "operator_array[%d] = lele[%d] -> expr.my_operator;\n", 
		          lhs -> node_id, lhs -> node_id ) ;

 /* changed Feb 18 */
        if (lhs -> func == IS_OPER_VARIABLE)
	{
	    fprintf (fp, "if (opt_arity[(int) (operator_array[%d])] != %d)\n",
			lhs -> node_id, lhs -> arity) ;
	    if (in_loop == FALSE) 
	    {
		if (current_rule_type == TRANS_RULE)
		    fprintf (fp, 
			"goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		else 
		    fprintf (fp, "goto try_rule; \n") ;
	    }
	    else
		fprintf (fp, 
		 "continue;\n") ;
	}

	if (lhs -> func == IS_OPERATOR) 
	{
	    fprintf (fp, "if (operator_array[%d] != %s)\n", 
		      lhs -> node_id, lhs -> name) ;
	    if (in_loop == FALSE) 
	    {
		if (current_rule_type == TRANS_RULE)
		    fprintf (fp, 
			"goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		else 
		    fprintf (fp, "goto try_rule; \n") ;
	    }
	    else
		fprintf (fp, 
		 "continue;\n") ;
	}

	fprintf (fp, "op_argument[%d] = &(lele[%d] -> expr.argument);\n",
		          lhs -> node_id, lhs -> node_id ) ;

	for (i = -1; ++i < lhs -> arity;)
	{
	    if (lhs -> input[i] -> func == IS_NULL)
	    {
		fprintf (fp, 
		 "if (lele[%d]->expr.input[%d] != NULL)\n",
			lhs -> node_id, i) ;
		if (in_loop == FALSE) 
		{
		    if (current_rule_type == TRANS_RULE)
			fprintf (fp, 
			    "goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		    else 
			fprintf (fp, "goto try_rule; \n") ;
		}
		else
		    fprintf (fp, 
		     "continue;\n") ;
	    }
	    else
	    {
		fprintf (fp, 
		     "if (lele[%d]->expr.input[%d] == NULL)\n",
			lhs -> node_id, i) ;
		if (in_loop == FALSE) 
		{
		    if (current_rule_type == TRANS_RULE)
			fprintf (fp, 
			    "goto incr_rule_iteration_%d; \n", trans_rule_num) ;
		    else 
			fprintf (fp, "goto try_rule; \n") ;
		}
		else
		    fprintf (fp, 
		     "continue;\n") ;
	    }
	}
    }	 

    if (lhs -> func == IS_OPER_VARIABLE) 
    {
	for (i = sym_tab_index; ++i <= var_tab_index;)
	    if (symbol_table[i].func == IS_OPER_VARIABLE &&
		symbol_table[i].id == lhs -> id)
	    {
		found = 1 ;
		if (symbol_table[i].bind_id == -1)
		{
		    symbol_table[i].bind_id = lhs -> node_id ; 
		}
		else	
		{
		    fprintf (fp, 
			"if (operator_array[%d] != operator_array[%d])\n",
			 symbol_table[i].bind_id, lhs -> node_id) ;
		    if (in_loop == FALSE) 
		    {
			if (current_rule_type == TRANS_RULE)
			    fprintf (fp, 
				"goto incr_rule_iteration_%d; \n", 
				trans_rule_num) ;
			else 
			    fprintf (fp, "goto try_rule; \n") ;
		    }
		    else
			fprintf (fp, 
			 "continue;\n") ;
		}
		break ;
	    }
	if (found == 0)
	    fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
			__FILE__, __LINE__) ;
    }

    if (lhs -> func != IS_INPUT_VARIABLE)
    {
	for (i = sym_tab_index; ++i <= var_tab_index;)
	    if (symbol_table[i].func == IS_OP_ARG_VARIABLE &&
		symbol_table[i].id == lhs -> arg_id)
	    {
		found = 1 ;
		if (symbol_table[i].bind_id == -1)
		{
		    symbol_table[i].bind_id = lhs -> node_id ; 
		}
		else	
		{
		    fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		}
		break ;
	    }
	if (found == 0)
	    fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
			__FILE__, __LINE__) ;
    }

    if (lhs -> func == IS_INPUT_VARIABLE)
    {
	if (lhs -> parent == NULL)
	    fprintf (stderr, "Error: File %s, line %d - parent NULL\n",
			__FILE__, __LINE__) ;
	fprintf (fp, "input[%d] = lele[%d] -> expr.input[%d];\n",
			lhs -> node_id, lhs -> parent -> node_id,
			input_no) ;
	fprintf (fp,"#if SAFETY>=1\n") ;
	fprintf (fp,
	  "get_eq_class_addr(ssa,input[%d]->eq_class_index,\n", lhs -> node_id) ;
	fprintf (fp,"&temp_eq_class) ;\n") ;
	fprintf (fp,
	  "if (input[%d] != temp_eq_class->log_expr_list.first)\n",
                lhs -> node_id) ;
	fprintf (fp,"FAIL (ERR_OPTIMIZER_INTERNAL_ERROR) ;\n") ;
	fprintf (fp, "#endif\n\n") ;
	for (i = sym_tab_index; ++i <= var_tab_index;)
	    if (symbol_table[i].func == IS_INPUT_VARIABLE &&
		symbol_table[i].id == lhs -> id)
	    {
		found = 1 ;
		if (symbol_table[i].bind_id == -1)
		{
		    symbol_table[i].bind_id = lhs -> node_id ; 
		}
		else	
		{
		    fprintf (stderr, "Error: File %s, line %d - bad symbol table.\n",
				__FILE__, __LINE__) ;
		}
		break ;
	    }
    }

    for (i = -1; ++i < lhs -> arity;)
	if (lhs -> input[i] != NULL)
	    write_trans_rule_lhs_code (fp, lhs -> input[i], i, rule_direction) ;

    return ;

}

#if DEBUG>=1
#if defined(__cplusplus) || defined(ANSI)
void print_expr_tree(FILE *fp, RULE_EXPR *ptr, int in)
#else
void print_expr_tree(fp, ptr, in)
RULE_EXPR 	*ptr ;
int	      	in ;
FILE		*fp ;
#endif
{
    int 	i ;
    char 	temp_name[MAX_NAME] ;
    BOOLEAN	found_group ;

    findent(fp,in) ;fprintf (fp,"Name: %s\n", ptr -> name) ;
    findent(fp,in);fprintf(fp,"Id:   %d\n", ptr -> id) ;
    findent(fp,in);fprintf(fp,"Address:       %#x\n", ptr) ;
    findent(fp,in);fprintf(fp,"Node id:       %d\n", ptr -> node_id) ;
    findent(fp,in);fprintf(fp,"Level:         %d\n", ptr -> level) ;
    
    find_rule_group_table (&(rule_group_table[0]), &(temp_name[0]),
			    &(ptr -> group_id), &found_group) ;

    if (found_group == TRUE)
    {
	findent(fp,in);fprintf(fp,"Rule group id: %d\n", ptr -> group_id) ;
	findent(fp,in);fprintf(fp,"Rule group :   %s\n", temp_name) ;
    }
    else
    {
	findent(fp,in);fprintf(fp,"Rule group not found.\n") ;
    }

    if (ptr -> level == 0)
    {
	findent(fp,in);fprintf(fp,"max level = %d\n", max_level) ;
    }
    if (ptr -> parent)
    {
	findent(fp,in);fprintf(fp,"Parent addr:   %#x\n", ptr -> parent) ;
    }
    else
    {
	findent(fp,in);fprintf(fp,"Parent addr:   NULL\n") ;
    }

    findent(fp,in);fprintf(fp,"Func: ") ;
    switch (ptr -> func)
    {
	case (IS_OPERATOR):
	    fprintf (fp, "OPERATOR\n") ;
	    break ;
	case (IS_ALGORITHM):
	    fprintf (fp, "ALGORITHM\n") ;
	    break ;
	case (IS_ENFORCER):
	    fprintf (fp, "ENFORCER\n") ;
	    break ;
	case (IS_OPER_VARIABLE):
	    fprintf (fp, "OPERATOR VARIABLE\n") ;
	    break ;
	case (IS_OP_ARG_VARIABLE):
	    fprintf (fp, "OPERATOR ARGUMENT VARIABLE\n") ;
	    break ;
	case (IS_AL_ARG_VARIABLE):
	    fprintf (fp, "ALGORITHM ARGUMENT VARIABLE\n") ;
	    break ;
	case (IS_INPUT_VARIABLE):
	    fprintf (fp, "INPUT VARIABLE\n") ;
	    break ;
	case (IS_NULL):
	    fprintf (fp, "NULL\n") ;
	    break ;
	default:
	    fprintf (fp, "error - UNKNOWN\n") ;
	    break ;
    }	

    findent(fp,in);fprintf(fp,"Expression?:   ") ;
    switch (ptr -> expr)
    {
	case (TRUE):
	    fprintf (fp, "TRUE\n") ;
	    break ;
	case (FALSE):
	    fprintf (fp, "FALSE\n") ;
	    break ;
	default:
	    fprintf (fp, "error - UNKNOWN\n") ;
	    break ;
    }	

    findent(fp,in);fprintf(fp,"Loop over?:    ") ;
    switch (ptr -> loop_over)
    {
	case (TRUE):
	    fprintf (fp, "TRUE\n") ;
	    break ;
	case (FALSE):
	    fprintf (fp, "FALSE\n") ;
	    break ;
	default:
	    fprintf (fp, "error - UNKNOWN\n") ;
	    break ;
    }	

    findent(fp,in);fprintf(fp,"Arity:         %d\n", ptr -> arity) ;

    findent(fp,in);fprintf(fp,"Argument name: %s\n", ptr -> arg_name) ;
    findent(fp,in);fprintf(fp,"Argument id:   %d\n", ptr -> arg_id) ;

    for (i = -1; ++i < ptr -> arity;)
    {
        findent(fp,in);fprintf(fp,"Input %d:\n", i) ;
	if (ptr -> input[i] == NULL)
	{
	    fprintf (stderr, "Error: File %s, Line %d - NULL pointer\n",
			__FILE__, __LINE__) ;
	}
	else
	    print_expr_tree (fp, ptr -> input[i], in + 2) ;
    }

    return ;
}
#endif

#if defined(__cplusplus) || defined(ANSI)
void number_nodes (RULE_EXPR *ptr, int start, CURRENT_RULE_SIDE side)
#else
void number_nodes (ptr, start, side)
RULE_EXPR       *ptr ;
int                     start ;
CURRENT_RULE_SIDE       side ;
#endif
{
    int i ;

    if (start == 1)
    {
	(side == LHS_CURRENT) ? 
		(lhs_tail = -1) : (rhs_tail = -1) ;
        for (i = MAX_EXPR_PER_RULE; --i >= 0; )
	{
	    (side == LHS_CURRENT) ? (lhs_queue[i] = NULL) : (rhs_queue[i] = NULL) ;
	}
	(side == LHS_CURRENT) ? 
		(lhs_queue[++lhs_tail] = ptr) : (rhs_queue[++rhs_tail] = ptr) ;
    }
  
    for (i = -1; ++i < ptr -> arity;)
	if (ptr -> input[i] -> func != IS_NULL)
	{
	    if (ptr -> input[i] -> func == IS_OPERATOR
	        || ptr -> input[i] -> func == IS_ALGORITHM
	    	|| ptr -> input[i] -> func == IS_ENFORCER
		|| ptr -> input[i] -> func == IS_OPER_VARIABLE)
	    {
		ptr -> input[i] -> loop_over = TRUE ;
	    }
	    (side == LHS_CURRENT) ? (lhs_queue[++lhs_tail] = ptr -> input[i]) 
				    : (rhs_queue[++rhs_tail] = ptr -> input[i]) ;
	}

    for (i = -1; ++i < ptr -> arity;)
	if (ptr -> input[i] -> func != IS_NULL)
	    number_nodes (ptr -> input[i], 0, side) ; 

    if (start == 1)
	if (side == LHS_CURRENT)  
	    for (i = -1; ++i <= lhs_tail;)
		 lhs_queue[i] -> node_id = i ;
	else	
	    for (i = -1; ++i <= rhs_tail;)
		 rhs_queue[i] -> node_id = i ;

    return ;
	    
} 

#if defined(__cplusplus) || defined(ANSI)
RULE_EXPR *make_expr(int id,int arg,char *name,char *arg_name,FUNC func,
                        int level)
#else
RULE_EXPR *make_expr(id,arg,name,arg_name,func,
                        level)
int                     id, arg ;
char                    *name, *arg_name ;
FUNC                    func ;
int                     level ;
#endif
{
    RULE_EXPR 	*ptr ;
    int		i ;
    ptr = (RULE_EXPR *) malloc (sizeof (RULE_EXPR)) ;

    if (name)
	strcpy (ptr -> name, name) ;
    else
	strcpy (ptr -> name, "NOT_SET") ;

    if (arg_name)
	strcpy (ptr -> arg_name, arg_name) ;
    else
	strcpy (ptr -> arg_name, "NOT_SET") ;

    ptr -> arity = 0 ;
    ptr -> func = func ; 
    ptr -> id = id ;
    ptr -> arg_id = arg ; 
    ptr -> node_id = -1 ;
    ptr -> group_id = -1 ;
    ptr -> parent = NULL ;
    ptr -> level = level ;
    ptr -> expr = FALSE ;
    ptr -> loop_over = FALSE ;
    for (i = -1; ++ i < MAX_INPUTS;)
	ptr -> input[i] = NULL ;
 
    return (ptr) ;
}

#if DEBUG>=1
void dump_symbol_table ()
{
    int i ;
    FILE *fp ;

    ((trans_rule_num == 0 && impl_rule_num == -1) 
	||
    (trans_rule_num == -1 && impl_rule_num == 0)) ? 
	(fp = fopen ("debug2.out", "w")) :
	(fp = fopen ("debug2.out", "a+")) ;

    if (current_rule_type == TRANS_RULE)
    {
	fprintf (fp, "\n\nTRANS_RULE\n") ;
	fprintf (fp, "Current rule = %d\n", trans_rule_num) ;
    }
    else
    {
	fprintf (fp, "\n\nIMPL_RULE\n") ;
	fprintf (fp, "Current rule = %d\n", impl_rule_num) ;
    }

    for (i = -1 ; ++i <= var_tab_index; )
    {
	fprintf (fp, "\nENTRY %d :\n", i) ;
	switch (symbol_table[i].func)
	{
	    case (IS_OPERATOR):


		fprintf (fp, "    Function: OPERATOR\n") ;
		break ;
	    case (IS_ALGORITHM):
		fprintf (fp, "    Function: ALGORITHM\n") ;
		break ;
	    case (IS_ENFORCER):
		fprintf (fp, "    Function: ENFORCER\n") ;
		break ;
	    case (IS_OPER_VARIABLE):
		fprintf (fp, "    Function: OPERATOR VARIABLE\n") ;
		break ;
	    case (IS_OP_ARG_VARIABLE):
		fprintf (fp, "    Function: OPER. ARGUMENT VARIABLE\n") ;
		break ;
	    case (IS_AL_ARG_VARIABLE):
		fprintf (fp, "    Function: ALGO. ARGUMENT VARIABLE\n") ;
		break ;
	    case (IS_INPUT_VARIABLE):
		fprintf (fp, "    Function: INPUT VARIABLE\n") ;
		break ;
	    default:
		fprintf (fp, "    Function: error - UNKNOWN\n") ;
		break ;
	}	
	fprintf (fp, "    Name:      %s\n", symbol_table[i].name) ;
	fprintf (fp, "    Id:        %d\n", symbol_table[i].id) ;
	fprintf (fp, "    Bind id:   %d\n", symbol_table[i].bind_id) ;
	fprintf (fp, "    arity:     %d\n", symbol_table[i].arity) ;
	fprintf (fp, "    rhs_count: %d\n", symbol_table[i].rhs_count) ;
	fprintf (fp, "    defined:   ") ;
	switch (symbol_table[i].defined_in)
	{
	    case (DEFINED_IN_RHS):
		fprintf (fp, "DEFINED IN RHS\n") ;
		break ;
	    case (DEFINED_IN_LHS):
		fprintf (fp, "DEFINED IN LHS\n") ;
		break ;
	    case (DEFINED_NOT_SET):
		fprintf (fp, "DEFINED NOT SET\n") ;
		break ;
	    default:
		fprintf (fp, "    Function: error - UNKNOWN\n") ;
		break ;
	}
	fprintf (fp, "    var_set:   ") ;
	switch (symbol_table[i].var_set)
	{
	    case (VAR_SET_IN_USER_CODE):
		fprintf (fp, "SET IN USER CODE\n") ;
		break ;
	    case (VAR_NOT_SET):
		fprintf (fp, "NOT SET\n") ;
		break ;
	    default:
		fprintf (fp, "    Function: error - UNKNOWN\n") ;
		break ;
	}
    }
    fclose (fp) ;
    return ;
}
#endif

int yywrap ()
{
    return (1) ;
}

void write_dbi_gen () 
{
    FILE *fp, *fp2 ;
    int i, j, k ;

    fp = fopen ("temp_dbi_gen.h", "w") ;
    fp2 = fopen ("temp_dbi_gen.c", "w") ;
    fprintf (fp2, "#include \"opt_defs.h\"\n") ;
    fprintf (fp2, "#include \"global.h\"\n") ;
    fprintf (fp2, "#include \"dbi_gen.h\"\n") ;
    fprintf (fp2, "#include \"dbi.h\"\n") ;
    fprintf (fp2, "#include \"opt.h\"\n") ;
    fprintf (fp2, "extern TRACE	trace ;\n") ;
    fprintf (fp2, "extern int	indent_value ;\n") ;
    fprintf (fp, "#define MAX_GROUPS	%d\n", (int) MAX_GROUPS) ;

    for (i = -1 ; ++i < rule_group_count;)
    {
	fprintf (fp, "#define %s        %d\n",
	                      rule_group_table[i].group_name, 
	                      rule_group_table[i].group_id) ; 
    }

    fprintf (fp, "\n#define TRANS_RULE_COUNT %d\n", (trans_rule_num + 1)) ;
    fprintf (fp, "\n#define IMPL_RULE_COUNT %d\n", (impl_rule_num + 1)) ;

   /*
    fprintf (fp2, "\nint trans_rule_priority[TRANS_RULE_COUNT]={") ;
    for (i = -1; ++i < trans_rule_num + 1;)
    {
	(i == 0) ?
	    fprintf (fp2, "%d", trans_rule_priority[i].priority) :
	    fprintf (fp2, ",%d", trans_rule_priority[i].priority) ;
    }
    fprintf (fp2, "};\n") ;
    fprintf (fp2, "\nint impl_rule_priority[IMPL_RULE_COUNT]={") ;
    for (i = -1; ++i < impl_rule_num + 1;)
    {
	(i == 0) ?
	    fprintf (fp2, "%d", impl_rule_priority[i].priority) :
	    fprintf (fp2, ",%d", impl_rule_priority[i].priority) ;
    }
    fprintf (fp2, "};\n") ;
   */

   /*
    fprintf (fp, "\n#if defined(__cplusplus)\n") ;
    fprintf (fp, "extern \"C\" int atoi (char *) ;\n") ;
    fprintf (fp, "extern \"C\" float atof (char *) ;\n") ;
    fprintf (fp, "\#endif\n") ;
   */

    fprintf (fp2, "\n#if defined(__cplusplus) || defined(ANSI)\n") ;
    fprintf (fp2, "char *error_text (int) ;\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
	    fprintf (fp2, 
		"STATUS derive_%s_sys_prop (", symbol_table[i].name);
	    fprintf (fp2, "    OPERATOR_ARGUMENT *,\n") ; 
	    fprintf (fp2, "    SYSTEM_PROPERTY **, LOGICAL_PROPERTY **,\n") ; 
	    fprintf (fp2, "    SYSTEM_PROPERTY *) ;\n") ; 
	    fprintf (fp2, 
		"STATUS derive_%s_log_prop (", symbol_table[i].name);
	    fprintf (fp2, "    OPERATOR_ARGUMENT *,\n") ; 
	    fprintf (fp2, "    LOGICAL_PROPERTY **,\n") ; 
	    fprintf (fp2, "    LOGICAL_PROPERTY *) ;\n") ; 
	}
	else if (symbol_table [i].func == IS_ALGORITHM
		 || symbol_table [i].func == IS_ENFORCER)
	{
	    fprintf (fp2, "STATUS cost_%s (", symbol_table[i].name) ;
	    fprintf (fp2, "    ALGORITHM_ARGUMENT *, SYSTEM_PROPERTY *,\n") ;
	    fprintf (fp2, "    LOGICAL_PROPERTY *, SYSTEM_PROPERTY **,\n") ;
	    fprintf (fp2, "    LOGICAL_PROPERTY **, PROPERTY_VECTOR *,\n") ;
	    fprintf (fp2, "    PROPERTY_VECTOR *, PROPERTY_VECTOR **,\n") ;
	    fprintf (fp2, "    COST *) ;\n") ;

	    fprintf (fp2, "STATUS do_any_good_%s (", symbol_table[i].name) ;
	    fprintf (fp2, "    OPERATOR *, OPERATOR_ARGUMENT *,\n") ;
	    fprintf (fp2, "    PROPERTY_VECTOR *, SYSTEM_PROPERTY *,\n") ;
	    fprintf (fp2, "    LOGICAL_PROPERTY *, PROPERTY_VECTOR *,\n") ;
	    fprintf (fp2, "    OPERATOR [], ALGO_APPL_RETURN *) ;\n") ;

	    fprintf (fp2, 
		"STATUS derive_%s_phy_prop (", symbol_table[i].name) ;
	    fprintf (fp2, "    ALGORITHM_ARGUMENT *, PROPERTY_VECTOR **,\n") ;
	    fprintf (fp2, "    SYSTEM_PROPERTY *, LOGICAL_PROPERTY *,\n") ;
	    fprintf (fp2, "    SYSTEM_PROPERTY **, LOGICAL_PROPERTY **,\n") ;
	    fprintf (fp2, "    PROPERTY_VECTOR *, PROPERTY_VECTOR *) ;\n") ;

	    fprintf (fp2, 
		"STATUS get_input_pv_%s (", symbol_table[i].name) ;
	    fprintf (fp2, "    ALGORITHM_ARGUMENT *, PROPERTY_VECTOR *,\n") ;
	    fprintf (fp2, "    PROPERTY_VECTOR *, int, int,\n") ;
	    fprintf (fp2, "    PROPERTY_VECTOR **, PROPERTY_VECTOR **,\n") ;
	    fprintf (fp2, "    SYSTEM_PROPERTY **, LOGICAL_PROPERTY **,\n") ;
	    fprintf (fp2, "    PHY_EXPR_LIST_ELEM **, COST **, BOOLEAN *);\n") ;
	}
    }
    fprintf (fp2, "\n#else\n") ;
    fprintf (fp2, "char *error_text () ;\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
	    fprintf (fp2, 
		"STATUS derive_%s_sys_prop () ;\n", symbol_table[i].name);
	    fprintf (fp2, 
		"STATUS derive_%s_log_prop () ;\n", symbol_table[i].name);
	}
	else if (symbol_table [i].func == IS_ALGORITHM
		 || symbol_table [i].func == IS_ENFORCER)
	{
	    fprintf (fp2, "STATUS cost_%s () ;\n", symbol_table[i].name) ;
	    fprintf (fp2, "STATUS do_any_good_%s () ;\n", symbol_table[i].name) ;
	    fprintf (fp2, 
		"STATUS derive_%s_phy_prop () ;\n", symbol_table[i].name) ;
	    fprintf (fp2, 
		"STATUS get_input_pv_%s () ;\n", symbol_table[i].name) ;
	}
    }
    fprintf (fp2, "\n#endif\n") ;


    fprintf (fp, "\n#define MAX_OPERATOR_ARITY %d\n", max_operator_arity);
    fprintf (fp, "#define NUM_OPERATORS %d\n", oper_count) ;
    fprintf (fp2, "\nint opt_arity[NUM_OPERATORS]={") ;
    for (i = -1, j = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
	   /* fprintf (fp, "#define %s %d\n", symbol_table[i].name, ++j) ; */
	     ++ j ;
	    (j == 0) ?
		fprintf (fp2, "%d", symbol_table[i].arity) :
		fprintf (fp2, ",%d", symbol_table[i].arity) ;
	}
    }
    fprintf (fp2, "};\n") ;

    fprintf (fp2, "\nint algo_arity[NUM_ALGORITHMS]={") ;
    fprintf (fp, "\n#define MAX_ALGORITHM_ARITY %d\n", max_algorithm_arity);
    fprintf (fp, "#define NUM_ALGORITHMS %d\n", algo_count) ;
    for (i = -1, j = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_ALGORITHM || 
	    symbol_table [i].func == IS_ENFORCER)  
	{
	   /* fprintf (fp, "#define %s %d\n", symbol_table[i].name, ++j) ; */
	     ++ j ;
	    (j == 0) ?
		fprintf (fp2, "%d", symbol_table[i].arity) :
		fprintf (fp2, ",%d", symbol_table[i].arity) ;
	}
    }
    fprintf (fp2, "};\n") ;

    fprintf (fp, "\n#define print_rule_group(rg)\\\n") ;
    for (i = -1, j = -1; ++i < rule_group_count;)
    {
	(++j == 0) ? 
	    fprintf (fp, "if ((rg) == %s) printf (\"%s\\n\") ;\\\n",
	                      rule_group_table[i].group_name, 
	                      rule_group_table[i].group_name) :
	    fprintf (fp, "else if ((rg) == %s) printf (\"%s\\n\") ;\\\n",
	                      rule_group_table[i].group_name, 
	                      rule_group_table[i].group_name) ; 
    }
    fprintf (fp, "else FAIL (ERR_UNKNOWN_RULE_GROUP) ;\n") ;

    fprintf (fp, "\n#define print_operator(op)\\\n") ;
    for (i = -1, j = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
	    (++j == 0) ?
		fprintf (fp, "if ((op) == %s) printf (\"%s\\n\") ;\\\n",
			 symbol_table[i].name, symbol_table[i].name) :
		fprintf (fp, "else if ((op) == %s) printf (\"%s\\n\") ;\\\n",
			 symbol_table[i].name, symbol_table[i].name) ;
	}
    }
    fprintf (fp, "else FAIL (ERR_UNKNOWN_OPERATOR) ;\n") ;

    fprintf (fp, "\n#define print_algorithm(m)\\\n") ;
    for (i = -1, j = -1, k = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_ALGORITHM ||
	    symbol_table[i].func == IS_ENFORCER)
	{
	    (++j == 0) ?
		fprintf (fp, "if ((m) == %s) printf (\"%s\\n\") ;\\\n",
			 symbol_table[i].name, symbol_table[i].name) :
		fprintf (fp, "else if ((m) == %s) printf (\"%s\\n\") ;\\\n",
			 symbol_table[i].name, symbol_table[i].name) ;

	    if (symbol_table[i].func == IS_ENFORCER)
	    {
		if (++k == 0)
		{
		    fprintf (fp2, 
			"\n#if defined(__cplusplus) || defined (ANSI)\n");
		    fprintf (fp2, 
		     "STATUS get_enforcer(int count,ALGORITHM *enforcer)\n") ;
		    fprintf (fp2, "#else\n") ;
		    fprintf (fp2, "STATUS get_enforcer(count,enforcer)\n") ;
		    fprintf (fp2, "int count ; ALGORITHM *enforcer;\n\n") ;
		    fprintf (fp2, "#endif\n") ;
		    fprintf (fp2, "{switch (count)\n{\n") ;
		}
		fprintf (fp2, "case %d: *enforcer = %s; break;\n",
			 k, symbol_table[i].name) ;
	    }
	   
	}
    }

    fprintf (fp, "else FAIL (ERR_UNKNOWN_ALGORITHM) ;\n") ;
    if (k >= 0)
    {
	fprintf (fp2, "default: *enforcer = (ALGORITHM) NO_MORE; break;\n    }\n") ;
    }
    else
    {
	fprintf (fp2, 
	    "\n#if defined(__cplusplus) || defined (ANSI)\n");
	fprintf (fp2, 
	     "STATUS get_enforcer(int count,ALGORITHM *enforcer)\n") ;
	fprintf (fp2, "#else\n") ;
	fprintf (fp2, "\nSTATUS get_enforcer(count,enforcer)\n") ;
	fprintf (fp2, "int count ; ALGORITHM *enforcer;\n\n") ;
	fprintf (fp2, "#endif\n") ;
	fprintf (fp2, "{\n*enforcer = NO_MORE;\n") ;
    }
    fprintf (fp2, "return (OKAY);\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, 
	"STATUS derive_sys_prop(OPERATOR my_operator,") ;
    fprintf (fp2, "OPERATOR_ARGUMENT *oa,SYSTEM_PROPERTY **input_sys_prop,\n") ;
    fprintf (fp2, 
	"LOGICAL_PROPERTY **input_log_prop,\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *derived_sys_prop)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, "STATUS derive_sys_prop(my_operator,oa,input_sys_prop,\n") ;
    fprintf (fp2, "input_log_prop,derived_sys_prop)\n") ;
    fprintf (fp2, "OPERATOR my_operator;OPERATOR_ARGUMENT *oa;\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *input_sys_prop[], *derived_sys_prop;\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *input_log_prop[];\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\n STATUS error;\n") ;

   /*
    fprintf (fp2, "{\nint i, num_inputs; STATUS error;\n") ;
    fprintf (fp2, "    num_inputs = opt_arity[(int) (my_operator)];\n") ;
    fprintf (fp2, "#if SAFETY>=1\n") ;
    fprintf (fp2, "for (i = -1;++i<num_inputs;)\n") ;
    fprintf (fp2, "if (sys_prop_not_set(input_sys_prop[i]))\n") ;
    fprintf (fp2, "{FAIL (ERR_NOT_INITIALIZED);}\n") ;
    fprintf (fp2, "#endif\n") ;
   */

    fprintf (fp2, "switch (my_operator) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
		fprintf (fp2, "case (%s): CALL (derive_%s_sys_prop\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(oa,input_sys_prop,input_log_prop,\n") ;
		fprintf (fp2, "derived_sys_prop)); break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_OPERATOR);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, 
	"STATUS derive_log_prop(OPERATOR my_operator,") ;
    fprintf (fp2, "OPERATOR_ARGUMENT *oa,LOGICAL_PROPERTY **input_log_prop,\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *derived_log_prop)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, "STATUS derive_log_prop(my_operator,oa,input_log_prop,\n") ;
    fprintf (fp2, "derived_log_prop)\n") ;
    fprintf (fp2, "OPERATOR my_operator;OPERATOR_ARGUMENT *oa;\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *input_log_prop[], *derived_log_prop;\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\n STATUS error;\n") ;

   /*
    fprintf (fp2, "{\nint i, num_inputs; STATUS error;\n") ;
    fprintf (fp2, "#if SAFETY>=1\n") ;
    fprintf (fp2, "num_inputs = opt_arity[(int) (my_operator)];\n") ;
    fprintf (fp2, "for (i = -1;++i<num_inputs;)\n") ;
    fprintf (fp2, "if (logical_prop_not_set(input_log_prop[i]))\n") ;
    fprintf (fp2, "{FAIL (ERR_NOT_INITIALIZED);}\n") ;
    fprintf (fp2, "#endif\n") ;
   */

    fprintf (fp2, "init_logical_prop(derived_log_prop) ;\n") ;
    fprintf (fp2, "switch (my_operator) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_OPERATOR)
	{
		fprintf (fp2, "case (%s): CALL (derive_%s_log_prop\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(oa, input_log_prop,\n") ;
		fprintf (fp2, "derived_log_prop)); break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_OPERATOR);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, "STATUS estimate_cost(") ;
    fprintf (fp2, "ALGORITHM algorithm_or_enforcer,") ;
    fprintf (fp2, "ALGORITHM_ARGUMENT *aa,") ;
    fprintf (fp2, "SYSTEM_PROPERTY *sys_prop,LOGICAL_PROPERTY *log_prop,\n") ;
    fprintf (fp2, 
     "SYSTEM_PROPERTY **input_sys_prop,LOGICAL_PROPERTY **input_log_prop,\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *needed_pv,PROPERTY_VECTOR *derived_pv,\n") ;
    fprintf (fp2, "PROPERTY_VECTOR **input_pv,\n") ;
    fprintf (fp2, "COST *ret_cost)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, 
	"STATUS estimate_cost(algorithm_or_enforcer,aa,sys_prop,log_prop,\n") ;
    fprintf (fp2, 
     "input_sys_prop,input_log_prop,needed_pv,derived_pv,input_pv,\n") ;
    fprintf (fp2, "ret_cost)\n") ;
    fprintf (fp2, "ALGORITHM algorithm_or_enforcer;ALGORITHM_ARGUMENT *aa;\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *input_sys_prop[], *sys_prop;\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *input_log_prop[], *log_prop;\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *needed_pv, *derived_pv, *input_pv[];\n") ;
    fprintf (fp2, "COST *ret_cost;\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\nSTATUS error;\n") ;

   /*
    fprintf (fp2, "#if SAFETY>=1\n") ;
    fprintf (fp2, "if (sys_prop_not_set(sys_prop))\n") ;
    fprintf (fp2, "{FAIL (ERR_NOT_INITIALIZED);}\n") ;
    fprintf (fp2, "#endif\n") ;
   */

    fprintf (fp2, "switch (algorithm_or_enforcer) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_ALGORITHM || symbol_table[i].func == IS_ENFORCER)
	{
		fprintf (fp2, "case (%s): CALL (cost_%s\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(aa, sys_prop,log_prop,\n") ;
		fprintf (fp2, "input_sys_prop,input_log_prop,\n") ;
		fprintf (fp2, "needed_pv, derived_pv, input_pv, ret_cost)) ; break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_ALGORITHM);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, 
	"STATUS do_any_good(") ;
    fprintf (fp2, "ALGORITHM * algorithm_or_enforcer,\n") ;
    fprintf (fp2, "OPERATOR * current_operator,\n") ;
    fprintf (fp2, "OPERATOR_ARGUMENT *oa,PROPERTY_VECTOR * needed_pv,\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *sys_prop,\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY * log_prop,PROPERTY_VECTOR *excl_pv,\n");
    fprintf (fp2, "OPERATOR input_operator[], \n") ;
    fprintf (fp2, "ALGO_APPL_RETURN *ret_value)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, "STATUS do_any_good(algorithm_or_enforcer,current_operator,\n") ;
    fprintf (fp2, "oa,needed_pv,sys_prop,\n") ;
    fprintf (fp2, "log_prop,excl_pv,input_operator,\n") ;
    fprintf (fp2, "ret_value)\n") ;
    fprintf (fp2, "ALGORITHM *algorithm_or_enforcer;OPERATOR *current_operator;\n") ;
    fprintf (fp2, "OPERATOR_ARGUMENT *oa;\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *sys_prop;\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *log_prop;\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *needed_pv, *excl_pv;\n") ;
    fprintf (fp2, "OPERATOR input_operator[] ;\n") ;
    fprintf (fp2, "ALGO_APPL_RETURN *ret_value ;\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\nSTATUS error;\n") ;
    fprintf (fp2, "switch (*algorithm_or_enforcer) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_ALGORITHM || symbol_table[i].func == IS_ENFORCER)
	{
		fprintf (fp2, "case (%s): CALL (do_any_good_%s\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(current_operator, oa, needed_pv,\n") ;
		fprintf (fp2, "sys_prop,log_prop,\n") ;
		fprintf (fp2, "excl_pv, input_operator, ret_value)) ; break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_ALGORITHM);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, 
	"STATUS derive_phy_prop(ALGORITHM algorithm_or_enforcer,\n") ;
    fprintf (fp2, "ALGORITHM_ARGUMENT *aa,PROPERTY_VECTOR **input_pv,\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *sys_prop,\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *log_prop,\n") ;
    fprintf (fp2, 
      "SYSTEM_PROPERTY **input_sys_prop,LOGICAL_PROPERTY **input_log_prop,\n");
    fprintf (fp2, "PROPERTY_VECTOR *needed_pv,PROPERTY_VECTOR *derived_pv)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, "STATUS derive_phy_prop(algorithm_or_enforcer,aa,input_pv,sys_prop,\n") ;
    fprintf (fp2, "log_prop,input_sys_prop,input_log_prop,needed_pv,derived_pv)\n") ;
    fprintf (fp2, "ALGORITHM algorithm_or_enforcer;ALGORITHM_ARGUMENT *aa;\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *sys_prop, *input_sys_prop[];\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *log_prop, *input_log_prop[];\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *input_pv[], *needed_pv, *derived_pv;\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\nSTATUS error;\n") ;
    fprintf (fp2, "switch (algorithm_or_enforcer) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if (symbol_table [i].func == IS_ALGORITHM || symbol_table[i].func == IS_ENFORCER)
	{
		fprintf (fp2, "case (%s): CALL (derive_%s_phy_prop\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(aa, input_pv,\n") ;
		fprintf (fp2, "sys_prop,log_prop,input_sys_prop,\n") ;
		fprintf (fp2, "input_log_prop,needed_pv,derived_pv)) ; break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_ALGORITHM);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fprintf (fp2, 
	"\n#if defined(__cplusplus) || defined (ANSI)\n");
    fprintf (fp2, 
	"STATUS get_input_pv(ALGORITHM algorithm_or_enforcer,\n") ;
    fprintf (fp2, "ALGORITHM_ARGUMENT *aa,PROPERTY_VECTOR *needed_pv,\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *excl_pv,\n") ;
    fprintf (fp2, 
	"int input_id,int iteration,PROPERTY_VECTOR **input_pv,\n") ;
    fprintf (fp2, "PROPERTY_VECTOR **input_excl_pv,\n") ;
    fprintf (fp2, 
      "SYSTEM_PROPERTY *input_sys_prop[],LOGICAL_PROPERTY *input_log_prop[],\n");
    fprintf (fp2, "PHY_EXPR_LIST_ELEM **input_ret_pele,\n") ;
    fprintf (fp2, "COST **input_ret_cost,BOOLEAN *fail)\n") ;
    fprintf (fp2, "#else\n") ;
    fprintf (fp2, "STATUS get_input_pv(algorithm_or_enforcer,aa,needed_pv,excl_pv,\n") ;
    fprintf (fp2, "input_id,iteration,input_pv,input_excl_pv,\n") ;
    fprintf (fp2, 
      "input_sys_prop,input_log_prop,input_ret_pele,input_ret_cost,fail)\n") ;
    fprintf (fp2, "ALGORITHM algorithm_or_enforcer;ALGORITHM_ARGUMENT *aa;\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *needed_pv, *excl_pv ;\n") ;
    fprintf (fp2, "PROPERTY_VECTOR *input_pv[], *input_excl_pv[] ;\n") ;
    fprintf (fp2, "int input_id, iteration;\n") ;
    fprintf (fp2, "SYSTEM_PROPERTY *input_sys_prop[] ;\n") ;
    fprintf (fp2, "LOGICAL_PROPERTY *input_log_prop[] ;\n") ;
    fprintf (fp2, "PHY_EXPR_LIST_ELEM *input_ret_pele[] ;\n") ;
    fprintf (fp2, "COST *input_ret_cost[] ;\n") ;
    fprintf (fp2, "BOOLEAN *fail ;\n") ;
    fprintf (fp2, "#endif\n") ;
    fprintf (fp2, "{\nSTATUS error;\n") ;
    fprintf (fp2, "switch (algorithm_or_enforcer) {\n") ;
    for (i = -1; ++i <= sym_tab_index; )
    {
	if ((symbol_table [i].func == IS_ALGORITHM || symbol_table[i].func == IS_ENFORCER)
		&& symbol_table[i].arity > 0)
	{
		fprintf (fp2, "case (%s): CALL (get_input_pv_%s\n",
			 symbol_table[i].name, symbol_table[i].name) ;
		fprintf (fp2, "(aa, needed_pv,\n") ;
		fprintf (fp2, "excl_pv,input_id,iteration,\n") ;
		fprintf (fp2, "input_pv,input_excl_pv,\n") ;
		fprintf (fp2, 
		 "input_sys_prop,input_log_prop,input_ret_pele,input_ret_cost,fail));\n");
		fprintf (fp2, "break;\n") ;
	}
    }
    fprintf (fp2, "default: FAIL (ERR_UNKNOWN_ALGORITHM);\n") ;
    fprintf (fp2, "}\n") ;
    fprintf (fp2, "return (OKAY) ;\n") ;
    fprintf (fp2, "}\n") ;

    fclose (fp) ;
    fclose (fp2) ;

    return ;

}

#if defined(__cplusplus) || defined(ANSI)
char *my_strstr (char *s1, char *s2)
#else
char *my_strstr (s1,s2)
char *s1, *s2 ;
#endif
{
    int len = strlen (s2) ;
    char *sp = NULL ;
  
    for (; s1 != NULL; ++s1, sp = NULL)
    {
	sp = index (s1, *s2) ;
	if (sp == NULL) break ; 
	if (strncmp (s2, sp, len) == 0) break ;
    }

    return ((char *) sp) ;
    
}

