
#include "bplusddl.h"

#include "index.h"


/****************************************************************************
 
        parses the following statement :
 
        ... [ UNIQUE ]  indexName ON className attributeName [ ASC | DESC ]
 
*****************************************************************************/

IndexCreate::parse()
{

        connectUserRequest();

        // check the UNIQUE keyword
        getToken();
        if( tokenType() == MS_UNIQUE ) setUniqueFlag();
        else backtrack();

        // get index name
        CHECK_TOKEN(IDENTIFIER, 9, "");
        setIndexName(tokenValue());

        // must be On keyword
        CHECK_TOKEN(ON, 15, "");

        // get class name
        CHECK_TOKEN(IDENTIFIER, 6, "");
        setClassName(tokenValue());
 
        attributeList = new LinkedList;

        do {
                IndexedAttribute *indexedAttribute;
                indexedAttribute = new IndexedAttribute;
 
                // get indexed attribute name
                CHECK_TOKEN(IDENTIFIER, 10, "");
                indexedAttribute->setAttributeName(tokenValue());
 
                // order for the attribute is determined here
                getToken();
                if(tokenType() == DESC)
                        indexedAttribute->setOrder();
                else if(tokenType() != ASC) backtrack();
 

      // insert indexed attribute to the list
                attributeList->insert(INDEXED_ATTRIBUTE, VOID indexedAttribute);

        } while(getToken() == COMMA ); // continue until not comma

        backtrack();
        return TRUE;
}

IndexCreate::IndexCreate( UserRequest *userReq)
{
        ur = userReq;
        strcpy(indexName,"");
        strcpy(className,"");
        uniqueFlag    = FALSE;
        attributeList = NULL;
}

IndexCreate::execute()
{
	CatBOOLEAN	status;
	IndexClass	ic;
	char		attName[80], dummy;
	strcpy( attName, 
	((IndexedAttribute *) attributeList->firstItem->item)->attributeName);
	ic.createIndex( indexName, className, attName, uniqueFlag, dummy, &status);
	if( status )
	{
		//setError( ?, ?, ? );	
		return FALSE;
	}
	return TRUE;
}


IndexDrop::parse()
{
        connectUserRequest();
 
        // get index name
        CHECK_TOKEN(IDENTIFIER, 9, "");
        setIndexName(tokenValue());
}

IndexDrop::execute()
{
	CatBOOLEAN	status;
	IndexClass	ic;
	ic.dropIndex( indexName, &status );
	if( status )
	{
		// setError
		return FALSE;
	}
	return TRUE;
}




