/* create class cccc inherits from xxxx */

#ifndef __CLASSDDL_H
#include "classddl.h"
#endif

#include "mood_parser.h"

#ifndef __KEYWORDS_H
#include "keywords.h"
#endif

#include "mood_catalog.h"

extern void readLine( char *, FILE *);



int ClassCreate::parse()
{
        return createClassParser();
}



int ClassCreate::createClassParser()
{
	MoodsTypes typeclass;

	connectUserRequest();

	memberNode *p = createMemberNode();

	//IS_THIS_KEYWORD(CREATE);

	//CHECK_TOKEN(CLASS, CLASS_KYWD_ERR, "" ); 

	CHECK_TOKEN(IDENTIFIER,CLS_NAME_EXPECTED, "");

	setClassName( tokenValue() );
    
	getToken();

	// TYPE definitions are not allowed to inherit
	if( tokenType() == INHERITS && typeFlag == 2 ){
		ur->setError( PARSE_ERROR, INH_NOT_ALLOWED, " " );
		return FALSE;
	}

	if( tokenType() == INHERITS  ){
		if( !inheritsClause() ) {
			myFree((char *) p );
			return FALSE;
		}
	}else backtrack();

	getToken();
	if( tokenType() != TUPLE ){
		if( !inheritList ){
			ur->setError( PARSE_ERROR, TUPLE_EXPECTED, " " );
			return FALSE;
		}
		backtrack();
		return TRUE; 
	}
	backtrack();
    
    	if(!typedefClause( p )){
       		myFree((char *)  p );
       		return FALSE;
    	}
    
    	/*getToken();
	if( tokenType() != SEMICOLON ){
		ur->setError( PARSE_ERROR, SEMICOLON_EXPECTED, " " );
		return FALSE;
	}*/

	return TRUE;

} 


int ClassCreate::semanticCheck( memberNode *p )
{
	MoodsTypes mt;
	CatBOOLEAN status=0;
	int typeId;

	if( p == NULL ) return TRUE;

	typeId = getTypeId( p->typeName );

	if( typeId == 0 ){
		ur->setError( PARSE_ERROR, INVLD_TYPE_NAME, p->typeName );
		return FALSE;
	}

	p->type = typeId;

	if( !semanticCheck( p->attributeList ) ) return FALSE;
	if( !semanticCheck( p->subAttributes ) ) return FALSE;

	return TRUE; 
}

int ClassCreate::createClass()
{
	MoodsTypes mt;
	CatBOOLEAN status=0;
	int typeId = getTypeId( className );

	if( typeId != 0 ) {
		ur->setError( PARSE_ERROR, CLASS_REDEFINITION, className );
		return FALSE;
	}
	if( !semanticCheck( members ) ) return FALSE;
	if( !mt.addType( className, inheritList, members, typeFlag ,&status)) 
		return FALSE;
	return TRUE;

}

int ClassCreate::inheritsClause()
{

	//IS_THIS_KEYWORD( INHERITS );

	CHECK_TOKEN( FROM, FROM_KYWD_ERR, " " );
	
	inheritList = new LinkedList;

	if( !parseClassList() ) return FALSE;

	return TRUE;

}

int ClassCreate::parseClassList() 
{
   
    inherits *p  = createInheritNode();
 
    getToken();

    if( tokenType() == VIRTUAL ){
        getToken(); 
        p->virtualFlag = 1;
    }

    /* set the PRIVATE flag if just an identifier exists 
       otherwise set the PUBLIC or PRIVATE flag by inspecting
       the keyword . */ 
    switch( tokenType()){
       case PUBLIC : p->protectionFlag = cpPUBLIC;
                     getToken();
                     break;

       case PRIVATE: p->protectionFlag = cpPRIVATE;
                     getToken();
                     break;
    }

    if( tokenType() != IDENTIFIER ){
        myFree((char *)  p  );
	ur->setError(PARSE_ERROR, CLS_NAME_EXPECTED, "");
        return FALSE;
    }

    /* add the name of the class into the inheritance list */ 
    strcpy( p->cname, tokenValue() ); 
    inheritList->insert( INHCONS, ( void * )p );

    getToken();
    if( tokenType() == COMMA ) return( parseClassList() );
    else backtrack();
  
    return TRUE;

}


int ClassCreate::tupleDef( memberNode * p )
{

     memberNode *localPtr = createMemberNode();

     getToken();


     if( tokenType() != LPAR ) {
	ur->setError(PARSE_ERROR, MISSING_LP, "");
        return FALSE;
     }
     
     p->subAttributes = localPtr;

     if( !declarationList( localPtr ) ) return FALSE;

     getToken();

     if( tokenType() != RPAR ) {
	ur->setError(PARSE_ERROR, MISSING_RP, "");
	return FALSE;
     }

     return TRUE;
}

void ClassCreate::strup( char *str )
{

    while( *str ){
        if( ( *str >= 'a' ) && ( *str <='z' ) ) 
            *str =  'A' + *str - 'a' ;
        str++;
    }

}

int ClassCreate::getTypeId( char * typeName )
{
     CatBOOLEAN status;
     OID typePtr;
     Catalog mycat;
     MoodsTypes mt;

     mt.getTypOid( typeName, &typePtr, &status );
     if( status ) return FALSE;

     mt.readType( typePtr, &mycat, &status );

     return mycat.TypeId;
}

int ClassCreate::setListRefDef( memberNode* p )
{

     memberNode *localPtr = createMemberNode();
	
     p->subAttributes = localPtr;

     getToken();
     if( tokenType() != LPAR ){
	myFree((char *) localPtr );
        p->subAttributes = NULL;
	ur->setError(PARSE_ERROR, MISSING_LP, "");
	return FALSE;
     }
 
     getToken();
     if( tokenType() == RPAR ){
	myFree((char *) localPtr );
        p->subAttributes = NULL;
        return TRUE;
     }
     backtrack();

     if( !allTypedefClauses( localPtr ) ){
	myFree((char *) localPtr );
	return FALSE;
     }

     getToken();

     if( tokenType() != RPAR ){
	myFree((char *) localPtr );
	ur->setError(PARSE_ERROR, MISSING_RP, "");
	return FALSE;
     }

     return TRUE;

}


int ClassCreate::declarationList( memberNode *p )
{
    memberNode *localPtr; 

    CHECK_TOKEN( IDENTIFIER, INVLD_ATT_NAME, tokenValue());

    if( notReplicated( members, tokenValue() ) )    
         strcpy( p->memberName, tokenValue() );
    else
         return FALSE;

    if( !allTypedefClauses( p ) )
       return FALSE;

    getToken();

    if( tokenType() == COMMA ){
       localPtr = createMemberNode();
       p->attributeList = localPtr;
       if( declarationList( localPtr ) ) return TRUE;
	myFree((char *) localPtr );
       return FALSE;
    }else backtrack();
  
    return TRUE;     
} 


int ClassCreate::stringDef( memberNode *p )
{
//    CHECK_TOKEN( IDENTIFIER, INVLD_ATT_NAME," " );

 /*   if( notReplicated( members, tokenValue() ) )
        strcpy( p->memberName, tokenValue() );
    else
        return FALSE;
*/
    getToken();
    
    if( tokenType() != LB )
         return FALSE;

     getToken();

     if( tokenType() != UNSIGNEDINT )
         return FALSE;

     p->size = atoi( tokenValue() );

     getToken();

     if( tokenType() != RB ) return FALSE;

     return TRUE; 
}

int delimeter( char *token )
{
   if( strcmp( token, "," ) == 0 ) return 1;
   if( strcmp( token, "." ) == 0 ) return 1;
   if( strcmp( token, ")" ) == 0 ) return 1;
   if( strcmp( token, "(" ) == 0 ) return 1;
   if( strcmp( token, "[" ) == 0 ) return 1;
   if( strcmp( token, "]" ) == 0 ) return 1;
   if( strcmp( token, "{" ) == 0 ) return 1;
   if( strcmp( token, "}" ) == 0 ) return 1;
   if( strcmp( token, "/" ) == 0 ) return 1;
   if( strcmp( token, "*" ) == 0 ) return 1;
   if( strcmp( token, "+" ) == 0 ) return 1;
   if( strcmp( token, "-" ) == 0 ) return 1;
   if( strcmp( token, "=" ) == 0 ) return 1;
   if( strcmp( token, "'" ) == 0 ) return 1;
   if( strcmp( token, "&" ) == 0 ) return 1;
   if( strcmp( token, "!" ) == 0 ) return 1;
   if( strcmp( token, "@" ) == 0 ) return 1;
   if( strcmp( token, "%" ) == 0 ) return 1;
   if( strcmp( token, "\\" ) == 0 ) return 1;
   if( strcmp( token, "|" ) == 0 ) return 1;
   return 0;
}

int ClassCreate::allTypedefClauses( memberNode* p )
{
   int typeid; 

   getToken();
   
   if( delimeter( tokenValue() ) ){ 
     printf("%s instead of an identifier !!!...\n", tokenValue() );
     return FALSE;
   }
 
//   typeid = getTypeId( tokenValue() );
//   printf("Type id:%d\n", typeid );

//   p->type = typeid;

   strcpy(p->typeName, tokenValue());

   switch( tokenType() ){
      case TUPLE   :
		     return tupleDef( p );
      case MS_SET  : 
		     return setListRefDef( p );
      case LIST    : 
		     return setListRefDef( p );
      case REF     : 
		     return setListRefDef( p );
      case INTEGER : 
                     return TRUE; 
      case FLOAT   : 
                     return TRUE; 
      case CHAR    : 
                     return TRUE; 
      case MS_BOOLEAN : 
                     return TRUE; 
      case STRING  : 
                     return stringDef( p );
   }     

   // this means that it is a reference 
   //if( typeid )
    //    return TRUE;

   //ur->setError(PARSE_ERROR, 1, " UNDEFINED CONSTRUCTOR"); 
   return TRUE; 
} 
 
int ClassCreate::notReplicated( memberNode *p, char *name )
{
    if( p == NULL ) return TRUE;
    
    if( strcmp( p->memberName, name ) == 0 ) return FALSE; 
    
    if( notReplicated( p->subAttributes, name ) )
        return notReplicated( p->attributeList, name );

    printf(" %s is used previously.\n", name );

    return FALSE;

    /* this part is added in order TUPLE definition to be forced 
    getToken();
    if( tokenType() != TUPLE ){
        printf("Only TUPLE definition is permitted...\n" );
        return FALSE;
    }
	 */
 
    backtrack();
    /* end of the forbidding zone */
}

int ClassCreate::typedefClause( memberNode *p )
{
    memberNode *localPtr;

    if( members == NULL ) members = p;

    strcpy( p->memberName, " " );
    
    if( !allTypedefClauses( p ) )
        return FALSE;


    /* this part is added in order to restrict the outermost definition to be TUPLE */
	 getToken(); /*
	 if( tokenType() != TUPLE ){
		  printf("Only TUPLE definition is permitted in the first level...\n" );
		  return FALSE;
    } */
    backtrack();
	 /* end of the restrictive part */

    return TRUE;
}

void ClassCreate::deleteMembers( memberNode *members ) 
{
   
    if( members == NULL ) return;
    
    deleteMembers( members->subAttributes );
    deleteMembers( members->attributeList );

    myFree((char *) members );
}

inherits *ClassCreate::createInheritNode()
{
    inherits *p = ( inherits * )myAlloc( sizeof( inherits ) );

    p->virtualFlag = 0;
    p->protectionFlag = cpPUBLIC;
    p->next = NULL;

    return p;
}


memberNode *ClassCreate::createMemberNode()
{
   
    memberNode *p = ( memberNode * )myAlloc( sizeof( memberNode ) );

    strcpy( p->memberName, "" );
    p->subAttributes = NULL;
    p->attributeList = NULL;
    p->size = 0;

    return p;

}

void ClassCreate::printMembers( struct memberNode *p )
{

    if( p == NULL ) return; 

    switch( p->type ){
         case ST:  printf("%s SET of ( ",p->memberName);
                      printMembers( p->subAttributes );
                      printf(" ) ");
                      printMembers( p->attributeList ); 
                      break;
         case RF:  printf(" %s REF of ( ",p->memberName);
                      printMembers( p->subAttributes );
                      printf(" ) ");
                      printMembers( p->attributeList ); 
                      break;
         case LST: printf(" %s LIST of( ",p->memberName);
                      printMembers( p->subAttributes );   
                      printf(" ) ");
                      printMembers( p->attributeList );
                      break;
         case TPL: printf(" %s TUPLE of ( ",p->memberName);
                       printMembers( p->subAttributes );
                       printMembers( p->attributeList );
                       printf(" ) ");
                       break;
         case STR:   printf(" %s STRING[ %d ] ", p->memberName, p->size );
                       printMembers( p-> attributeList );
                       break;
         case INT:   printf(" %s %s ", p->memberName, "INTEGER");
                       printMembers( p->subAttributes );
                       printMembers( p->attributeList );
                       break;
         case CHR:  printf(" %s %s ", p->memberName, " CHAR " );
                       printMembers( p->attributeList );
                       break; 
         case FLT: printf(" %s %s ", p->memberName, "FLOAT");
                       printMembers( p->attributeList );
                       break;
         case BL: printf(" %s %s ", p->memberName, "BOOLEAN");
                         printMembers( p->attributeList );
                         break;
    }
}



ClassDrop::parse()
{
	connectUserRequest();
	char	*p;

	do {

		CHECK_TOKEN(IDENTIFIER, CLS_NAME_EXPECTED, "");

		p = (char *) myAlloc( MAXSTRLEN );
		strcpy( p,  tokenValue() );

		clsList->insert( STRCONS, p  );

		getToken();

	} while( tokenType() == COMMA );

	backtrack();

	return TRUE;
}

deleteIncludeLine( char *className )
{
        FILE    *source, *dest;
        char    fileName[120], backupName[120];
        char    line[180], searchLine[2*MAXSTRLEN];

        sprintf( fileName, "%s/database.h",  DatabaseHome );
        sprintf( backupName, "%s.old", fileName );
        rename( fileName, backupName );

	sprintf(searchLine, "#include <%s.h>\n", className );

        dest   = fopen( fileName,  "w" );
        source = fopen( backupName, "r");

        if( source == NULL || dest == NULL) return FALSE;

        while( !feof( source ) )
        {
                readLine( line, source );
                if( strcmp( line, searchLine ) ) fprintf( dest, "%s", line );
        }
        fclose( source );
        fclose( dest );

        return TRUE;

}


deleteClassLine( char *className )
{
        FILE    *source, *dest;
        char    fileName[120], backupName[120];
        char    line[180], searchLine[2*MAXSTRLEN];

        sprintf( fileName, "%s/classes.h",  DatabaseHome );
        sprintf( backupName, "%s.old", fileName );
        rename( fileName, backupName );

	sprintf(searchLine, "class %s;\n", className );

        dest   = fopen( fileName,  "w" );
        source = fopen( backupName, "r");

        if( source == NULL || dest == NULL) return FALSE;

        while( !feof( source ) )
        {
                readLine( line, source );
                if( strcmp( line, searchLine ) ) fprintf( dest, "%s", line );
        }
        fclose( source );
        fclose( dest );

        return TRUE;

}


ClassDrop::execute()
{
	ListItem	*p, *q;
	MoodsTypes	moodsType;
	Catalog		ct;
	AttrList	*al;
	CatBOOLEAN	notFound = 0;
	char		*className;
	SortedStringList	*refList;
	LinkedList	*superCls, *subCls;

	refList = new SortedStringList(); 

	// Check the existence of classes
	for( p = clsList->firstItem; p; p=p->next ) 
	{
		className  = (char *) p->item;

		superCls = new LinkedList();
		subCls = new LinkedList();

		if( !moodsType.getClassInfo(className,&ct,&al,superCls,subCls,&notFound ))
		{
			delete al;
			delete subCls;
			delete superCls;
			ur->userError = catalogError;
			ur->operationType = HANDLE_ERROR;
			return FALSE;
		}

		if(al) 	delete al;
		if( superCls) delete superCls;
		if( subCls ){
			for( q = subCls->firstItem; q; q = q->next )
				refList->insertSorted( STRCONS, q->item );
			subCls->deleteItems();
		}

		if( notFound ) {
			ur->setError( PARSE_ERROR, INVLD_CLSNAME_ERR, className );
			return FALSE;
		}
	}
	

	if( !moodsType.checkReference( clsList, refList ) ) {
		delete clsList;
		delete refList;
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	//refList->printOn();

	char sqlText[3*MAXSTRLEN];
	char fileName[2*MAXSTRLEN];
	UserRequest *childRequest;
	int	force = 1, isClsFound = 1;

	// Check if there exist refrenced class
	for( p = refList->firstItem; p; p=p->next )
	{
		isClsFound = 0;
		for( q = clsList->firstItem; q; q=q->next ) {
			if( !strcmp( (char *) p->item, (char *) q->item ) ) {
				isClsFound = 1;
				break;
			}
		}
		if( !isClsFound ) break;
	}

	if( !isClsFound )
	{
		ur->setError( RUNTIME_ERROR, HAS_REFERENCE, (char *) p->item );
		delete refList;
		delete clsList;
		return FALSE;
	}

	// Delete Instances of the dropped classes
	for( p = clsList->firstItem; p; p=p->next ) 
	{
		className  = (char *) p->item;
		sprintf( sqlText, "DELETE dsfgh FROM %s dsfgh;", className );
		childRequest = new UserRequest( sqlText );
		childRequest->parseText();
		childRequest->handleOperation();
	}

	// Now delete the catalog info for these classes
	for( p = clsList->firstItem; p; p=p->next ) 
	{
		className  = (char *) p->item;
		if( !moodsType.deleteClass( className, &force, &notFound ) )
			return FALSE;

		// Now remove the associated  header files
		sprintf( fileName, "%s/%s.h", DatabaseHome, className );
		unlink( fileName );

		deleteClassLine( className );
		deleteIncludeLine( className );
	}

	return TRUE;
}

