
#include "mood_catalog.h"
#include "mood_parser.h"
#include "linklist.h"




char *catalogText = " Begin Declaration {

create class FunctionType
tuple(
	retType		Integer,
	funcName	String[ 100 ],
	isMember	Integer,
	className	String[ 100 ],
	fsignature	String[ 100 ],
	argc		Integer,
	retTypeName	String[ 32 ],
	prototypeStr	String[ 200 ]
)


create class AttributeType
tuple (
	AttributeName		String[100],
	AttributeType		REF( Catalog ),
	AttrTypeName		String[100],
	AttributeOffset 	Integer,
	AttributeConstructor	Char,
	ClassMember		Char
)

create class Catalog
tuple(
	TypeName		String[100],
	TypeId			Integer,
	Size			Integer,
	ClassType		Char,
	List_Attributes		List( AttributeType ),
	List_SuperClass		List( Catalog ),
	List_SubClass		List( Catalog ),
	List_Function		List( FunctionType ),
	Instance_File		FID,
	Cardinality		Integer
) 

create class MOODS_INDEX
tuple(
	IndexName		String[100],
	ClassName		String[100],
	AttributeName		String[100],
	AttributeTypeId		Integer,
	UniqueFlag		Char,
	IndexType		Char,
	IndexId			IID
)

}; ";



CreateDatabase::parse()
{
	connectUserRequest();
        CHECK_TOKEN( IDENTIFIER, DB_NAME_EXPECTED, tokenValue() );
        setDatabaseName( tokenValue() );
        return TRUE;
}

CreateDatabase::execute()
{
        Database        db;
	UserRequest	*childReq;

        if( !db.createDatabase( databaseName ) ) {
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

/********************
	// do not create catalog classes now
	childReq = new UserRequest( catalogText );

	if( !childReq->parseText() ) {
		ur->userError = childReq->userError;
		delete childReq;
		return FALSE;
	}
	
	if( !childReq->handleOperation() ) {
		ur->userError = childReq->userError;
		delete childReq;
		return  FALSE;
	}

	delete childReq;
********************/

        return TRUE;
}

UseDatabase::parse()
{
	connectUserRequest();
        CHECK_TOKEN( IDENTIFIER, DB_NAME_EXPECTED, tokenValue() );
        setDatabaseName( tokenValue() );
        return TRUE;
}
 
UseDatabase::execute()
{
        Database        db;
	FILE		*fp;
	char		fileName[2*MAXSTRLEN];

        if( !db.useDatabase( databaseName ) ) {
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	sprintf( fileName, "%s/DATABASES/%s/database.h", MoodHome, databaseName );
	fp = fopen( fileName, "r" );
	if( fp == NULL )
	{
		sprintf( fileName, "%s/%s", MoodHome, databaseName );
		ur->userError.setError( RUNTIME_ERROR, DIR_NOT_FOUND, fileName );
		noDBInUse = 1;
		return FALSE;
	}
	fclose( fp );

        return TRUE;
}


DropDatabase::parse()
{
	connectUserRequest();
        CHECK_TOKEN( IDENTIFIER, DB_NAME_EXPECTED, tokenValue() );
        setDatabaseName( tokenValue() );
        return TRUE;
}
 
DropDatabase::execute()
{
        Database        db;
	FILE		*fp;
	char		fileName[ 2*MAXSTRLEN ];

        if( !db.useDatabase( databaseName ) ) {
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

        sprintf( fileName, "%s/DATABASES/%s/database.h", MoodHome, databaseName );
        fp = fopen( fileName, "r" );

        if( fp == NULL )
        {
                sprintf( fileName, "%s/%s", MoodHome, databaseName );
                ur->userError.setError( RUNTIME_ERROR, DIR_NOT_FOUND, fileName );
                noDBInUse = 1;
                return FALSE;
        }

        fclose( fp );

        if( !db.deleteDatabase( databaseName ) ) {
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}
        return TRUE;
}

ListDatabase::execute()
{
	Database	db;
	ListItem	*p;

	dbList = new LinkedList();

	if( !db.listDatabases( dbList ) )
	{
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		delete dbList;
		return FALSE;
	}

	if( interfaceMode == TEXTUAL_MODE ) {
		for( p = dbList->firstItem; p; p=p->next)
			printf("%s\n", (char *) p->item );
	}

	return TRUE;
}


ListType::execute()
{
	MoodsTypes	mt;
	ListItem	*p;	
	CatalogTypeItem	*cti;

	typeList = new	LinkedList();

	if( !mt.listTypes( typeList ) ) {
		delete typeList;
		ur->userError = catalogError;
		ur->operationType = HANDLE_ERROR;
		return FALSE;
	}

	if( interfaceMode == TEXTUAL_MODE ) {

	printf("%30s %4s %4s %12s \n", "TYPE NAME", "SIZE", "ID", "TYPE TYPE!");
	printf("%30s %4s %4s %12s \n", 
		"_____________________________", 
		"____", "____",  "____________" );
	for( p = typeList->firstItem; p; p=p->next)
	{
		cti = (CatalogTypeItem * ) p->item;
		if( strcmp( cti->TypeName, " ")) {
			printf("%30s %4d %4d ", cti->TypeName,cti->Size, cti->TypeId) ;
			switch( cti->ClassType )
			{
				case 'C' :
					printf("SYSTEM CLASS"); break;
				case 'c' :
					printf("USER CLASS"); break;
				case 'S' :
				case 'R' :
				case 'L' :
				case 'T' :
					printf("SYSTEM TYPE"); break;
				case 't' :
					printf("USER TYPE"); break;
				case 'B' :
					printf("BASIC TYPE"); break;
			}
			printf("\n");
		}
	}
	}

	return TRUE;
}

int DisplayHierarchy::parse()
{
	connectUserRequest();
	// put a propoer number for 1, later !!!!!!
	CHECK_TOKEN( HIERARCHY, 1, "" );
	return TRUE;
}

int DisplayHierarchy::execute()
{
	CatBOOLEAN	status;
	MoodsTypes	mt;
	if( !mt.getClassHierarchy( "MoodsRoot", &hier, &status) )  return FALSE;
	return TRUE;
}

StartFile::parse()
{
	connectUserRequest();
        CHECK_TOKEN( STRCONS, MISSING_FILE_NAME, "" );
        strcpy( fileName, tokenValue() );
	return TRUE;
}	

StartFile::execute()
{
	FILE *fp;
	int ch;
	char msqlText[30000], fileContent[30000], *textPtr, *textEnd;
	UserRequest	*childRequest;

        // Check the existence of the filename
        fp = fopen( fileName, "r" );
        if( fp == NULL ) {
                ur->setError( RUNTIME_ERROR, INVALID_FILENAME, fileName );
                return FALSE;
        }

	// read file contents
	textPtr = fileContent;
	ch = getc( fp );
	while( ch != EOF )
	{
		*textPtr = ch;
		textPtr++;
		ch = getc( fp );
	}
	*textPtr = '\0';
        fclose( fp );
	
	// Now execute file
	textPtr = fileContent;
	textEnd = strchr( textPtr, ';' );

	while( textEnd )
	{
		strncpy( msqlText, textPtr, (int) (textEnd - textPtr + 1) );
		msqlText[ (int) (textEnd - textPtr + 1) ] = '\0';

		childRequest = new UserRequest( msqlText );

		if( childRequest->parseText() )
			childRequest->handleOperation();

		if( childRequest->operationType == HANDLE_ERROR  &&
		    childRequest->userError.errorType != NO_ERROR ) {
			ur->operationType = HANDLE_ERROR;
			ur->userError = childRequest->userError;
			return FALSE;
		}

		//childRequest->handleError();

		childRequest->commitXAct();

                childRequest->startXAct();

		//delete childRequest;

		textPtr = textEnd + 1;
		textEnd = strchr( textPtr, ';' );
	}

	return TRUE;
}
