#include "decldml.h"

void DeclStmnt::connectUserRequest( )
{
	ur->operation		= VOID this;
	ur->operationType	= DECLARE_STMNT;
	ur->operationTypeOnError= DECLARE_STMNT;	
}

int DeclStmnt::isElement( char *className, LinkedList *list )
{

	if( !list ) return FALSE;

	char *item = ( char * )(list->getFirstEntry());
	ClassCreate *cl;

	while( item ){
		if( strcmp( className, item ) == 0 ){
			ur->setError( PARSE_ERROR, CYCLIC_DEFINITION, className );
			return TRUE;
		}
		if( ( cl=findClass(item) ) )
			if( cyclic( className, cl ) ){
				ur->setError( PARSE_ERROR, CYCLIC_DEFINITION, className );
				return TRUE;
			}
		item = ( char * )(list->getNextEntry());
	}

	return FALSE;
}

ClassCreate *DeclStmnt::findClass( char *className )
{
	ClassCreate	*cl = NULL;
	int 		size = list.size, i;

	for( i = 0; i < size; i++ ){
		cl = (ClassCreate *)(list.listAt(i)->item);
		if( strcmp( cl->className, className ) == 0 ) return cl;		
	}
	return NULL;
}


int DeclStmnt::cyclic( char *className, ClassCreate *p )
{
	memberNode *members;
	ClassCreate *newClass;

	if( isElement( className, p->inheritList )) return TRUE;

	members = p->members->subAttributes;
	while( members ){
		if( strcmp( className, members->typeName ) == 0 ){
			ur->setError( PARSE_ERROR, CYCLIC_DEFINITION, className ); 
			return TRUE;
		}
		if( (newClass=findClass( members->typeName ) ) )
			if( cyclic( className, newClass ) ) return TRUE;
		members = members->attributeList;
	} 
	return FALSE;
}

int DeclStmnt::cycleCheck()
{
	int		size = list.size, i;
	ClassCreate	*p;

	for( i = 0; i < size; i++ ){
		p = ( ClassCreate * )(list.listAt(i)->item);
		if( cyclic( p->className, p )  ) return FALSE;
	}

	return TRUE;
}

int DeclStmnt::defined( char * cname, int pos = -1 )
{
        int i, listLength = list.size;
        ClassCreate     *p;
        MoodsTypes      mt;
        CatBOOLEAN      status = FALSE;
        Catalog         cat;

	if( mt.readType( cname, &cat, &status ) && !status ) return TRUE;

	for( i = 0 ; i < listLength; i++ )
		if( i != pos ){
			p = ( ClassCreate *)(list.listAt( i )->item);
			if( !strcmp(p->className, cname ) ) return TRUE;
		}

        return FALSE;
}

/* checks whether all the types in the member list are defined */
int DeclStmnt::memberCheck( memberNode *member )
{
	if( member == NULL ) return TRUE;

	if( !defined( member->typeName ) ){
		ur->setError( PARSE_ERROR, INVLD_TYPE_NAME, member->typeName );
		return FALSE;
	}
	
	if( !memberCheck( member->attributeList ) ) return FALSE;
	if( !memberCheck( member->subAttributes ) ) return FALSE;

	return TRUE;
}

/* check whether a type name exists in the catalog or 
   is declared currently */
int DeclStmnt::typeNameSemantics()
{
        int i, listLength = list.size;
        ClassCreate     *p;

	for( i = 0; i < listLength; i++ ){
		p = ( ClassCreate *)(list.listAt( i )->item);
		if( !memberCheck( p->members ) ) return FALSE;
	}

	return TRUE;
}

/* check whether a class is multiply defined */
int DeclStmnt::multipleDeclarationCheck()
{
	int i, listLength = list.size;
	ClassCreate     *p;

	for( i = 0; i < listLength; i++ ){
		p = ( ClassCreate *)(list.listAt( i )->item);
		if( defined( p->className, i )){
			ur->setError( PARSE_ERROR, CLASS_REDEFINITION, p->className );
			return FALSE;
		}
        }

	return TRUE;
}

int DeclStmnt::checkInheritanceList( LinkedList * p )
{
	ClassCreate *node;
	Catalog cat;
	MoodsTypes  mt;
	CatBOOLEAN  status=FALSE;
	int		i, listLength = list.size;

	if( p == NULL ) return TRUE;

	char *item = ( char * )(p->getFirstEntry());

	while( item ){
		if( mt.readType( item, &cat, &status ) && !status ){
			if( cat.ClassType == 't' || cat.ClassType == 'T' ){
				ur->setError( PARSE_ERROR, TYPE_INHERITANCE, cat.TypeName );
				return FALSE;
			}		
		} else {
			for( i = 0; i < listLength; i++ ){
				node = ( ClassCreate *)(list.listAt( i )->item);	
				if( !strcmp( node->className, item ) && node->typeFlag == 2 ){
					ur->setError( PARSE_ERROR, TYPE_INHERITANCE, node->className );
                                	return FALSE;
                        	}
			}
		}
                item = ( char * )(p->getNextEntry());
        }

        return TRUE; 
}

int DeclStmnt::inheritanceConflicts()
{
	ClassCreate	*p;
	int i, listLength = list.size;
	/* cyclic inheritance check */

	/* type inheritance error */
	for( i = 0; i < listLength; i++ ){
		p = ( ClassCreate *)(list.listAt(i )->item);
		if( !checkInheritanceList( p->inheritList ) ) return FALSE;
	}

	return TRUE;
}

int DeclStmnt::semanticCheck()
{
	if( !cycleCheck() ) return FALSE;
	if( !typeNameSemantics() ) return FALSE;
	if( !multipleDeclarationCheck() ) return FALSE;
	if( !inheritanceConflicts() ) return FALSE;
	return TRUE;
}


int DeclStmnt::parse()
{
	ClassCreate *p;

	connectUserRequest();

//	check_token( begin, begin_kywd_expected, " " );
	CHECK_TOKEN( DECLARATION, DECLARATION_KYWD_EXPECTED, " " );
	CHECK_TOKEN( LCB, LCB_EXPECTED, " " );

	getToken();
	while( tokenType() == CREATE ){
		getToken();	
		if( !( tokenType() == CLASS || tokenType() == TYPE ) ){
			ur->setError( PARSE_ERROR, CLASS_KYWD_ERR, " " );
			return FALSE;
		}
	
		if( tokenType() == TYPE ) p = new ClassCreate( ur, 2 );
		else p = new ClassCreate( ur, 1 );	

		if( !p->createClassParser() ) return FALSE;

		list.insert( CLSCONS, (void *)p );

		getToken();
	}

	connectUserRequest();

	if( tokenType()!=RCB ){
		ur->setError( PARSE_ERROR, RCB_EXPECTED, " " );
                return FALSE;
	}

	return TRUE;
}

int DeclStmnt::simpleParse()
{
	ClassCreate *p;

	connectUserRequest();

	getToken();	
	if( !( tokenType() == CLASS || tokenType() == TYPE ) ){
		ur->setError( PARSE_ERROR, CLASS_KYWD_ERR, " " );
		return FALSE;
	}
	
	if( tokenType() == TYPE ) p = new ClassCreate( ur, 2 );
	else p = new ClassCreate( ur, 1 );	

	if( !p->createClassParser() ) return FALSE;

	list.insert( CLSCONS, (void *)p );

	connectUserRequest();

	return TRUE;
}

int DeclStmnt::createClasses( )
{
	int i, listLength = list.size;
	ClassCreate *p;
	MoodsTypes  mt;
	CatBOOLEAN status = FALSE;

	if( !semanticCheck() ) return FALSE;

	/* create the empty classes beforehand */
	for( i = 0; i < listLength; i++ ){
		p = ( ClassCreate *)(list.listAt(i)->item);
		if( !mt.createEmptyClass( p->className ) ) return FALSE; 
	}

	/* fill the classes that were created empty */
	for( i = 0; i < listLength; i++ ){
		p = ( ClassCreate *)(list.listAt(i)->item);
                if( !mt.fillEmptyClass( p->className, p->inheritList, p->members,p->typeFlag, &status ) ) return FALSE;
	}

	return TRUE;
}

