#ifndef lint
static	char sccsid[] = "@(#)play.c 1.2 90/01/02 Copyr 1989 Sun Micro";
#endif
/* Copyright (c) 1989 by Sun Microsystems, Inc. */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <signal.h>

#include <stropts.h>
#include <sys/ioctl.h>

#include <libaudio.h>
#include <audio_device.h>
#include <audio_filehdr.h>

#define	Error		(void) fprintf


/* Local variables */
char prog1[80];

#define	MAX_GAIN		(100)		/* maximum gain */

/*
 * This defines the tolerable sample rate error as a ratio between the
 * sample rates of the audio data and the audio device.
 */
#define	SAMPLE_RATE_THRESHOLD	(.01)

unsigned        Volume = ~0;            /* output volume */
double		Savevol;		/* saved volume level */
int		Verbose = FALSE;	/* verbose messages */
int		Immediate = FALSE;	/* don't hang waiting for device */
char		*Audio_dev = "/dev/audio";

int		Audio_fd = -1;		/* file descriptor for audio device */
Audio_hdr	Dev_hdr;		/* audio header for device */

int reconfig(Audio_hdr);

/* Global variables */


void sigint()
{
	/* flush output queues before exiting */
	if (Audio_fd >= 0) {
		(void) audio_flush_play(Audio_fd);
		if (Volume != ~0)
			(void) audio_set_play_gain(Audio_fd, &Savevol);
		(void) close(Audio_fd);			/* close output */
	}
}

/*
 * Play an audio file.
 */
PlaySound(unsigned char *buf, long fsize, unsigned V_level)
{
	int		i;
	int		err;
	double		vol;
	struct stat	st;
	struct sigvec	vec;
	unsigned char  *temp_filehdr; 
        Audio_hdr       File_hdr;
        unsigned        ilenp;

	if( buf == NULL ) return 1;


	err = 0;
        temp_filehdr = (unsigned char *)malloc(sizeof(Audio_filehdr));
        memcpy((char *)temp_filehdr, (char *)buf, sizeof(Audio_filehdr));

        err = audio_decode_filehdr(temp_filehdr, &File_hdr, &ilenp);
        if (err < 0) {
		Error(stderr, "%s: cannot decode the file header.",prog1);
		return(1);
	}
 
        buf += (sizeof(Audio_filehdr)+ilenp);
        fsize -= (sizeof(Audio_filehdr)+ilenp);
	strcpy(prog1, "Playsound");	/* save program initiation name */
        Volume = V_level;

	/* Validate and open the audio device */
	err = stat(Audio_dev, &st);
	if (err < 0) {
		Error(stderr, "%s: cannot stat ", prog1);
		perror(Audio_dev);
		return(1);
	}
	if (!S_ISCHR(st.st_mode)) {
		Error(stderr, "%s: %s is not an audio device\n", prog1,
		    Audio_dev);
		return(1);
	}

	/* Try it quickly, first */
	Audio_fd = open(Audio_dev, O_WRONLY | O_NDELAY);
	if ((Audio_fd < 0) && (errno == EBUSY)) {
		if (Immediate) {
			Error(stderr, "%s: %s is busy\n", prog1, Audio_dev);
			return(1);
		}
		if (Verbose) {
			Error(stderr, "%s: waiting for %s...", prog1, Audio_dev);
			(void) fflush(stderr);
		}
		/* Now hang until it's open */
		Audio_fd = open(Audio_dev, O_WRONLY);
		if (Verbose)
			Error(stderr, "%s\n", (Audio_fd < 0) ? "" : "open");
	}
	if (Audio_fd < 0) {
		Error(stderr, "%s: error opening ", prog1);
		perror(Audio_dev);
		return(1);
	}

	/* Get the device output encoding configuration */
	if (audio_get_play_config(Audio_fd, &Dev_hdr) != AUDIO_SUCCESS) {
		Error(stderr, "%s: %s is not an audio device\n",
		    prog1, Audio_dev);
		return(1);
	}

	/* If -v flag, set the output volume now */
	if (Volume != ~0) {
		vol = (double) Volume / (double) MAX_GAIN;
		(void) audio_get_play_gain(Audio_fd, &Savevol);
		err = audio_set_play_gain(Audio_fd, &vol);
		if (err != AUDIO_SUCCESS) {
			Error(stderr,
			    "%s: could not set output volume for %s\n",
			    prog1, Audio_dev);
			return(1);
		}
	}

	/* Set up SIGINT handler to flush output */
	vec.sv_handler = sigint;
	vec.sv_mask = 0;
	vec.sv_flags = 0;
	(void) sigvec(SIGINT, &vec, (struct sigvec *)NULL);


	/* Check the device configuration */
	if (audio_cmp_hdr(&Dev_hdr, &File_hdr) != 0) {
		/*
		 * The device does not match the input file.
		 * Wait for any old output to drain, then attempt
		 * to reconfigure the audio device to match the
		 * input data.
		 */
		if (audio_drain(Audio_fd, FALSE) != AUDIO_SUCCESS) {
			Error(stderr, "%s: ", prog1);
			perror("AUDIO_DRAIN error");
			return(1);
		}
		if (!reconfig(File_hdr))
			goto close_dev;
	}

	/*
	 * At this point, we're all ready to output the data.
	 */

        if (fsize >= 0) {
        	/* If input EOF, write an eof marker */
        	err = write(Audio_fd, (char *)buf, fsize);

        	if (err != fsize) {
                	Error(stderr, "%s: output error: ", prog1);
                	perror("");
                }
        }


	/*
	 * Though drain is implicit on close(), it's performed here
	 * for the sake of completeness, and to ensure that the volume
	 * is reset after all output is complete.
	 */
close_dev:
        (void) audio_drain(Audio_fd, FALSE);
	if (Volume != ~0) {
		(void) audio_set_play_gain(Audio_fd, &Savevol);
	}
	
	(void) close(Audio_fd);			/* close output */
	free(temp_filehdr);
	return(1);
/*NOTREACHED*/
}


/*
 * Try to reconfigure the audio device to match the file encoding.
 * If this fails, we should attempt to make the input data match the
 * device encoding.  For now, we give up on this file.
 *
 * Returns TRUE if successful.  Returns FALSE if not.
 */
reconfig(Audio_hdr File_hdr)
{
	int	err;
	char	msg[AUDIO_MAX_ENCODE_INFO];

	Dev_hdr = File_hdr;
	err = audio_set_play_config(Audio_fd, &Dev_hdr);

	switch (err) {
	case AUDIO_SUCCESS:
		return (TRUE);

	case AUDIO_ERR_NOEFFECT:
		/*
		 * Couldn't change the device.
		 * Check to see if we're nearly compatible.
		 * audio_cmp_hdr() returns >0 if only sample rate difference.
		 */
		if (audio_cmp_hdr(&Dev_hdr, &File_hdr) > 0) {
			double	ratio;

			ratio = (double) abs((int)
			    (Dev_hdr.sample_rate - File_hdr.sample_rate)) /
			    (double) File_hdr.sample_rate;
			if (ratio <= SAMPLE_RATE_THRESHOLD) {
				if (Verbose) {
					Error(stderr,
			    "%s: WARNING: Audio file sampled at %d, playing at %d\n",
					    prog1, File_hdr.sample_rate,
					    Dev_hdr.sample_rate);
				}
				return (TRUE);
			}
			Error(stderr, "%s: Audio file sample rate %d not available\n",
			    prog1, File_hdr.sample_rate);
			return (FALSE);
		}
		(void) audio_enc_to_str(&File_hdr, msg);
		Error(stderr, "%s: Audio file encoding not available: %s\n",
		    prog1, msg);
		return (FALSE);

	default:
		Error(stderr, "%s: i/o error (set config)\n");
		return(1);
/*NOTREACHED*/
	}
}


