#include "opt_defs.h"
#include "dbi_gen.h"
#include "opt.h"
#include "dbi.h"
#include "global.h"
#include <malloc.h>

extern displayFID( FID * );
extern char *myAlloc( int );

void printJoinArg( JOIN_ARG* pred )
{
	if( !pred ) return;
	printJoinArg( (JOIN_ARG *) pred->left );
	switch( pred->type )
	{
		case OPER :
			switch( pred->data.op )
			{
				case EQUAL_OP : printf(" = "); break;
				case LESS_THAN_OP : printf(" < "); break;
				case GREATER_THAN_OP : printf(" > "); break;
				case LESS_THAN_EQUAL_OP : printf(" <= "); break;
				case GREATER_THAN_EQUAL_OP : printf(" >= "); break;
				case NOT_EQUAL_OP : printf(" <> "); break;
				case AND_OP : printf(" AND ");
				default : printf(" UNDEF OP ");
			}
			break;
		case ATTRIB :
			printf(" %s ", pred->data.attr );
			break;
			
		case CONST :
			switch( pred->data.cnst.kind )
			{
				case REAL :
				   printf(" %f ",pred->data.cnst.value.real_value );
				   break;
				case TEXT :
				    printf(" %s ",pred->data.cnst.value.text_value );
				    break;
				case INTEGER :
				   printf(" %d ",pred->data.cnst.value.int_value );
				   break;
				default :
					printf(" UNDEF CONS ");
			}	
		default :
			printf(" UNDEF TYPE  : %d ", pred->type );
	}
	printJoinArg( (JOIN_ARG *) pred->right );
	
}

void RELATION_SCHEMA::copy( RELATION_SCHEMA *copyFrom )
{
	int i = 0;
	*this = *copyFrom;
	for(i=0; i<copyFrom->arity; i++) 
	{
		attr[i] = (ATTRIBUTE * ) myAlloc( sizeof(ATTRIBUTE));
		*attr[i] = *(copyFrom->attr[i]);
	}

}

ATTRIBUTE* RELATION_SCHEMA::contains( char *attName )
{
	int i;
	for(i=0; i<arity; i++ )
		if( !strcmp(attName, attr[i]->name) )
			return attr[i];	
	return NULL;
}

/*RELATION_SCHEMA::RELATION_SCHEMA()
{
	arity = 0;
	cardinality = 0;
	record_width = 0;
}

RELATION_SCHEMA::~RELATION_SCHEMA()
{
	int i;
       	for(i=0; i<arity; i++ ) myFree(  attr[i] );
}


RELATION_SCHEMA::freeAll()
{
	int i;
	for(i=0; i<arity; i++ )
		free attr[i];
}*/

void RELATION_SCHEMA::printOn()
{
        int i;
        ATTRIBUTE *a;
 
        printf("name : %s  arity : %d card : %d  width : %d typeId : %d\n",
                name, arity, cardinality, record_width, TypeId );
 
        printf(" ATTRIBUTES :\n");
	printf("%20s %20s %4s %5s %6s %5s %6s\n",
		"NAME", "CLASS NAME", "TYPE", "WIDTH", "TOTREF", "BTREE", "OFFSET" );

        for(i=0; i<arity; i++ ) {
                a = attr[i];
		printf("%20s %20s %4d %5d %6.2f %5d %6d\n",
		a->name, a->class_name, a->attr_type, a->attr_width, a->totref, a->btree_avail, a->offset );
        } // for
 
}

/*
void RELATION_SCHEMA::printOn()
{
	int i;
	printf(" Name : %s \n",  name );
	printf(" cardinality : %d \n", cardinality );
	printf(" size        : %d \n", record_width );
	displayFID( &fid );
	printf(" ATTRIBUTES :\n");
	for(i=0; i<arity; i++ )
	{
		printf(" Att Name    : %s\n", attr[i]->name );
		printf(" Cls Name    : %s\n", attr[i]->class_name );
		printf(" type        : %d\n", attr[i]->attr_type );
	} 
}
*/


void printPlan( QUERY *aplan )
{
	if(!aplan) return;
	switch(aplan->my_operator )
	{
		case UNION :
			printf(" operator : UNION \n");
			//printJoinArg( &aplan->argument.join);
			printPlan( (QUERY *) aplan->input[0] );
			printPlan( (QUERY *) aplan->input[1] );
			break;
		case JOIN :
			printf(" operator : JOIN \n");
			printJoinArg( &aplan->argument.join);
			printPlan( (QUERY *) aplan->input[0] );
			printPlan( (QUERY *) aplan->input[1] );
			break;
		case SELECT :
			printf(" operator : SELECT \n");
			printJoinArg( aplan->argument.select.cond[0] );
			printf("\n");
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case GET_SET :
			printf(" operator : GET_SET \n");
			printf(" attribute : %s \n", aplan->argument.get_set.attr );
			if( aplan->argument.get_set.use_self )
				printf(" USE SELF : TRUE \n");
			else
				printf(" USE SELF : FALSE \n");
			aplan->argument.get_set.input.printOn();
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case MAT :
			printf(" operator : MAT \n");
			printf(" attribute : %s\n ", aplan->argument.mat.attr );
			aplan->argument.mat.input[0].printOn();
			printPlan( (QUERY *) aplan->input[0] );
			break;
		case UNNEST :
			printf(" operator : UNNEST \n");
			printPlan( (QUERY *) aplan->input[0] );
			break;
		default :
			printf(" Unknown Operator \n ");
	}
} 
	
		
